; Routines only used by bytecode instructions in the range $E0 to $FF.

; Waits until beam reaches bottom of screen following next interrupt.
; Registers unaffected.
WAIT_BEAMLOW:
PUSH AF                      ; Preserve registers.
PUSH BC
HALT                         ; Beam now at top of screen.
LD BC, 1923                  ; Closest to delay target of 50,000 cycles.
                             ; (Assumes 5,000-ish cycles for scheduler etc.)
  WAIT_BEAMLOW_JP1:                ; Each pass, except last, is 26 Ts.
  DEC BC                     ; 6 Ts.
  LD A, B                    ; 4 Ts.
  OR C                       ; 4 Ts.
  JR NZ, WAIT_BEAMLOW_JP1    ; 12 Ts (if true.)
  POP BC                     ; Restore registers.
  POP AF
  RET

; Creates eight character form number in TextVar A.
CREATE_FORMNUM:
LD B, A                      ; TextVar.
LD HL, VT_CREATE_FORMNUM
PUSH HL                      ; Store copy of variable table start.
LD A, R
RRC A
LD (HL), A                   ; RND_0.
INC HL
CALL RNG
LD A, (V_RNG_NUMBER)
LD (HL), A                   ; RND_1.
INC HL
LD A, R
RLC A
LD (HL), A                   ; RND_2.
LD HL, I_TXTOKENS
LD A, B                      ; TextVar.
CALL GET_ADDRESS_UNALIGNED   ; HL now points to first byte of TextVar.
POP DE                       ; Restore copy of variable table start.

LD C, 75                     ; ASCII code for "K".
LD B, 00001111B              ; Mask out upper four bits.
LD A, (DE)                   ; RND_0
AND B
ADD A, C                     ; A now contains "K" + 4 bit offset (K to Z.)
LD (HL), A                   ; Store first letter.
INC DE
INC HL
LD A, (DE)
AND B
ADD A, C
LD (HL), A                   ; Store second letter.
DEC DE

LD IX, VT_CREATE_FORMNUM
LD (IX+3), 3                 ; PASSCOUNT.
LD C, 48                     ; ASCII code for "0".
LD B, 00000111B              ; Mask out upper three bits.

  CREATE_FORMNUM_JP1:
  LD A, (DE)                 ; Get next random number.
  PUSH AF
  AND B                      ; Mask out upper four bits.
  ADD A, C
  INC HL
  LD (HL), A                 ; Store digit.
  POP AF                     ; Restore BCD.
  SRA A                      ; Shift upper number to lower number.
  SRA A
  SRA A
  SRA A
  AND B                      ; Mask out upper four bits.
  ADD A, C
  INC HL
  LD (HL), A                 ; Store digit.
  INC DE
  DEC (IX+3)                 ; PASSCOUNT.
  JR NZ, CREATE_FORMNUM_JP1
  RET

VT_CREATE_FORMNUM:
DB 0                         ;+0 RND_0
DB 0                         ;+1 RND_1
DB 0                         ;+2 RND_2
DB 0                         ;+3 PASSCOUNT

; Assigns each group of people to a Department (four possible outcomes.)
; Random number generator must be called first.
SET_PEOPLE:
LD HL, V_GVAR                ; RNG output is stored in first entry of table.
LD A, 00000011B              ; Mask off all but lower two bits.
AND (HL)                     ; Each entry is two bytes, only lower one needed.
SLA A                        ; Multiply by four.
SLA A
ADD A, 16                    ; Add to starting value.
LD L, 64                     ; HL $00xx aligned, no overflow, now at 32nd entry.
LD C, 16                     ; Number of people to be assigned.
LD B, 32                     ; Wraparound value.


  SET_PEOPLE_JP1:
  LD (HL), A                 ; Assign person to role and department (low byte.)
  INC HL                     ; Go to high byte.
  LD (HL), 0                 ; Clear high byte (just in case.)
  INC HL                     ; Go to next person.
  INC A                      ; Next role and department.
  CP B                       ; Reached wraparound?
  JR NZ, SET_PEOPLE_JP2
  LD A, 16                   ; Wraparound to starting value.

  SET_PEOPLE_JP2:
  DEC C                      ; Last person assigned?
  JR NZ, SET_PEOPLE_JP1
  RET

; Display current time in HH:MM:SS:FF format as defined in VT_TIMER_CLOCK.
; The output can be clipped at both ends, e.g. to only show minutes and seconds.
SHOW_TIME:
; Not used, removed to save memory.
RET

;LD HL, V_TIMER_CLOCK_HOURS
;LD DE, DOUBLE_TO_STRING_BUFFER      ; Temporarily redirect double string output.
;LD (DOUBLE_TO_STRING_ADDRESS), DE
;PUSH DE                             ; Copy unit buffer start address to IX.
;POP IX
;LD DE, MB_PS_NEXTROW                ; Combined string buffer start.
;LD C, 4                             ; Loop counter (four numbers to transfer.)
;
;  SHOW_TIME_JP1:
;  PUSH HL                           ; Store time unit pointer.
;  LD L, (HL)                        ; Copy value to HL
;  LD H, 0
;  CALL DOUBLE_TO_STRING
;  POP HL                            ; Restore time unit pointer.
;  DEC HL                            ; Move to next unit (decreasing order.)
;  LD A, (IX+3)                      ; Get 4th unit character.
;  LD (DE), A                        ; Copy to combined string buffer.
;  INC DE
;  LD A, (IX+4)                      ; Get 5th unit character.
;  LD (DE), A                        ; Copy to combined string buffer.
;  INC DE
;  DEC C                             ; All unit strings processed?
;  JR Z, SHOW_TIME_JP2
;  LD A, ":"                         ; Add colon between strings.
;  LD (DE), A
;  INC DE
;  JR SHOW_TIME_JP1                  ; Process next time unit.
;
;  SHOW_TIME_JP2:
;  LD DE, MB_PS_NEXTROW              ; Restore double string output address.
;  LD (DOUBLE_TO_STRING_ADDRESS), DE
;
;  PUSH DE
;  POP HL                            ; DE and HL now at start of combined string.
;  LD A, (V_SHOW_TIME_XOFFSET)
;  LD B, 0
;  LD C, A                           ; BC now contains XOFFSET.
;  ADD HL, BC                        ; Increase HL accordingly.
;  LD C, 11                          ; Maximum output length.
;  LDIR                              ; Move string along buffer.
;
;  LD DE, VT_SHOW_TIME               ; Redirect PS_DRAWROW variable source.
;  LD (V_PS_DRAWROWVT), DE
;  CALL PS_DRAWROW                   ; Draw timer string.
;  LD DE, VT_PS_DRAWROW              ; Restore PS_DRAWROW variable source.
;  LD (V_PS_DRAWROWVT), DE
;  RET
;
VT_SHOW_TIME:
;DB 0                           ;+0 XSTART
;DB 0                           ;+1 YSTART
;DB 11                          ;+2 XLENGTH
;
V_SHOW_TIME_XOFFSET:
;DB 0


; Called with number to dial in the A register.
DIAL_NUMBER:
LD HL, I_PHONENOS              ; N.B. max. phoneno data length is 255.
CALL INDEX_TO_ADDRESS          ; HL now holds phoneno pointer, C data length.
LD DE, M_SCRATCHPAD            ; Pointer to where sequence will be generated.
LD (V_SET_SOUNDSEQ_PTR), DE
LD IX, VT_DIAL_NUMBER
LD B, (IX+0)                   ; Discard high length byte (so max 255 length.)

INC B
DEC B
JR Z, DN_TONEDIAL              ; If PHNUMREF was 0.
DEC B
JR Z, DN_PULSEDIAL             ; If PHNUMREF was 1.
DEC B
JR Z, DN_ROTDIAL               ; If PHNUMREF was 2.
JR DN_PULSEDIAL                ; If PHNUMREF was 3 (Payphone, also pulse dial.)

  DN_ROTDIAL
  LD B, (IX+15)                ; ROTBMPXPOS.
  LD C, (IX+16)                ; ROTBMPYPOS.
  LD (V_LASTDRAWPOS), BC       ; Pass XY for dial animation.

  DN_ROTSEQ:                   ; If PHNUMREF was 2.
  LD A, (IX+12)                ; Pass STARTROTBMP...
  LD (IX+13), A                ; ...to NEXTROTBMP.
  LD A, (IX+11)                ; Pass REWINDGAP...
  LD (IX+1), A                 ; ...to SOUNDGAP.
  PUSH BC
  LD B, (IX+10)                ; REWINDSND.

  LD (IX+14), 1                ; Set STEPROTBMP to +1.
  CALL DNRD_NEXTNUM            ; Write rewind sound sequence for digit.

  DEC HL                       ; Repeat digit.
  LD A, (IX+7)                 ; Pass PULSEGAP...
  LD (IX+1), A                 ; ...to SOUNDGAP.
  LD B, (IX+9)                 ; UNWINDSND.

  DEC (IX+13)                  ; Compensate for surplus animation frame INC.
  LD (IX+14), 255              ; Set STEPROTBMP to -1.
  CALL DNRD_NEXTNUM            ; Write unwind sound sequence for digit.

  POP BC
  DEC C
  JR NZ, DN_ROTSEQ
  JR DN_STARTSEQ

  DN_TONEDIAL:
  CALL DNTD_NEXTNUM
  DEC C
  JR NZ, DN_TONEDIAL
  JR DN_STARTSEQ

  DN_PULSEDIAL:
  PUSH BC
  LD A, (IX+7)                 ; Pass PULSEGAP...
  LD (IX+1), A                 ; ...to SOUNDGAP.
  LD B, (IX+6)                 ; PULSESND.
  CALL DNPD_NEXTNUM
  POP BC
  DEC C
  JR NZ, DN_PULSEDIAL

  DN_STARTSEQ:
  LD A, (IX+2)                 ; EOSCODE.
  LD (DE), A                   ; Mark end of sequence.
  LD IX, VT_SOUNDSEQ
  LD (IX+3), 1                 ; Set RPTCOUNT to 1.
  CALL RESET_SOUNDSEQ
  RET

DNTD_NEXTNUM:
LD A, (HL)                     ; Get next digit.
ADD A, (IX+4)                  ; Offset by DTOFFSET to get sound number.
LD (DE), A                     ; Write sound number to sequence.
INC DE                         ; Move sequence pointer.
LD A, (IX+5)                   ; DTDELAY.
LD (DE), A                     ; Add delay to sequence.
INC DE                         ; Move sequence and phoneno pointers.
INC HL
RET

DNPD_NEXTNUM:
XOR A
LD C, (HL)                     ; Get next digit.
CP C
JR NZ, DNPD_NEXTNUM_JP1
LD C, 10                       ; Zero corresponds to ten pulses.

  DNPD_NEXTNUM_JP1:
  LD A, B                      ; Copy sound.
  LD (DE), A                   ; Write sound to sequence.
  INC DE                       ; Move sequence pointer.
  LD A, (IX+1)                 ; SOUNDGAP.
  LD (DE), A                   ; Add gap between pulses to sequence.
  INC DE                       ; Move sequence pointer.
  DEC C                        ; Pulse counter.
  JR NZ, DNPD_NEXTNUM_JP1
  LD A, (IX+8)                 ; PSEQGAP.
  DEC DE
  LD (DE), A                   ; Overwrite gap after pulse sequence.
  INC DE                       ; Move sequence and phoneno pointers.
  INC HL
  RET

; Per DNPD_NEXTNUM, but with rotary dial animation frame insertion.
DNRD_NEXTNUM:
LD A, (HL)                     ; Get next digit.
ADD A, 4                       ; Offset for distance between rest and stop.
LD C, A
CP 4                           ; Was digit zero before offset?
JR NZ, DNRD_NEXTNUM_JP1
LD C, 14                       ; Zero corresponds to ten pulses plus offset.

  DNRD_NEXTNUM_JP1:
  LD A, B                      ; Copy sound.


  LD A, 64                     ; "Draw again" code for dial animation.
  LD (DE), A
  INC DE                       ; Move sequence pointer.
  LD A, (IX+13)                ; Dial animation frame to draw.
  LD (DE), A
  INC DE                       ; Move sequence pointer.
  ADD A, (IX+14)               ; Add STEPROTBMP to get NEXTROTBMP.
  LD (IX+13), A                ; Store updated NEXTROTBMP.

  LD A, B                      ; Restore sound.

  LD (DE), A                   ; Write sound to sequence.
  INC DE                       ; Move sequence pointer.
  LD A, (IX+1)                 ; SOUNDGAP.
  LD (DE), A                   ; Add gap between pulses to sequence.
  INC DE                       ; Move sequence pointer.

  DEC C                        ; Pulse counter.
  JR NZ, DNRD_NEXTNUM_JP1
  LD A, (IX+8)                 ; PSEQGAP.
  DEC DE
  LD (DE), A                   ; Overwrite gap after pulse sequence.
  INC DE                       ; Move sequence and phoneno pointers.
  INC HL
  RET

; Create sequence in M_SCRATCHPAD, then pass pointer to RESET_SOUNDSEQ.
; - will need to tweak RESET_SOUNDSEQ to take a custom pointer (e.g. DE > 0.)

VT_DIAL_NUMBER:
DB 0                           ;+0 PHONETYPE.
DB 0                           ;+1 SOUNDGAP.
DB 127                         ;+2 EOSCODE.
DB 0                           ;+3 (Spare.)
DB 32                          ;+4 DTOFFSET.
DB 2                           ;+5 DTDELAY. (Was 10)
DB 15                          ;+6 PULSESND.
DB 2                           ;+7 PULSEGAP.
DB 20                          ;+8 PSEQGAP.
DB 2                           ;+9 UNWINDSND.
DB 12                          ;+10 REWINDSND.
DB 1                           ;+11 REWINDGAP.
DB 16                          ;+12 STARTROTBMP.
DB 0                           ;+13 NEXTROTBMP.
DB 0                           ;+14 STEPROTBMP.
DB 8                           ;+15 ROTBMPXPOS.
DB 6                           ;+16 ROTBMPYPOS.

; Manual entry, as only a few phone numbers.
I_PHONENOS:
DW PHONENO_000
DW PHONENO_001
DW PHONENO_002
DW PHONENO_003
DW PHONENO_004

PHONENO_000:
DB 0,3,7,4,5,7,8,3,4

PHONENO_001:
DB 0,3,7,4,5,9,8,1,4

PHONENO_002:
DB 0,3,7,4,5,5,0,3,4

PHONENO_003:
DB 0,3,7,4,5,3,9,4,3

; Null final entry (for penultimate entry's length calculation.)
PHONENO_004:


; Called with phone type in the A register, repeat length in the E register.
RING_PHONE:
LD HL, M_SCRATCHPAD            ; Set pointer to sequence generation start.
LD (V_SET_SOUNDSEQ_PTR), HL
LD IX, VT_RING_PHONE
ADD A, (IX+0)                  ; Add SNDSTART to phone type.
LD D, A                        ; D now holds the relevant sound to use.
LD A, (IX+1)                   ; PULSEFOR.

LD C, (IX+2)                   ; PAUSEONE.
CALL RP_PULSE                  ; Generate first pulse, including pause at end.
LD C, (IX+3)                   ; PAUSETWO.
CALL RP_PULSE                  ; Generate second pulse, including pause at end.
LD (HL), 127                   ; Mark end of sequence.
LD IX, VT_SOUNDSEQ
;LD (IX+3), 2                   ; Set RPTCOUNT to 2.
LD (IX+3), E                   ; Set RPTCOUNT.
CALL RESET_SOUNDSEQ            ; Queue sequence to start playing next frame.
RET                            ; Done.

RP_PULSE:
LD B, A                        ; Reset pulse counter.

  RP_PULSE_JP1:
  LD (HL), D                   ; Sound.
  INC HL
  LD (HL), 1                   ; One frame of silence.
  INC HL
  DJNZ RP_PULSE_JP1            ; Next period in pulse.

  DEC HL                       ; Overwrite pause at end of pulse.
  LD (HL), C
  INC HL
  RET

VT_RING_PHONE:
DB 22                        ;+0 SNDSTART
DB 10                        ;+1 PULSEFOR.
DB 10                        ;+2 PAUSEONE.
DB 100                       ;+3 PAUSETWO.


; Draws vertical beam line up or downwards, with sound effect.
; Downwards masks against the source image, upward copies the source image.
; Called with "Down" variable in the A register.
VFX_BEAMV:
LD IX, VT_VFX_BEAMV
LD B, 1
CP B                           ; Beam going down?
JR Z, VFX_BEAMV_JP1
LD B, 255                      ; Otherwise, YVEL is negative (-1 on 8 bit add.)

  VFX_BEAMV_JP1:
  LD (IX+6), B                 ; Set YVEL.

  VFX_BEAMV_JP2:               ; Start of cell row.
  LD B, (IX+0)                 ; XFROM.
  LD C, (IX+1)                 ; YFROM.
  CALL GET_SCR_CELLSTART
  EX DE, HL                    ; HL now contains left cell of destination row.
  LD C, (IX+5)                 ; YSOURCE.
  CALL GET_SCR_CELLSTART       ; DE now contains left cell of source row.
  LD C, 8                      ; Eight bytes per row.
  BIT 7, (IX+6)                ; Is YVEL negative?
  JR Z, VFX_BEAMV_JP3          ; If not, start from top of cell.
  LD B, 7                      ; If not, move HL and DE to bottom of cell.
  LD A, D                      ; (Achieved by adding 256*7.)
  ADD A, B
  LD D, A
  LD A, H
  ADD A, B
  LD H, A

  VFX_BEAMV_JP3:
  LD A, (IX+7)                 ; DELAY.
  INC A                        ; Allow for refresh synchronisation halt.

  VFX_BEAMV_JP3_DELAY:         ; Start of byte row within cell.
  HALT                         ; Wait for refresh.
  DEC A
  JR NZ, VFX_BEAMV_JP3_DELAY

  EXX                          ; REG'START
  LD A, (IX+8)                 ; 'SOUND.
  LD DE, 0                     ; 'Clear sound duration override.
  CALL PLAY_SOUND              ; 'Play sound effect.
  EXX                          ; REG'END

  LD B, (IX+2)                 ; XLENGTH.
  PUSH HL                      ; Preserve destination pointer.
  PUSH DE                      ; Preserve source pointer

  VFX_BEAMV_JP4:               ; Row processing.
  LD A, (DE)                   ; Get source byte.
  BIT 7, (IX+6)                ; Is YVEL negative?
  JR NZ, VFX_BEAMV_JP5         ; If it is, don't process the source byte.

  XOR (HL)                     ; Mask destination against the source.
  OR (IX+4)                    ; Apply LINEMASK for vertical stripe effect.

  VFX_BEAMV_JP5:
  LD (HL), A                   ; Set destination byte.
  INC HL                       ; Next byte in destination row.
  INC DE                       ; Next byte in source row.
  DJNZ VFX_BEAMV_JP4

  POP DE                       ; Restore source pointer.
  POP HL                       ; Restore destination pointer.
  LD B, (IX+6)                 ; YVEL.
  LD A, H                      ; Move up/down one row within destination cell.
  ADD A, B
  LD H, A
  LD A, D                      ; Move up/down one row within source cell.
  ADD A, B
  LD D, A
  DEC C                        ; Repeat for height of cell.
  JR NZ, VFX_BEAMV_JP3

  LD A, (IX+1)                 ; YFROM.
  ADD A, B                     ; Move up/down per YVEL to next destination row.
  LD (IX+1), A
  LD A, (IX+5)                 ; YSOURCE.
  ADD A, B                     ; Move up/down per YVEL to next source row.
  LD (IX+5), A
  DEC (IX+3)                   ; YLENGTH.
  JR NZ, VFX_BEAMV_JP2         ; Draw next cell row.
  RET


VT_VFX_BEAMV:
DB 0                    ;+0 XFROM.
DB 0                    ;+1 YFROM.
DB 0                    ;+2 XLENGTH.
DB 0                    ;+3 YLENGTH.
DB 10101010B            ;+4 LINEMASK.
DB 0                    ;+5 YSOURCE.
DB 0                    ;+6 YVEL.
DB 1                    ;+7 DELAY.
DB 12                   ;+8 SOUND.

; Working.
VFX_SCROLLPIXV:
LD IX, VT_VFX_SCROLLPIXV
LD A, (IX+9)                   ; SOUNDNUM.
CALL MODCOPY_SOUND
LD B, (IX+5)                   ; STEPS.
XOR A
CP (IX+6)                      ; YVEL.
JR NZ, VFX_SCROLLPIXV_JP1      ; If not zero, leave as-is.
LD (IX+6), 255                 ; Otherwise, set YVEL to -1 (rollover addition.)

  VFX_SCROLLPIXV_JP1:
  LD C, (IX+3)                 ; YLENGTHPX.
  LD D, (IX+1)                 ; YFROMPX.
  LD E, (IX+4)                 ; YSOURCEPX.
  LD A, (IX+7)                 ; Delay.

  VFX_SCROLLPIXV_JP2:
  CALL WAIT_BEAMLOW            ; Wait for refresh, then further to reduce tears.
  AND A                        ; Delay check.
  JR Z, VFX_SCROLLPIXV_JP3
  DEC A                        ; Reduce delay counter.
  JR VFX_SCROLLPIXV_JP2        ; Wait another frame.

  VFX_SCROLLPIXV_JP3:
  XOR A
  CP (IX+8)                    ; Check WITHSND.
  JR Z, VFX_SCROLLPIXV_JP5

  EXX                          ; Preserve registers.

  CP (IX+10)                   ; 'Check BENDSND.
  JR Z, VFX_SCROLLPIXV_JP4

  LD HL, SOUND_000_L           ; 'LOW phase delay.
  INC HL                       ; 'LOW phase delay (lower byte.)
  LD A, (IX+4)                 ; 'YSOURCEPX.
  LD (HL), A                   ; 'Set lower byte to YSOURCEPX.

  VFX_SCROLLPIXV_JP4:
  LD A, 0                      ; 'Play customised sound 0.
  LD DE, 0                     ; 'Clear sound duration override.
  CALL PLAY_SOUND

  EXX                          ; Restore registers.

  VFX_SCROLLPIXV_JP5:
  PUSH BC                      ; Preserve STEPS and YLENGTHPX.
  LD B, 0
  LD C, (IX+0)                 ; BC now contains XFROMCL.
  LD A, D                      ; YFROMPX.
  INC D                        ; Move YFROMPX down one pixel row.
  CALL GET_YPIXPTR
  ADD HL, BC                   ; Offset address by XFROMCL.
  LD A, E                      ; YSOURCEPX.
  INC E                        ; Move YSOURCEPX down one pixel row.
  PUSH DE                      ; Preserve YFROMPX and YSOURCEPX.
  EX DE, HL                    ; DE now points to left start of destination row.
  CALL GET_YPIXPTR

  LD C, (IX+11)                 ; BC now contains XSOURCECL.

  ADD HL, BC                   ; HL now points to left start of source row.

  LD C, (IX+2)                 ; BC now contains XLENGTHCL.
  LDIR                         ; Copy source row to destination row.
  POP DE                       ; Restore YFROMPX and YSOURCEPX.
  POP BC                       ; Restore STEPS and YLENGTHPX.
  DEC C                        ; Next row.
  JR NZ, VFX_SCROLLPIXV_JP5

  LD A, (IX+4)                 ; YSOURCEPX.
  ADD A, (IX+6)                ; YVEL.
  LD (IX+4), A                 ; YSOURCEPX now updated per YVEL.
  DEC B                        ; Next step.
  JR NZ, VFX_SCROLLPIXV_JP1
  RET                          ; Finished.

VT_VFX_SCROLLPIXV:
DB 0                    ;+0 XFROMCL.
DB 0                    ;+1 YFROMPX.
DB 0                    ;+2 XLENGTHCL.
DB 0                    ;+3 YLENGTHPX.
DB 0                    ;+4 YSOURCEPX.
DB 0                    ;+5 STEPS.
DB 0                    ;+6 YVEL.
DB 0                    ;+7 DELAY.
DB 0                    ;+8 WITHSND.
DB 0                    ;+9 SOUNDNUM.
DB 0                    ;+10 BENDSND.
DB 0                    ;+11 XSOURCECL.

; For single-channel sound modifications.
SFX_BENDSND:
LD IX, VT_SFX_BENDSND
LD A, (IX+0)                   ; SOURCESND.
CALL MODCOPY_SOUND
LD B, (IX+3)
LD C, (IX+2)                   ; BC now contains HPITCHSTEP.
LD D, (IX+5)
LD E, (IX+4)                   ; DE now contains LPITCHSTEP.

  SFX_BENDSND_JP1:
  EXX                          ; Preserve registers.
  LD DE, 32                    ; 'Hack to fix number of cycles.
  LD (V_SOUNDQUEUE_DURO), DE
  XOR A
  CALL PLAY_SOUND              ; 'Play customised Sound 0.
  EXX
  DEC (IX+1)                   ; STEPS.
  RET Z                        ; Done.

  LD HL, (SOUND_000_H)         ; Modify the HIGH phase delay.
  LD A, L                      ; Swap endianess, due to sound generator quirk.
  LD L, H
  LD H, A
  ADD HL, BC                   ; Addition will now give correct result.
  LD A, L                      ; Swap endianess, due to sound generator quirk.
  LD L, H
  LD H, A
  LD (SOUND_000_H), HL         ; Store swapped result.

  LD HL, (SOUND_000_L)         ; Modify the LOW phase delay.
  LD A, L                      ; Swap endianess, due to sound generator quirk.
  LD L, H
  LD H, A
  ADD HL, DE                   ; Addition will now give correct result.
  LD A, L                      ; Swap endianess, due to sound generator quirk.
  LD L, H
  LD H, A
  LD (SOUND_000_L), HL         ; Store swapped result.

  JR SFX_BENDSND_JP1           ; Next step.

VT_SFX_BENDSND:
DB 0                    ;+0 SOURCESND.
DB 0                    ;+1 STEPS.
DW 0                    ;+2 HPITCHSTEP.
DW 0                    ;+4 LPITCHSTEP.

; N.B. does not check for line overflow.
; - Done for simplicity at expense of visual glitches at far horizontal edges.
SHFTAREA_H:
LD IX, VT_SHFTAREA_H
XOR A
CP (IX+4)                       ; XPLUS.
JR NZ, SHAH_R
LD A, (IX+0)                    ; XSTART.
ADD A, (IX+2)                   ; XLENGTH.
LD (IX+0), A                    ; Left shift works from right to left.

  SHAH_L:
  CALL SHAH_ROWSTART            ; HL now points to top row of rightmost cell.
  INC HL                        ; Start from cell after shaking area.

    SHAH_L_JP1:
    PUSH HL                     ; Preserve starting position.
    LD B, (IX+2)                ; XLENGTH.
    LD A, (HL)                  ; Get first ReadByte.

    SHAH_L_JP2:
    RLA                         ; Set Carry to MSB of ReadByte.
    DEC HL                      ; Go to WriteByte.
    LD A, (HL)                  ; Store copy for use as next ReadByte.
    RL (HL)                     ; Shift WriteByte left, placing Carry in LSB.
    DJNZ SHAH_L_JP2             ; XLENGTH.

    POP HL                      ; Restore starting position.
    INC H                       ; Move down to next row of cell.
    DEC D                       ; Check lines per cell counter.
    JR NZ, SHAH_L_JP1           ; Next line.

    INC (IX+1)                  ; YSTART (move down one cell row.)
    DEC (IX+3)                  ; YLENGTH
    JR NZ, SHAH_L               ; Next cell row.
    RET                         ; Done.

  SHAH_R:
  CALL SHAH_ROWSTART            ; HL now points to top row of leftmost cell.
  DEC HL                        ; Start from cell before shaking area.

    SHAH_R_JP1:
    PUSH HL                     ; Preserve starting position.
    LD B, (IX+2)                ; XLENGTH.
    LD A, (HL)                  ; Get first ReadByte.

    SHAH_R_JP2:
    RRA                         ; Set Carry to LSB of ReadByte.
    INC HL                      ; Go to WriteByte.
    LD A, (HL)                  ; Store copy for use as next ReadByte.
    RR (HL)                     ; Shift WriteByte right, placing Carry in MSB.
    DJNZ SHAH_R_JP2             ; XLENGTH.

    POP HL                      ; Restore starting position.
    INC H                       ; Move down to next row of cell.
    DEC D                       ; Check lines per cell counter.
    JR NZ, SHAH_R_JP1           ; Next line.

    INC (IX+1)                  ; YSTART (move down one cell row.)
    DEC (IX+3)                  ; YLENGTH
    JR NZ, SHAH_R               ; Next cell row.
    RET                         ; Done.

  SHAH_ROWSTART:
  LD B, (IX+0)                  ; XSTART.
  LD C, (IX+1)                  ; YSTART.
  CALL GET_SCR_CELLSTART        ; DE now points to top row of cell.
  EX DE, HL                     ; HL now points to top row of cell.
  LD D, 8                       ; Eight lines per cell.
  RET                           ; Done.


VT_SHFTAREA_H:
DB 0                         ;+0 XSTART
DB 0                         ;+1 YSTART
DB 0                         ;+2 XLENGTH
DB 0                         ;+3 YLENGTH
DB 0                         ;+4 XPLUS

; Called with memory start address in HL, rumble length in C.
PLAY_RUMBLE:
LD B, 255                    ; Set initial click length.

  PLAY_RUMBLE_JP1:
  LD A, (HL)                 ; Get byte from memory.
  LD DE, 48                  ; Delay multiplier (for lower-frequency sound.)
  CALL MULT_DE_BY_A_QND
  INC DE                     ; Prevent zero values causing excessive delay.

  PLAY_RUMBLE_JP2:
  DEC DE                     ; Delay loop.
  LD A, D
  OR E
  JR NZ, PLAY_RUMBLE_JP2

  PUSH BC                    ; Preserve click length.
  CALL PLAY_CLICK            ; Play click of length B.
  POP BC                     ; Restore click length.
  INC HL                     ; Move to next byte in memory.
  DEC C                      ; Rumble length.
  JR NZ, PLAY_RUMBLE_JP1
  RET

; Draw pre-calculated sequence of lines.
HLINE_NEXT:
LD IX, VT_HLINE_NEXT
XOR A
CP (IX+2)                    ; WIPELAST.
JR Z, HLINE_NEXT_JP1
; To do: wipe current line.

  HLINE_NEXT_JP1:
  LD A, (IX+0)               ; LISTSTEP.
  INC A
  CP (IX+1)                  ; LISTLENGTH.
  JR NZ, HLINE_NEXT_JP2
  XOR A                      ; Restart at beginning of list.

  HLINE_NEXT_JP2:
  LD (IX+0), A               ; Store updated LISTSTEP.
  LD DE, 3                   ; List entry size (hard-coded.)
  CALL MULT_DE_BY_A_QND      ; DE now contains list offset.
  LD HL, (V_HLNEXT_ADDR)
  ADD HL, DE                 ; HL now points to list entry.
  LD DE, VT_HLINE            ; Copy list entry across.
  LD BC, 3
  LDIR
  CALL HLINE                 ; Draw line.
  RET                        ; Done.

VT_HLINE_NEXT:
DB 15                        ;+0 LISTSTEP.
DB 16                        ;+1 LISTLENGTH.
DB 0                         ;+2 WIPELAST.

V_HLNEXT_ADDR:
DW V_HLNEXT_LIST

V_HLNEXT_LIST:
DB 118, 57, 20
DB 81, 94, 94
DB 116, 59, 24
DB 116, 59, 24
DB 114, 61, 28
DB 114, 61, 28
DB 112, 63, 32
DB 112, 63, 32
DB 110, 65, 36
DB 110, 65, 36
DB 122, 54, 12
DB 105, 69, 46
DB 121, 55, 14
DB 100, 74, 56
DB 120, 56, 16
DB 94, 80, 68

; Fill attribute row A with data beginning at DE.
VFX_SETATTRSROW:
PUSH AF                      ; Preserve row number.
                             ; Fill attributes first.
PUSH DE                      ; Store source data address.
LD DE, 32                    ; Length of attibute row.
CALL MULT_DE_BY_A_QND        ; DE now contains destination offset.
LD HL, (V_ATTR_START)        ; Start of attributes map.
ADD HL, DE
EX DE, HL                    ; DE now points to start of attribute row.
POP HL                       ; Restore source data address.
LD BC, 32                    ; 32 bytes per attribute row.
LDIR                         ; Copy from source to row.
                             ; Fill pixels next.
POP AF                       ; Restore row number.
LD B, 0
LD C, A
CALL GET_SCR_CELLSTART       ; DE now contains start of pixel data.
LD A, 8                      ; Eight pixel rows per attribute cell.

  VFX_SETATTRSROW_JP1:
  LD BC, 32                  ; 32 bytes per pixel row.
  PUSH DE                    ; Preserve pixel row start.
  LDIR                       ; Copy from source to pixel row.
  POP DE                     ; Restore pixel row start.
  INC D                      ; Move destination down one pixel row.
  DEC A                      ; Next pixel row
  JR NZ, VFX_SETATTRSROW_JP1
  RET

; Fill screen pixel row with solid ink (255.)
; Starting from row B, fill C rows with D gaps between each row.
; - Tested and working, but not used, so removed to save memory.
;VFX_WIPEROWS:
;LD A, B
;CALL GET_YPIXPTR             ; HL now points to start of row.
;LD A, 11111111B              ; Value for all ink pixels.
;LD E, 32                     ; 32 bytes per row.
;
;  VFX_WIPEROWS_JP1:          ; Fill row.
;  LD (HL), A
;  INC HL
;  DEC E
;  JR NZ, VFX_WIPEROWS_JP1
;
;  LD A, B                    ; Move row down by gap.
;  ADD A, D
;  LD B, A
;  DEC C                      ; Fill next row.
;  JR NZ, VFX_WIPEROWS
;  RET Z
