; Super Enquiry Simulator - January 2018.
; Most code taken from the Xenoxx Optimised Revision.

; Full screen attributes buffer.
; - must be 0xXX00 aligned.
; - can only be used after BASIC call to machine code.
; - can instead be used as 1k scratchpad.
; - may need to expand for half-size images to allow buffered H/V flipping.
; WARNING: Do not use whole buffer, or stacks will be overwritten!
org $5B00
ATTR_BUFFER:
; Maximum size is 256 bytes, otherwise stack will be overwritten.
M_SCRATCHPAD:

; Used to store runtime-generated mapping table for flipped bytes (256.)
; - if necessary, can be used as temporary expansion for M_SCRATCHPAD.
org $5C00
FLIPX_LOOKUP:

; Stacks squashed down to 31 entries each. Unlikely to get near that deep...
; - N.B. PUSH_CALLSTACK in ZX-PARSER checks for max 128 entries.
; - N.B. PUSH_IFKYSTACK in ZX-PARSER checks for max 128 entries.
org $5D01
PSR_CALL_STACK:

org $5D3F
PSR_IFKY_STACK:

org $5D7D
STACKLIMIT:
; Last byte before IFKY_STACK is overwritten.

;org $5EFE
org $5DFE
STACKSTART:

org $5E00
V_GVAR:

; $5F00 to $5FFF will be overwritten by I_TXLINES after initialisation.
; I_TXLINES contains up to 128 runtime-generated text line addresses.
org $5F00
I_TXLINES:
LOADER:
; Headerless loader.
;LD IX, $4000        ; Screen start address.
;LD DE,6912          ; Screen data length.
;LD A, $FF
;SCF
;CALL $0556          ; ROM loading routine.
;LD IX, $5F20         ; Game data start address.
;LD DE,41184         ; data length.
;LD A, $FF
;SCF
;CALL $0556          ; ROM loading routine.
; End headerless loader.

org $5F20
ENTRY:
LD A, $AC              ; Interrupt address table is from AC00 to AD00 inclusive.
LD I, A
IM 2
LD SP, STACKSTART      ; Set stack pointer.
LD HL, V_GVAR          ; Wipe the global variables table.
PUSH HL                ; (Just in case any were not initialised in Nodes.)
POP DE
INC DE
LD (HL), 0
LD BC, 256
LDIR
CALL NODE_PARSER       ; Start the parser (never returns.)

org $6000
INCLUDE "SES-SOUNDS.asm"

; Currently 3 byte gap between sounds and text.

; Approximately 7.25k available for text, including all index tables.
org $61BC
I_TXTOKENS:
INCLUDE "Data/TXTOKENS-I.asm"

TXTOKENS:
INCLUDE "Data/TXTOKENS.asm"

I_TXDICTWD:
INCLUDE "Data/TXDICTWD-I.asm"

TXDICTWD:
INCLUDE "Data/TXDICTWD.asm"

I_TXBANKS:
INCLUDE "Data/TXBANKS-I.asm"

TXBANKS:
INCLUDE "Data/TXBANKS.asm"

TXLINES:
INCLUDE "Data/TXLINES.asm"

; Not currently in use.
I_ATTRIBUTES:
;INCLUDE "Data/ATTRS-I.asm"
ATTRIBUTES:
;INCBIN "Data/ATTRS.bin"

; Currently 11 byte gap between text and Node data.

; 6k (6144 bytes) available for Node data.
org $7820
D_NODES:
INCBIN "Data/NODES.bin"

; Characters 1 to 32 are not used, so Node data can run up to $9020.
org $8F20
D_TEXT_CHAR:

org $9020
; Characters 33 to 127 inclusive.
     DB  0,   0,   0,   0,   0,   0,   0,   0 ; Tile number 32.
     DB  0,   0,  16,  16,  16,   0,  16,   0 ; Tile number 33.
     DB  0,  36,  36,   0,   0,   0,   0,   0 ; Tile number 34.
     DB  0,  36, 126,  36,  36, 126,  36,   0 ; Tile number 35.
     DB  0,  28,  40,  24,  12,  12,  56,   0 ; Tile number 36.
     DB  0, 114,  84, 120,  30,  42,  78,   0 ; Tile number 37.
     DB  0,  16,  40,  16,  52,  72,  52,   0 ; Tile number 38.
     DB  0,  16,  16,   0,   0,   0,   0,   0 ; Tile number 39.
     DB  0,  16,  32,  32,  32,  32,  16,   0 ; Tile number 40.
     DB  0,   8,   4,   4,   4,   4,   8,   0 ; Tile number 41.
     DB  0,   0,   8,  28,   8,  20,   0,   0 ; Tile number 42.
     DB  0,  16,  16, 124,  16,  16,   0,   0 ; Tile number 43.
     DB  0,   0,   0,   0,   4,   4,   8,   0 ; Tile number 44.
     DB  0,   0,   0,   0,  60,   0,   0,   0 ; Tile number 45.
     DB  0,   0,   0,   0,   0,   0,   8,   0 ; Tile number 46.
     DB  0,   0,   4,  12,  24,  48,  32,   0 ; Tile number 47.
     DB  0,  56,  76,  84,  84, 100,  56,   0 ; Tile number 48.
     DB  0,  16,  48,  16,  16,  16,  56,   0 ; Tile number 49.
     DB  0,  56,   4,   4,  24,  32,  60,   0 ; Tile number 50.
     DB  0,  56,   4,   4,  24,   4,  56,   0 ; Tile number 51.
     DB  0,   8,  24,  40, 126,   8,   8,   0 ; Tile number 52.
     DB  0,  60,  32,  56,   4,   4,  56,   0 ; Tile number 53.
     DB  0,  56,  64, 120,  68,  68,  56,   0 ; Tile number 54.
     DB  0, 124,   4,  12,  24,  16,  16,   0 ; Tile number 55.
     DB  0,  60,  66,  60,  66,  66,  60,   0 ; Tile number 56.
     DB  0,  60,  66,  62,   2,   2,  60,   0 ; Tile number 57.
     DB  0,   0,  16,   0,   0,  16,   0,   0 ; Tile number 58.
     DB  0,   0,  16,   0,  16,  16,  32,   0 ; Tile number 59.
     DB  0,   0,  12,  48,  64,  48,  12,   0 ; Tile number 60.
     DB  0,   0,   0,  60,   0,  60,   0,   0 ; Tile number 61.
     DB  0,   0,  48,  12,   2,  12,  48,   0 ; Tile number 62.
     DB  0,  56,   4,   4,  24,   0,  16,   0 ; Tile number 63.
     DB  0,  62,  72,  72, 126,  72,  72,   0 ; Tile number 64, now AF logo.
     DB  0,  60,  66,  66, 126,  66,  66,   0 ; Tile number 65.
     DB  0, 124,  66, 124,  66,  66, 124,   0 ; Tile number 66.
     DB  0,  60,  64,  64,  64,  64,  60,   0 ; Tile number 67.
     DB  0, 124,  66,  66,  66,  66, 124,   0 ; Tile number 68.
     DB  0, 126,  64, 124,  64,  64, 126,   0 ; Tile number 69.
     DB  0, 126,  64, 124,  64,  64,  64,   0 ; Tile number 70.
     DB  0,  62,  64,  78,  66,  66,  62,   0 ; Tile number 71.
     DB  0,  66,  66, 126,  66,  66,  66,   0 ; Tile number 72.
     DB  0, 124,  16,  16,  16,  16, 124,   0 ; Tile number 73.
     DB  0, 126,   8,   8,   8,  72,  48,   0 ; Tile number 74.
     DB  0,  68,  72, 112,  80,  72,  68,   0 ; Tile number 75.
     DB  0,  64,  64,  64,  64,  64, 126,   0 ; Tile number 76.
     DB  0,  66, 102,  90,  66,  66,  66,   0 ; Tile number 77.
     DB  0,  66,  98,  82,  74,  70,  66,   0 ; Tile number 78.
     DB  0,  60,  66,  66,  66,  66,  60,   0 ; Tile number 79.
     DB  0, 124,  66,  66, 124,  64,  64,   0 ; Tile number 80.
     DB  0,  60,  66,  66,  74,  68,  58,   0 ; Tile number 81.
     DB  0, 124,  66,  66, 124,  66,  66,   0 ; Tile number 82.
     DB  0,  62,  64,  60,   2,   2, 124,   0 ; Tile number 83.
     DB  0, 126,  16,  16,  16,  16,  16,   0 ; Tile number 84.
     DB  0,  66,  66,  66,  66,  66,  60,   0 ; Tile number 85.
     DB  0,  66,  66,  36,  36,  24,  24,   0 ; Tile number 86.
     DB  0,  68,  68,  84, 124,  40,  40,   0 ; Tile number 87.
     DB  0,  66,  36,  24,  24,  36,  66,   0 ; Tile number 88.
     DB  0,  68,  68,  40,  16,  16,  16,   0 ; Tile number 89.
     DB  0, 126,   2,  12,  48,  64, 126,   0 ; Tile number 90.
     DB  0,  48,  32,  32,  32,  32,  48,   0 ; Tile number 91.
     DB  0,   0,  32,  48,  24,  12,   4,   0 ; Tile number 92.
     DB  0,  12,   4,   4,   4,   4,  12,   0 ; Tile number 93.
     DB  0,   0,  16,  40,  68,   0,   0,   0 ; Tile number 94.
     DB  0,   0,   0,   0,   0,   0, 126,   0 ; Tile number 95.
     DB  0,  32,  16,   8,   0,   0,   0,   0 ; Tile number 96.
     DB  0,   0,  24,   4,  28,  36,  28,   0 ; Tile number 97.
     DB  0,  32,  32,  56,  36,  36,  56,   0 ; Tile number 98.
     DB  0,   0,  28,  32,  32,  32,  28,   0 ; Tile number 99.
     DB  0,   4,   4,  28,  36,  36,  28,   0 ; Tile number 100.
     DB  0,   0,  24,  36,  56,  32,  28,   0 ; Tile number 101.
     DB  0,  12,  16,  56,  16,  16,  16,   0 ; Tile number 102.
     DB  0,   0,  28,  36,  36,  28,   4,  24 ; Tile number 103.
     DB  0,  32,  32,  56,  36,  36,  36,   0 ; Tile number 104.
     DB  0,  16,   0,  16,  16,  16,  16,   0 ; Tile number 105.
     DB  0,   4,   0,   4,   4,   4,  36,  24 ; Tile number 106.
     DB  0,  32,  32,  40,  48,  40,  36,   0 ; Tile number 107.
     DB  0,  32,  32,  32,  32,  32,  24,   0 ; Tile number 108.
     DB  0,   0,  52,  42,  42,  42,  42,   0 ; Tile number 109.
     DB  0,   0,  60,  34,  34,  34,  34,   0 ; Tile number 110.
     DB  0,   0,  28,  34,  34,  34,  28,   0 ; Tile number 111.
     DB  0,   0,  56,  36,  36,  56,  32,  32 ; Tile number 112.
     DB  0,   0,  28,  36,  36,  28,   6,   4 ; Tile number 113.
     DB  0,   0,  44,  50,  32,  32,  32,   0 ; Tile number 114.
     DB  0,   0,  28,  32,  24,   4,  56,   0 ; Tile number 115.
     DB  0,  16,  56,  16,  16,  16,  12,   0 ; Tile number 116.
     DB  0,   0,  34,  34,  34,  34,  30,   0 ; Tile number 117.
     DB  0,   0,  34,  34,  20,  20,   8,   0 ; Tile number 118.
     DB  0,   0,  34,  34,  42,  42,  20,   0 ; Tile number 119.
     DB  0,   0,  34,  20,   8,  20,  34,   0 ; Tile number 120.
     DB  0,   0,  36,  36,  36,  28,   4,  24 ; Tile number 121.
     DB  0,   0,  62,   4,   8,  16,  62,   0 ; Tile number 122.
     DB  0,  16,  32,  64,  64,  32,  16,   0 ; Tile number 123.
     DB  0,   8,   8,   8,   8,   8,   8,   0 ; Tile number 124.
     DB  0,   8,   4,   2,   2,   4,   8,   0 ; Tile number 125.
     DB  0,   0,   0,  48,  90,  12,   0,   0 ; Tile number 126.
     DB  0,   0,   0,   0,   0,   0,   0,   0 ; Tile number 127.

org $9320
; Not used by this game, so empty.
D_GFX_CHAR:

; 6.5k-ish for code.
org $9320:
INCLUDE "ZX-SCHEDULER.asm"
INCLUDE "ZX-PARSER.asm"
INCLUDE "ZX-UTILITIES.asm"
INCLUDE "ZX-DRAWING.asm"
INCLUDE "ZX-PRINTING.asm"
INCLUDE "ZX-SOUND.asm"
INCLUDE "SES-SPECIFIC.asm"

; Global constants.
C_SCREEN_BITMAP_START:
DB 00H, 40H

C_BUFFER_BITMAP_START:
DB 00H, 40H                   ; Same as screen bitmap (no buffer used.)

C_TILEBANK_START:
DW D_TEXT_CHAR

C_TILEBANK_SIZE:
DW $0400                     ; 128 tiles of 8 bytes each.
                             ; N.B. gfx tiles must immediately follow text.

; Random Number Generator (one byte.)
RNG:
PUSH AF
LD A, (VT_TIMER_FRAMES)
LD (V_RNG_NUMBER), A
POP AF
RET

V_RNG_KEYHOLDCOUNT:
DB 0

V_RNG_NUMBER:
DB 0

; Fast and compact.
; Called with attribute value in the A register.
SET_ALLATTRS:
LD HL, (V_ATTR_START)   ; Start of attributes map.
LD (HL), A              ; Set first attribute byte.
PUSH HL
POP DE
INC DE                  ; DE now points to start+1.
LD BC, $2FF             ; Remaining length of attributes map.
LDIR
RET

;(VT already converted to actual values, where applicable, when this is called.)
SET_ATTRIBUTES:
CALL SETUP_ATTRCHANGE
LD A, (IX+4)            ; ATTRVALUE.

  SET_ATTRIBUTES_JP1:
  LD (HL), A
  INC HL
  DEC B                 ; End of row?
  JR NZ, SET_ATTRIBUTES_JP1
  DEC C                 ; Last row?
  RET Z                 ; Done.

  LD B, D               ; Reset row position counter.
  PUSH DE               ; Preserve registers.
  LD D, 0               ; DE now contains 32-XLENGTH.
  ADD HL, DE            ; HL now points to start of the next row.
  POP DE                ; Restore registers.
  JR SET_ATTRIBUTES_JP1 ; Do next row.

; Not used, removed to save memory.
;(VT already converted to actual values, where applicable, when this is called.)
;INVERT_ATTRIBUTES:
;CALL SETUP_ATTRCHANGE
;
;  INVERT_ATTRIBUTES_JP1:
;  PUSH BC               ; Preserve registers.
;  LD A, (HL)
;  AND 11000000B         ; Clear all except upper two bits.
;  LD B, A               ; Store upper two bits in B.
;  LD A, (HL)
;  AND 00000111B         ; Clear all except Pen bits.
;  SLA A                 ; Shift three places left to Paper position.
;  SLA A
;  SLA A
;  LD C, A               ; Store new Paper bits in C.
;  LD A, (HL)
;  AND 00111000B         ; Clear all except Paper bits.
;  SRA A                 ; Shift three places right to Pen position.
;  SRA A
;  SRA A
;  OR B                  ; Add upper two bits (unchanged.)
;  OR C                  ; Add new Paper bits.
;  POP BC                ; Restore registers.
;  LD (HL), A            ; Store modified attribute.
;
;  INC HL
;  DEC B                 ; End of row?
;  JR NZ, INVERT_ATTRIBUTES_JP1
;  DEC C                 ; Last row?
;  RET Z                 ; Done.
;
;  LD B, D               ; Reset row position counter.
;  PUSH DE               ; Preserve registers.
;  LD D, 0               ; DE now contains 32-XLENGTH.
;  ADD HL, DE            ; HL now points to start of the next row.
;  POP DE                ; Restore registers.
;  JR INVERT_ATTRIBUTES_JP1; Do next row.

; Returns with:
; - memory location of attribute byte (XSTART, YSTART) in HL.
; - XLENGTH in B, and YLENGTH in C.
; - XLENGTH (copy) in D, and 32-XLENGTH in E.
; - 32 in DE (row pitch.)
SETUP_ATTRCHANGE:
LD IX, VT_ATTRIBUTES
LD HL, (V_ATTR_START)
LD L, (IX+0)            ; HL now offset by XSTART (no risk of overflow.)
LD D, 0
LD E, (IX+1)            ; DE now contains YSTART.
LD A, 32                ; There are 32 bytes per row.
CALL MULT_DE_BY_A_QND   ; No risk of overflow (if YSTART in valid range 0-31.)
ADD HL, DE              ; HL now also offset by (YSTART * 32)
LD B, (IX+2)            ; XLENGTH.
LD D, B
LD C, (IX+3)            ; YLENGTH.
LD A, 32
SUB B
LD E, A
RET

; Called from Parser.
CHECKGVAR_ATTRIBUTES:
LD HL, VT_ATTRIBUTES
LD C, 4                 ; Only want to check first four values (Xs and Ys.)
CALL GET_GVAR_BLOCK
RET

VT_ATTRIBUTES:
DB 0                    ;+0 XSTART
DB 0                    ;+1 YSTART
DB 32                   ;+2 XLENGTH
DB 24                   ;+3 YLENGTH
DB 01000000B            ;+4 ATTRVALUE (default: black pen, black paper.)
DB 0                    ;+5 FADEATTR


; Not used, removed to save memory.
; Loads attributes from file number in A register.
;LOAD_ATTRIBUTES:
;LD HL, I_ATTRIBUTES
;CALL GET_ADDRESS_UNALIGNED ; HL now contains start address of source file.
;LD DE, (V_ATTR_START)
;LD B, D                 ; Store copy of upper byte+3 for length check later.
;INC B
;INC B
;INC B

;  LOAD_ATTRIBUTES_JP1:
;  LD A, (HL)            ; Get next byte.
;  BIT 7, A              ; Run length byte?
;  JR NZ, LDAT_RLESUB
;  LD (DE), A            ; Load data byte to screen.

;  LOAD_ATTRIBUTES_JP2:
;  INC HL                ; Source pointer.
;  INC DE                ; Destination pointer.
;  LD A, D
;  CP B                  ; D will reach B at end of attributes ($0300 long.)
;  RET Z
;  JR LOAD_ATTRIBUTES_JP1

;LDAT_RLESUB:
;RES 7, A                ; Clear RLE flag.
;LD C, A                 ; Set run length counter.
;DEC HL                  ; Get previous attribute byte.
;LD A, (HL)
;INC HL

;  LDAT_RLESUB_JP1:
;  LD (DE), A            ; Write previous byte to screen.
;  DEC C                 ; End of run?
;  JR Z, LOAD_ATTRIBUTES_JP2
;  INC DE                ; Go to next screen byte.
;  JR LDAT_RLESUB_JP1    ; Fill with previous value.


; Not used, removed to save memory.
; Copies rectangular region from attributes buffer to video memory.
;COPYATTR_AREA:
;LD IX, VT_ATTRIBUTES
;LD HL, $5800                 ; Start of attributes in video memory.
;LD D, (IX+0)                 ; XSTART.
;LD E, (IX+1)                 ; YSTART.
;LD C, 32                     ; Width of screen and buffer attributes files.
;CALL GET_ARRAY_ADDRESS       ; HL now points to first video memory destination.
;PUSH HL                      ; Preserve video memory pointer.
;LD HL, ATTR_BUFFER
;CALL GET_ARRAY_ADDRESS       ; HL now points to first buffer memory source.
;POP DE                       ; Restore video memory pointer.
;XOR A                        ; Clear A register.

;  COPYATTR_AREA_JP1:         ; Per row.
;  LD C, (IX+2)               ; XLENGTH.
;  PUSH DE                    ; Preserve source & destination pointers.
;  PUSH HL                    ; Preserve destination.

;  COPYATTR_AREA_JP2:
;  LD A, (HL)                 ; Copy byte.
;  BIT 0, (IX+5)              ; Check FADEATTR.
;  CALL NZ, COPYATTR_FADE
;  LD (DE), A
;  INC HL                     ; Move along row.
;  INC DE
;  DEC C                      ; End of row?
;  JR NZ, COPYATTR_AREA_JP2
;  POP HL                     ; Restore source & destination pointers.
;  POP DE
;  LD BC, 32                  ; Width of screen and buffer attributes files.
;  EX DE, HL                  ; Offset both pointers by BC (move down one row.)
;  ADD HL, BC
;  EX DE, HL
;  ADD HL, BC
;  DEC (IX+3)                 ; YLENGTH (end of rows?)
;  JR NZ, COPYATTR_AREA_JP1
;  RET

;COPYATTR_FADE:
;EX DE, HL
;BIT 6, (HL)                  ; Check attribute in video memory.
;JR NZ, COPYATTR_FADE_JP1     ; Do not fade if BRIGHT flag already set.
;RES 6, A

;  COPYATTR_FADE_JP1:
;  EX DE, HL
;  RET

; Enable copy to buffer (default value is direct to video memory.)
V_ATTR_START:
DW $5800

; This fills pixel video memory (excluding attributes) with the A register.
FILL_SCREEN:
LD HL, 16384            ; Start of pixel video memory.
LD DE, 6144             ; Length of pixel video memory.
LD B, A                 ; Pixel value.

  FILL_SCREEN_JP1:
  LD (HL), B
  INC HL
  DEC DE
  LD A, E
  OR D
  JR NZ, FILL_SCREEN_JP1
  RET

FILL_AREA:
LD IX, VT_FILL_AREA
LD B, (IX+0)                 ; XSTART.
LD C, (IX+1)                 ; YSTART.

  ; Per row:
  FILL_AREA_JP1:
  CALL GET_SCR_CELLSTART     ; DE now contains row and column start address.
  PUSH BC                    ; Store X/Y position.
  LD B, (IX+2)               ; Use copy of XLENGTH as counter.
  LD A, (IX+4)               ; FILLVALUE.

  ; Per cell.
  FILL_AREA_JP2:
  LD C, 8                    ; Eight bytes per cell.
  PUSH DE                    ; Store initial screen address.

  FILL_AREA_JP3:
  LD (DE), A                 ; Copy byte to video memory.
  INC D                      ; Move down one pixel row in cell.
  DEC C                      ; All pixels rows in cell filled?
  JR NZ, FILL_AREA_JP3
  POP DE                     ; Restore initial screen address.
  INC DE                     ; Move to next cell along row.
  DEC B                      ; End of row length to be filled?
  JR NZ, FILL_AREA_JP2
  POP BC                     ; Restore X/Y position.
  INC C                      ; Move Y position down one row.
  DEC (IX+3)                 ; End of rows to be filled?
  JR NZ, FILL_AREA_JP1
  RET                        ; Done.

VT_FILL_AREA:
DB 0                         ;+0 XSTART
DB 0                         ;+1 YSTART
DB 0                         ;+2 XLENGTH
DB 0                         ;+3 YLENGTH
DB 0                         ;+4 FILLVALUE

; Called with minimum wait in BC, maximum wait in DE.
WAITFOR:
LD A, D                      ; Check if DE is zero.
OR E
RET Z
LD A, B                      ; Check if BC is zero.
OR C
JR Z, WAITFOR_JP2            ; If zero, can be broken immediately by keypress.

  WAITFOR_JP1:               ; Cannot break from this loop with a keypress.
  HALT
  DEC DE
  DEC BC
  LD A, B
  OR C
  JR NZ, WAITFOR_JP1

  LD A, D                    ; Covers situation where DE == BC.
  OR E                       ; (DE < BC is not covered.)
  RET Z

  WAITFOR_JP2:               ; Can break from this loop with a keypress.
  HALT
  LD A, (V_KEYBOARD_KEY_ANY) ; A=1 if any key was pressed this frame.
  DEC A
  JR Z, WAITFOR_JP3
  DEC DE
  LD A, D
  OR E
  JR NZ, WAITFOR_JP2
  LD BC, 0                  ; Clear BROKESEQ flag (in GVAR 0.)
  LD (V_GVAR), BC
  RET

  WAITFOR_JP3:
  LD BC, 1                  ; Set BROKESEQ flag (in GVAR 0.)
  LD (V_GVAR), BC
  CALL RESET_KEYDELAYS      ; Set all keys to delay value.
  RET

ONKEY:                      ; Called with key code in the A register.
LD HL, V_KEYBOARD_KEYS      ; Start of key status table.
LD B, 0
LD C, A
ADD HL, BC                  ; Offset address by key code low nibble.
LD A, (HL)                  ; A now contains the relevant key variable.
CP B                        ; Check if key was not pressed (A==0).
JR NZ, ONKEY_JP1            ; Otherwise, go on to check repeat delay.
LD A, 00000111B             ; Flags: B2=1(KEY), B1=1(SKIP), B0=1(INC DEPTH.)
CALL SET_DEPTH_SKIP
RET

  ONKEY_JP1:
  CP 1                      ; Check if key variable is 1.
  JR NZ, ONKEY_JP2          ; Otherwise, key variable is greater than 1.
  LD A, (ONKEY_DELAY)       ; Set key variable to delay value.
  LD (HL), A                ; This prevents retrigger until delay counts down.
  LD A, 00000101B           ; Flags: B2=1(KEY), B1=0(DON'T SKIP), B0=1(INC DEPTH.)
  CALL SET_DEPTH_SKIP
  RET

  ONKEY_JP2:                ; Key variable was greater than 1.
  LD A, 00000111B           ; Flags: B2=1(KEY), B1=1(SKIP), B0=1(INC DEPTH.)
  CALL SET_DEPTH_SKIP
  RET

ONKEY_DELAY:
DB 15                       ; Delay in frames before a held key repeats.

WAITFOR_ANYKEY:
PUSH AF                     ; Preserve registers.
PUSH BC
PUSH HL

  WAITFOR_ANYKEY_JP1:
  HALT                      ; Wait until next frame.
  LD A, (V_KEYBOARD_KEY_ANY); A=1 if any key was pressed this frame.
  DEC A
  JR Z, WAITFOR_ANYKEY_JP2
  JR WAITFOR_ANYKEY_JP1

  WAITFOR_ANYKEY_JP2:
  CALL RESET_KEYDELAYS      ; Set all keys to delay value.
  POP HL                    ; Restore registers.
  POP BC
  POP AF
  RET

; Resets all key timers to delay value.
RESET_KEYDELAYS:
LD A, (ONKEY_DELAY)
LD HL, V_KEYBOARD_KEYS
LD C, 5                   ; There are five key values to set

  RESET_KEYDELAYS_JP1:
  LD (HL), A
  INC HL
  DEC C
  JR NZ, RESET_KEYDELAYS_JP1
  RET

TILE_DRAW:                   ; Called with A=tile number, B=X position, C=Y position.
LD H, 0
LD L, A                      ; HL now contains tile value.
ADD HL, HL
ADD HL, HL
ADD HL, HL                   ; HL now multiplied by 8.
LD DE, (V_TILE_DRAW_BANK_START)
ADD HL, DE                   ; HL now contains tilebank source of first byte.
CALL GET_SCR_CELLSTART       ; DE now contains buffer destination of first byte.
LD C, 8                      ; There are 8 bytes per tile.

  TILE_DRAW_JP1:
  LD A, (HL)
  LD (DE), A
  INC HL                     ; Tile bank bytes are in direct order.
  INC D                      ; 256 byte step per line of character in video memory.
  DEC C
  JR NZ, TILE_DRAW_JP1
  RET

V_TILE_DRAW_BANK_START:      ; Used to store the address of tile 0 in the bank to be used.
DW 0


; Not used, removed to save memory.
; Called with GPOINTER in DE.
; (Needs some tuning because grain does not look random enough.)
;ADD_MEMGRAIN:
;LD IX, VT_ADD_MEMGRAIN
;EX DE, HL                    ; Move GPOINTER to HL.
;LD B, (IX+0)                 ; XSTART.
;LD C, (IX+1)                 ; YSTART.
;
;  ; Per row:
;  ADD_MEMGRAIN_JP1:
;  CALL GET_SCR_CELLSTART     ; DE now contains row and column start address.
;  EX DE, HL                  ; Now screen pointer in HL and GPOINTER in DE.
;  PUSH BC                    ; Store X/Y position.
;  LD B, (IX+2)               ; Use copy of XLENGTH as counter.
;
;  ; Per cell.
;  ADD_MEMGRAIN_JP2:
;  LD C, 8                    ; Eight bytes per cell.
;  PUSH HL                    ; Store initial screen address.
;
;  ADD_MEMGRAIN_JP3:
;  LD A, (DE)                 ; Get grain byte.
;  AND (IX+4)                 ; Bitmask grain byte.
;  OR (HL)                    ; Add screen byte bits.
;  LD (HL), A                 ; Update screen byte.
;  INC H                      ; Move down one pixel row in cell.
;  INC DE                     ; Next grain byte.
;  RRC (IX+4)                 ; Rotate mask byte (to reduce pattern.)
;  DEC C                      ; All pixels rows in cell filled?
;  JR NZ, ADD_MEMGRAIN_JP3
;  POP HL                     ; Restore initial screen address.
;  INC HL                     ; Move to next cell along row.
;  DEC B                      ; End of row length to be filled?
;  JR NZ, ADD_MEMGRAIN_JP2
;  POP BC                     ; Restore X/Y position.
;  INC C                      ; Move Y position down one row.
;  DEC (IX+3)                 ; End of rows to be filled?
;  JR NZ, ADD_MEMGRAIN_JP1
;  RET                        ; Done.
;
;VT_ADD_MEMGRAIN:
;DB 0                         ;+0 XSTART.
;DB 0                         ;+1 YSTART.
;DB 0                         ;+2 XLENGTH.
;DB 0                         ;+3 YLENGTH.
;DB 0                         ;+4 BITMASK.

org $AC00
INTADDRESS_TABLE:
; The interrupt address lookup table (to cover all possible bus byte values.)
        DS $101, $FD

; 19.875k available for bitmap data.
org $AD01
D_BITMAPS:
INCBIN "Data/BITMAPS.bin"


; Maximum of 27 nodes (2x2 bytes per entry.)
org $FD90
I_NODE_ENTRY:
INCBIN "Data/NODES-I-ENT.bin"

I_NODE_CURRENT:
INCBIN "Data/NODES-I-CUR.bin"

org $FDFD
INTERRUPT:
JP SCHEDULER

; Maximum of 128 subnodes (2 byte entry per subnode.)
org $FE00
I_NODE_SUB:
INCBIN "Data/NODES-I-SUB.bin"

; Maximum of 127 bitmaps (and 128th null entry.)
org $FF00
I_BGSTREAMS:
INCLUDE "Data/BITMAPS-I.asm"
