; Could also be used by an uncompressed drawing routine.
GET_IMAGE_INFO:
LD (IX+0), B                   ; Store XCellPos as XSTART.
LD (IX+1), C                   ; Store YCellPos as YSTART.
CALL GET_ADDRESS_UNALIGNED     ; HL now at start of bitmap data.
LD D, (HL)                     ; Width (B0-B6) and Alpha (B7.)
INC HL
LD E, (HL)                     ; Height.
INC HL                         ; HL now at 3rd byte of bitmap data.
LD A, 10000000B                ; Alpha mask.
AND D                          ; Get Alpha only.
RLCA                           ; Move Alpha from B7 to B0.
LD (IX+2), A                   ; Store ALPHA.
LD A, 01111111B                ; Width mask.
AND D                          ; Get Width only.
LD (IX+3), A                   ; Store WIDTH.
LD (IX+4), E                   ; Store HEIGHT.
RET

; Called with BackgroundNum in A, XCellPos in B, and YCellPos in C.
DRAW_BGSTREAM:
LD (V_LASTDRAWPOS), BC         ; Store copy of XY position for future reference.
LD IX, VT_DRAW_BGSTREAM
LD HL, I_BGSTREAMS
CALL GET_IMAGE_INFO            ; Fills VT +0 to +4, HL points to 3rd byte.
LD E, (HL)                     ; Get 3rd & 4th byte of header (little endian.)
INC HL
LD D, (HL)                     ; DE now points to start of pixel data stream.
INC HL                         ; HL now points to start of cellmap.
CALL DBGS_SET_SHADOWREGS
EX DE, HL                      ; HL now points to start of pixel data stream.
XOR A
LD (IX+5), A                   ; Reset RLCOUNT.
LD (IX+6), A                   ; Reset ROWPASS.

  DRAW_BGSTREAM_JP1:
  CALL GET_SCR_CELLSTART       ; Returns pointer in DE.
  LD C, (IX+3)                 ; WIDTH.

  DRAW_BGSTREAM_JP2:
  PUSH BC                      ; Preserve column counter.
  PUSH DE                      ; Preserve screen pointer.
  LD A, (IX+5)                 ; RLCOUNT.
  CP 0
  JR NZ, DRAW_BGSTREAM_JP3
  EXX                          ; New handler address is required.
  CALL DBGS_NEXT_CELLCODE      ; IY now holds new handler address.
  EXX
  LD C, 8                      ; Byte counter for filling cells.
  JP (IY)                      ; Fill cell as instructed (by code & Alpha flag.)

  DRAW_BGSTREAM_JP3:           ; Mid-run.
  DEC (IX+5)                   ; RLCOUNT.
  LD C, 8                      ; Byte counter for filling cells.
  CALL DBGS_CELLIS_REPT

  DRAW_BGSTREAM_JP4:           ; Re-entry from all instruction handlers.
  POP DE                       ; Restore screen pointer.
  POP BC                       ; Restore column counter.
  INC DE                       ; DE now at start of next screen cell.
  DEC C                        ; Width counter.
  JR NZ, DRAW_BGSTREAM_JP2     ; Draw next cell.

  LD A, (IX+6)                 ; Get ROWPASS.
  XOR 00000001B                ; Flip it each time that a row is completed.
  LD (IX+6), A                 ; Store ROWPASS.
  AND (IX+2)                   ; Check if ALPHA and Rowpass are both set.
  JR Z, DRAW_BGSTREAM_JP5      ; If not, this row is complete.
  LD B, (IX+0)                 ; Get XSTART.
  LD C, (IX+1)                 ; Get YSTART/YPOS (not updated yet.)
  JR DRAW_BGSTREAM_JP1         ; Run another pass over the row.

  DRAW_BGSTREAM_JP5:           ; Prepare for next row down.
  DEC (IX+4)                   ; HEIGHT/counter.
  RET Z
  LD B, (IX+0)                 ; XSTART.
  INC (IX+1)                   ; YSTART/YPOS.
  LD C, (IX+1)
  JR DRAW_BGSTREAM_JP1         ; Draw next row.

; Fills shadow registers for parsing cell codes.
; Called with pointer to start of image (cell code data) in HL.
DBGS_SET_SHADOWREGS:
PUSH HL                        ; Store start of image pointer.
EXX                            ; REG'START
POP HL                         ; 'Get copy of start of image pointer.
LD B, (HL)                     ; 'Get first index byte.
LD C, 0                        ; 'Reset index byte shift tracker.
LD D, (IX+2)                   ; 'Get ALPHA.
EXX                            ; REG'END
RET

; Returns jump address to handle next cell code in IY.
DBGS_NEXT_CELLCODE:
LD A, B                        ; Current index byte.
AND 00000011B                  ; Mask upper six bits.
SLA A                          ; Move code to B2/B1.
OR D                           ; Copy Alpha flag to B0.
PUSH HL                        ; Preserve index byte pointer.
LD HL, WT_DRAW_BGSTREAM
CALL GET_JUMP_FROMTABLE        ; IY now contains the command address.
POP HL                         ; Restore index byte pointer.
SRL B                          ; Shift next index code bits into position.
SRL B
INC C                          ; Increment shift tracker.
BIT 2, C                       ; Check if shift tracker has reached 4.
RET Z
LD C, 0                        ; Reset shift tracker.
INC HL                         ; Get next cell code index byte.
LD B, (HL)
RET

DBGS_CELL_AND_DATA:
LD A, (DE)                     ; Get screen byte.
AND (HL)                       ; Mask screen byte.
LD (DE), A                     ; Return masked byte to screen.
INC D
INC HL                         ; Prepare for next byte.
DEC C                          ; Check if all bytes copied.
JR NZ, DBGS_CELL_AND_DATA
RET

DBGS_CELL_OR_DATA:
LD A, (DE)                     ; Get screen byte.
OR (HL)                        ; Blend data byte.
LD (DE), A                     ; Return blended byte to screen.
INC D
INC HL                         ; Prepare for next byte.
DEC C                          ; Check if all bytes copied.
JR NZ, DBGS_CELL_OR_DATA
RET

DBGS_CELL_SET_DATA:
LD A, (HL)
LD (DE), A                     ; Copy byte.
INC D
INC HL                         ; Prepare for next byte.
DEC C                          ; Check if all bytes copied.
JR NZ, DBGS_CELL_SET_DATA
RET

DBGS_CELL_SET_VALUE:           ; Inlined for speed.
LD (DE), A                     ; Set byte.
INC D
DEC C                          ; Check if all bytes set.
JR NZ, DBGS_CELL_SET_VALUE
RET

DBGS_CODETOPTR:
LD A, (HL)                     ; Get first byte of code from data stream.
CP 96                          ; Minimum high byte for two-byte pointer.
JR NC, DBGS_CODETOPTR_JP4
CP 64                          ; Minimum code for RLE.
JR NC, DBGS_CODETOPTR_JP3

  DBGS_CODETOPTR_JP1:          ; Index code (0-63).
  PUSH HL                      ; Preserve data stream pointer.
  PUSH DE                      ; Preserve DE.
  LD D, 0
  LD E, A
  SLA E                         ; Multiply by eight (cell size) to get offset.
  SLA E
  SLA E
  JR NC, DBGS_CODETOPTR_JP2    ; Check for overflow on final shift.
  INC D

  DBGS_CODETOPTR_JP2:          ; DE now contains the offset to the indexed cell.
  LD HL, D_BITMAPS             ; Indexed cells are stored at the start of this.
  ADD HL, DE                   ; Add offset to calculate pointer.
  POP DE                       ; Restore DE.
  JR DBGS_CODETOPTR_JP5

  DBGS_CODETOPTR_JP3:          ; RLE code (64-95).
  PUSH HL                      ; Preserve data stream pointer.
  AND 00011111B                ; Lower five bits are run length -1.
  LD (IX+5), A                 ; RLCOUNT (checked after drawing, so not +1'd.)
  LD HL, (V_LASTIMAGEPTR)      ; Temporary pointer will be the same as before.
  JR DBGS_CODETOPTR_JP5

  DBGS_CODETOPTR_JP4:          ; Full pointer high byte (96-255).
  INC HL                       ; Move to next byte in data stream.
  PUSH HL                      ; Preserve data stream pointer.
  LD L, (HL)                   ; Get low byte of two-byte pointer.
  LD H, A                      ; HL now contains the full pointer.

  DBGS_CODETOPTR_JP5:
  LD (V_LASTIMAGEPTR), HL      ; Store temporary pointer.
  POP HL                       ; Restore data stream pointer.
  INC HL                       ; Move to next entry in data stream.
  RET

DBGS_CELLIS_REPT:
PUSH HL                        ; Preserve data stream pointer.
LD HL, (V_LASTIMAGEPTR)
BIT 0, (IX+2)                  ; ALPHA.
JR NZ, DBGS_CELLIS_REPT_JP1
CALL DBGS_CELL_SET_DATA        ; Solid.
POP HL                         ; Restore data stream pointer.
RET

  DBGS_CELLIS_REPT_JP1:
  BIT 0, (IX+6)                ; ROWPASS.
  CALL Z, DBGS_CELL_AND_DATA   ; Alpha, first pass.
  BIT 0, (IX+6)                ; ROWPASS.
  CALL NZ, DBGS_CELL_OR_DATA   ; Alpha, second pass.
  POP HL                       ; Restore data stream pointer.
  RET

DBGS_CELLIS_000:
LD (V_LASTIMAGEPTR), HL        ; Pointer may be reused by subsequent RLE.
CALL DBGS_CELL_SET_DATA
JP DRAW_BGSTREAM_JP4

DBGS_CELLIS_001:
LD (V_LASTIMAGEPTR), HL        ; Pointer may be reused by subsequent RLE.
BIT 0, (IX+6)                  ; ROWPASS.
CALL Z, DBGS_CELL_AND_DATA
BIT 0, (IX+6)                  ; ROWPASS.
CALL NZ, DBGS_CELL_OR_DATA
JP DRAW_BGSTREAM_JP4

DBGS_CELLIS_010:
CALL DBGS_CODETOPTR
PUSH HL                        ; Preserve data stream pointer.
LD HL, (V_LASTIMAGEPTR)        ; Get pointer returned by DBGS_CODETOPTR.
CALL DBGS_CELL_SET_DATA
POP HL                         ; Restore data stream pointer.
JP DRAW_BGSTREAM_JP4

DBGS_CELLIS_011:
CALL DBGS_CODETOPTR
PUSH HL                        ; Preserve data stream pointer.
LD HL, (V_LASTIMAGEPTR)        ; Get pointer returned by DBGS_CODETOPTR.
BIT 0, (IX+6)                  ; ROWPASS.
CALL Z, DBGS_CELL_AND_DATA
BIT 0, (IX+6)                  ; ROWPASS.
CALL NZ, DBGS_CELL_OR_DATA
POP HL                         ; Restore data stream pointer.
JP DRAW_BGSTREAM_JP4

DBGS_CELLIS_100:
LD A, 00000000B
CALL DBGS_CELL_SET_VALUE
JP DRAW_BGSTREAM_JP4

DBGS_CELLIS_101:
LD A, 00000000B
BIT 0, (IX+6)                  ; ROWPASS.
CALL Z, DBGS_CELL_SET_VALUE    ; No action required if white on 2nd pass.
JP DRAW_BGSTREAM_JP4

DBGS_CELLIS_110:
LD A, 11111111B
CALL DBGS_CELL_SET_VALUE
JP DRAW_BGSTREAM_JP4

DBGS_CELLIS_111:
LD A, 11111111B
BIT 0, (IX+6)                  ; ROWPASS.
CALL NZ, DBGS_CELL_SET_VALUE   ; No action required if black on 1st pass.
JP DRAW_BGSTREAM_JP4

WT_DRAW_BGSTREAM:
DW DBGS_CELLIS_000             ; Code 000: Data, Solid.
DW DBGS_CELLIS_001             ; Code 001: Data, Alpha.
DW DBGS_CELLIS_010             ; Code 010: Code, Solid.
DW DBGS_CELLIS_011             ; Code 011: Code, Alpha
DW DBGS_CELLIS_100             ; Code 100: White, Solid.
DW DBGS_CELLIS_101             ; Code 101: White, Alpha.
DW DBGS_CELLIS_110             ; Code 110: Black, Solid.
DW DBGS_CELLIS_111             ; Code 111: Black, Alpha.

VT_DRAW_BGSTREAM:
DB 0                           ;+0 XSTART
DB 0                           ;+1 YSTART (initially, then YPOS tracker.)
DB 0                           ;+2 ALPHA
DB 0                           ;+3 WIDTH
DB 0                           ;+4 HEIGHT (initially, then counter.)
DB 0                           ;+5 RLCOUNT
DB 0                           ;+6 ROWPASS

; Stored as two-byte word.
V_LASTDRAWPOS:
DB 0                           ; YSTART.
DB 0                           ; XSTART.

V_LASTIMAGEPTR:
DW 0

; Generates FLIPX_LOOKUP table.
GEN_FLIPX_LOOKUP:
LD C, 0
LD HL, FLIPX_LOOKUP

  GEN_FLIPX_LOOKUP_JP1:
  LD B, 8                      ; Eight bits to be flipped.
  LD D, C                      ; Copy of byte to be flipped.

  GEN_FLIPX_LOOKUP_JP2:        ; Flip copy of byte
  RL D
  RRA
  DJNZ GEN_FLIPX_LOOKUP_JP2
  LD (HL), A                   ; Stored flipped byte.
  INC HL                       ; Move to next table position.
  INC C                        ; Set next byte to flip.
  RET Z                        ; All 256 values have been flipped.
  JR GEN_FLIPX_LOOKUP_JP1      ; Flip next byte.

SCRFLIP_X:
LD IX, VT_SCRFLIP
LD A, (IX+4)                   ; XTOSTART.
ADD A, (IX+2)                  ; XLENGTH.
DEC A
LD (IX+5), A                   ; Store calculated XTOEND.
EXX                            ; REG'START
LD HL, FLIPX_LOOKUP            ; 'Set 'HL to start of byte flip lookup table.
EXX                            ; REG'END

  SCRFLIP_X_JP1:               ; Start of cell row.
  LD B, (IX+0)                 ; XFROM.
  LD C, (IX+1)                 ; YFROM.
  CALL GET_SCR_CELLSTART
  EX DE, HL                    ; HL now contains left cell of row to be flipped.
  LD B, (IX+5)                 ; XTOEND
  CALL GET_SCR_CELLSTART       ; DE now contains right cell of row destination.
  LD C, 8                      ; Eight bytes per row.

  SCRFLIP_X_JP2:               ; Start of byte row within cell.
  LD B, (IX+2)                 ; XLENGTH.
  PUSH HL                      ; Preserve source & destination screen pointers.
  PUSH DE

  SCRFLIP_X_JP3:               ; Flip row of bytes.
  LD A, (HL)                   ; Get byte to flip.
  EXX                          ; REG'START
  LD L, A                      ; 'Table is 0x--00 aligned, no overflow risk.
  LD A, (HL)                   ; Get flipped byte from table.
  EXX                          ; REG'END
  LD (DE), A                   ; Store flipped byte.
  INC L                        ; Move right in source area.
  DEC E                        ; Move left in destination area.
  DJNZ SCRFLIP_X_JP3           ; Repeat for XLENGTH.
  POP DE                       ; Restore source & destination screen pointers.
  POP HL
  INC H                        ; Move down one row within cell.
  INC D
  DEC C                        ; Repeat for height of cell.
  JR NZ, SCRFLIP_X_JP2

  INC (IX+1)                   ; YFROM (move down to next cell.)
  DEC (IX+3)                   ; YLENGTH.
  JR NZ, SCRFLIP_X_JP1         ; Draw next cell row.
  RET

; Many simalarities to SCRFLIP_X, so shared variable table.
SCRFLIP_Y:
LD IX, VT_SCRFLIP
LD A, (IX+4)                   ; YTOSTART.
ADD A, (IX+3)                  ; YLENGTH.
DEC A
LD (IX+5), A                   ; Store calculated YTOEND.

  SCRFLIP_Y_JP1:               ; Start of cell row.
  LD B, (IX+0)                 ; XFROM.
  LD C, (IX+1)                 ; YFROM.
  CALL GET_SCR_CELLSTART
  EX DE, HL                    ; HL now contains left cell of row to be flipped.
  LD C, (IX+5)                 ; YTOEND
  CALL GET_SCR_CELLSTART       ; DE now contains left cell of destination row.
  LD A, D                      ; Move DE to bottom row of destination cell.
  ADD A, 7
  LD D, A
  LD A, 8                      ; Eight bytes per row.
  LD B, 0                      ; Clear B.

  SCRFLIP_Y_JP2:               ; Start of byte row within cell.
  LD C, (IX+2)                 ; XLENGTH.
  PUSH HL                      ; Preserve source & destination screen pointers.
  PUSH DE
  LDIR                         ; Copy row.
  POP DE                       ; Restore source & destination screen pointers.
  POP HL
  INC H                        ; Move down one row within source cell.
  DEC D                        ; Move up one row within destincation cell.
  DEC A                        ; Repeat for height of cell.
  JR NZ, SCRFLIP_Y_JP2

  INC (IX+1)                   ; YFROM (move down to next source cell.)
  DEC (IX+5)                   ; YTOEND (move up to next destination cell.)
  DEC (IX+3)                   ; YLENGTH.
  JR NZ, SCRFLIP_Y_JP1         ; Draw next cell row.
  RET

; Many simalarities to SCRFLIP_X, so shared variable table.
SCRCOPY:
LD IX, VT_SCRFLIP

  SCRCOPY_JP1:                 ; Start of cell row.
  LD B, (IX+0)                 ; XFROM.
  LD C, (IX+1)                 ; YFROM.
  CALL GET_SCR_CELLSTART
  EX DE, HL                    ; HL now contains left cell of row to be flipped.
  LD B, (IX+4)                 ; XTO.
  LD C, (IX+5)                 ; YTO.
  CALL GET_SCR_CELLSTART       ; DE now contains left cell of destination row.
  LD A, 8                      ; Eight bytes per row.
  LD B, 0                      ; Clear B.

  SCRCOPY_JP2:                 ; Start of byte row within cell.
  LD C, (IX+2)                 ; XLENGTH.
  PUSH HL                      ; Preserve source & destination screen pointers.
  PUSH DE
  LDIR                         ; Copy row.
  POP DE                       ; Restore source & destination screen pointers.
  POP HL
  INC H                        ; Move down one row within source cell.
  INC D                        ; Move down one row within destincation cell.
  DEC A                        ; Repeat for height of cell.
  JR NZ, SCRCOPY_JP2

  INC (IX+1)                   ; YFROM (move down to next source cell.)
  INC (IX+5)                   ; YTO (move down to next destination cell.)
  DEC (IX+3)                   ; YLENGTH.
  JR NZ, SCRCOPY_JP1         ; Draw next cell row.
  RET

; This order fits use of NPS_00_COPYXYDIMS in parser.
VT_SCRFLIP:
DB 0                    ;+0 XFROM.
DB 0                    ;+1 YFROM.
DB 0                    ;+2 XLENGTH.
DB 0                    ;+3 YLENGTH.
DB 0                    ;+4 XTOSTART / YTOSTART / XTO.
DB 0                    ;+5 XTOEND / YTOEND / YTO.

; Draw horizontal line.
HLINE:
LD IX, VT_HLINE
LD A, (IX+1)            ; YSTART.
CALL GET_YPIXPTR        ; HL now contains start address of pixel row.

LD C, (IX+2)            ; XLENGTH.
LD B, (IX+0)            ; XSTART.
LD E, (IX+3)            ; COLOUR.
LD A, B                 ; Make working copy of XSTART.
SRL A                   ; Upper five bits are byte offset within row.
SRL A
SRL A
ADD A, L                ; No overflow risk.
LD L, A                 ; HL now points to start byte within row.

LD A, 00000111B         ; Get lower three bits of XSTART.
AND B                   ; A now contains the start pixel offset within byte.
LD D, A                 ; Store copy of pixel offset.
JR Z, HLINE_JP5         ; Skip starting part-byte if offset is zero.

LD A, 8                 ; Calculate pixel length within byte.
SUB D
LD B, A                 ; Store pixel length counter.
XOR A                   ; Prepare to fill first byte (00000000).

  HLINE_JP1:
  SCF
  RRA                   ; Net result is shift right, then set B7.
  DEC C                 ; Line width counter.
  JR NZ, HLINE_JP3

  HLINE_JP2:            ; Line fits entirely within first cell.
  SRL A                 ; Shift byte by offset, filling from left with 0.
  DEC D
  JR NZ, HLINE_JP2
  JR HLINE_JP9         ; Shortcut to write final byte.

  HLINE_JP3:
  DJNZ HLINE_JP1        ; Pixel length counter.

  HLINE_JP4:
  SRL A                 ; Shift byte by offset, filling from left with 0.
  DEC D
  JR NZ, HLINE_JP4
  CALL HLINE_BYCOL      ; Mask line by colour, update screen byte, move pointer.

  HLINE_JP5:            ; Starting part-byte now written (or not required.)
  LD A, 11111000B       ; Get upper five bits of remaining line length.
  AND C                 ; If zero, then there are no whole bytes to fill
  JR Z, HLINE_JP7
  LD B, A               ; Otherwise, store value contained in upper five bits.
  SRL B
  SRL B
  SRL B
  CPL                   ; Clear upper five bits of remaining line length.
  AND C
  LD C, A
  LD A, (IX+3)          ; Set colour for whole bytes.

  HLINE_JP6:            ; Fill whole bytes in line.
  LD (HL), A            ; Write whole byte to display.
  INC HL                ; Move to next byte in line.
  DJNZ HLINE_JP6        ; Upper five bits of remaining line length.

  HLINE_JP7:            ; Calculate final byte (except for single-cell lines.)
  XOR A                 ; Prepare to fill final byte (00000000).
  CP C                  ; Not required if line width counter already at zero.
  RET Z

  HLINE_JP8:
  SCF
  RRA                   ; Net result is shift right, then set B7.
  DEC C                 ; Line width counter.
  JR NZ, HLINE_JP8

  HLINE_JP9:           ; Final byte in line.
  CALL HLINE_BYCOL      ; Mask line by colour, update screen byte, move pointer.
  RET                   ; Done.

HLINE_BYCOL:
BIT 0, E                ; Check colour.
JR Z, HLINE_BYCOL_JP1

OR (HL)                 ; Black line, mask against screen byte.
JR HLINE_BYCOL_JP2      ; Done.

  HLINE_BYCOL_JP1:      ; White line.
  CPL                   ; Invert.
  AND (HL)              ; Mask against screen byte.

  HLINE_BYCOL_JP2:
  LD (HL), A            ; Write byte to display.
  INC HL                ; Move to next byte in line.
  RET                   ; Done.

VT_HLINE:
DB 0                    ;+0 XSTART.
DB 0                    ;+1 YSTART.
DB 0                    ;+2 XLENGTH.
DB 00000000B            ;+3 COLOUR. - N.B. must be 00000000 or 11111111.
