NODE_PARSER:
LD A, (V_NODE_CURRENT)
LD HL, V_NODE_NEXT
CP (HL)
JR Z, NODE_PARSER_JP1

LD A, (HL)                      ; V_NODE_NEXT.
LD (V_NODE_CURRENT), A

  NODE_PARSER_JP1:
  LD A, (V_NODE_STATE)
  LD B, A
  XOR A                         ; Set A to 0.
  CP B                          ; Check for node state 0 (entry.)
  JR NZ, NODE_PARSER_JP2
  LD HL, I_NODE_ENTRY
  JR NODE_PARSER_JP3

  NODE_PARSER_JP2:
  INC A
  CP B                          ; Check for node state 1 (current.)
  RET NZ                        ; Return if none of the above.
  LD HL, I_NODE_CURRENT

  NODE_PARSER_JP3:
  LD A, (V_NODE_CURRENT)        ; This is also the low address offset byte / 2.
  SLA A                         ; A now contains the offset (no overflow risk.)
  LD E, A
  LD D, 0                       ; DE now contains the offset.
  ADD HL, DE                    ; HL now points to the correct index entry.
  LD E, (HL)
  INC HL
  LD D, (HL)
  EX DE, HL                     ; HL now points to the start of code.
  DEC HL                        ; Compensates for INC HL in NODE_PARSER_MAIN.

  NODE_PARSER_MAIN:
  INC HL
  LD B, 0
  LD C, (HL)                    ; Get the next instruction byte (in BC.)
  SLA C                         ; Multiply by two.
  JR NC, NODE_PARSER_MAIN_JP1
  INC B

    NODE_PARSER_MAIN_JP1:
    LD IY, I_NODE_PARSER_INSTRUCTIONS
    ADD IY, BC                  ; IY points at the instruction handler address.
    LD E, (IY+0)
    LD D, (IY+1)                ; Now DE contains the handler address.
    PUSH DE
    POP IY                      ; Now IY contains the handler address.

    LD A, (V_NODE_IFSKIP)       ; Check for skipping all but END instructions.
    LD B, A
    LD A, (V_NODE_KEYSKIP)
    OR B                        ; If IFSKIP and/or KEYSKIP was not zero...
    JR NZ, NODE_PARSER_MAIN_JP2 ; ...then possibly skip the next instruction.

    PUSH HL                     ; Preserve bytecode address pointer.
    LD E, (IY-1)                ; Get bytecode instruction length.
    LD D, 0
    DEC DE
    ADD HL, DE
    LD (V_NODE_POINTER), HL     ; Store end address bytecode instruction.
    POP HL                      ; Restore bytecode address pointer.
    JP (IY)                     ; Go and interpret the instruction.

    NODE_PARSER_MAIN_JP2:
    LD A, (HL)                  ; Get the instruction byte.
    AND 11110000B               ; Mask lower nibble.

    CP $C0                      ; Check for match with END instruction range.
    JR NZ, NODE_PARSER_MAIN_JP3 ; If no match, check for IF range.
    JP (IY)                     ; Otherwise, interpret the instruction.

    NODE_PARSER_MAIN_JP3:
    CP $80                      ; Check for match with IF instruction range.
    JR NZ, NODE_PARSER_MAIN_JP4 ; If no match, check for ONKEY range.

    PUSH HL                     ; If match, increment IFDEPTH.
    LD A, (V_NODE_IFSKIP)       ; Set next IFSKIPSTACK bit to current status.
    SLA A                       ; Move SKIP flag to BIT 1.
    OR 00000001B                ; Also set BIT 2=0 (IF) and BIT 0=1 (IncDepth.)
    CALL SET_DEPTH_SKIP         ; A register flags now set, call routine.
    POP HL
    JR NODE_PARSER_MAIN_JP5     ; Skip the actual instruction.

    NODE_PARSER_MAIN_JP4:
    CP $A0                      ; Check for match with ONKEY instruction range.
    JR NZ, NODE_PARSER_MAIN_JP5 ; If no match, simply skip the instruction.

    PUSH HL                     ; If match, increment KEYDEPTH.
    LD A, (V_NODE_KEYSKIP)      ; Set next KEYSKIPSTACK bit to current status.
    SLA A                       ; Move SKIP flag to BIT 1.
    OR 00000101B                ; Also set BIT 2=1 (KEY) and BIT 0=1 (IncDepth.)
    CALL SET_DEPTH_SKIP         ; A register flags now set, call routine.
    POP HL                      ; Now proceed to skip the actual instruction.

    NODE_PARSER_MAIN_JP5:
    LD A, (IY-1)                ; Each instruction's length precedes its label.

    NODE_PARSER_MAIN_JP6:       ; Increment HL by instruction length-1.
    INC HL
    DEC A
    JR NZ, NODE_PARSER_MAIN_JP6
    DEC HL
    JR NODE_PARSER_MAIN         ; Interpret the next instruction.

    NODE_PARSER_MAIN_0X:        ; 0X instructions.

      DB 5      ; Instruction length.
      NODE_PARSER_MAIN_00_PRINTLINE:
      LD IX, VT_PS_DRAWROW
      CALL GET_NEXT_NPBYTE      ; LineNum.
      PUSH AF                   ; Store LineNum.
      CALL GET_NEXT_NPBYTE      ; XCellPos.
      LD (IX+0), A              ; XPOS.
      CALL GET_NEXT_NPBYTE      ; YCellPos.
      LD (IX+1), A              ; YPOS.
      CALL GET_NEXT_NPBYTE      ; XCellClearLength.
      LD (IX+2), A              ; XLENGTH.
      POP AF                    ; Restore LineNum.
      CALL PRINT_STRING_ASLINE
      JP RETURN_TO_PARSER       ; Done.

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_01_PRINTPGPH:
      CALL GET_NEXT_NPBYTE      ; LineNum.
      CALL PRINT_STRING_ASPGPH
      JP RETURN_TO_PARSER       ; Done.

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_02_CLEARPGPH:
      CALL PS_CLEAR_PGPH
      JP RETURN_TO_PARSER       ; Done.

      DB 5      ; Instruction length.
      NODE_PARSER_MAIN_03_SETPGPHAREA:
      LD IX, VT_PS_ASPGPH
      CALL NPS_00_COPYXYDIMS
      LD A, (IX+1)              ; YSTART.
      DEC A
      LD (IX+7), A              ; Reset YTRACK to YSTART-1.
      JP NODE_PARSER_MAIN       ; Done.

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_04_SETPGPHSPACE:
      LD IX, VT_PS_ASPGPH
      CALL GET_NEXT_NPBYTE      ; Spacing.
      LD (IX+5), A              ; ROWGAP.
      JP NODE_PARSER_MAIN       ; Done.

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_05_SETPGPHMODE:
      LD IX, VT_PS_ASPGPH
      CALL GET_NEXT_NPBYTE      ; LineMode.
      LD (IX+4), A              ; LFMODE.
      JP NODE_PARSER_MAIN       ; Done.

      DB 3      ; Instruction length.
      NODE_PARSER_MAIN_06_SETPGPHDELAY:
      CALL GET_NEXT_NPB_FULLVAR ; MinDelay.
      LD (V_PSAP_MINWAIT), DE
      CALL GET_NEXT_NPB_FULLVAR ; MaxDelay.
      LD (V_PSAP_MAXWAIT), DE
      JP NODE_PARSER_MAIN       ; Done.

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_07_SETPGPHPROMPT:
      LD IX, VT_PS_ASPGPH
      CALL GET_NEXT_NPBYTE      ; DictWord.
      LD (IX+6), A              ; DPDICT.
      JP NODE_PARSER_MAIN       ; Done.

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_08_SETLINEBANKHI:
      LD A, (V_SLBANK_OFFSET)   ; Copy destination index offset.
      LD E, A
      CALL GET_NEXT_NPBYTE      ; BankNum.
      CALL SET_LINEBANK
      JP RETURN_TO_PARSER       ; Done.

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_09_SETLINEBANKLO:
      CALL GET_NEXT_NPBYTE      ; BankNum.
      LD E, 0                   ; Copy destination index offset.
      CALL SET_LINEBANK
      JP RETURN_TO_PARSER       ; Done.

      DB 3      ; Instruction length.
      NODE_PARSER_MAIN_0A_SETPGPHSOUND:
      LD IX, VT_PS_ASPGPH
      CALL GET_NEXT_NPBYTE      ; PlaySound.
      LD (IX+10), A             ; PLAYSOUND.
      CALL GET_NEXT_NPBYTE      ; SoundNum.
      LD (IX+11), A             ; SOUNDNUM.
      JP NODE_PARSER_MAIN       ; Done.

      ; Below is currently spare.
      NODE_PARSER_MAIN_0B_???:

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_0C_FILLSCREEN:
      INC HL
      LD A, (HL)                ; FillValue (from node data.)
      CALL FILL_SCREEN
      JP RETURN_TO_PARSER       ; Done.

      DB 6      ; Instruction length.
      NODE_PARSER_MAIN_0D_FILLAREA:
      LD IX, VT_FILL_AREA
      CALL NPS_00_COPYXYDIMS
      CALL GET_NEXT_NPBYTE      ; FillValue.
      LD (IX+4), A              ; FILLVALUE.
      CALL FILL_AREA
      JP RETURN_TO_PARSER       ; Done.

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_0E_PRINTFIXED:
      CALL GET_NEXT_NPBYTE      ; LineNum (get before HL is modified.)
      LD HL, VT_PS_DRAWFIXED    ; Copy SETPRINTFIXED table.
      LD DE, VT_PS_DRAWROW
      LDI
      LDI
      LDI
      CALL PRINT_STRING_ASLINE
      JP RETURN_TO_PARSER       ; Done.

      DB 4      ; Instruction length.
      NODE_PARSER_MAIN_0F_SETPRINTFIXED:
      LD IX, VT_PS_DRAWFIXED
      CALL GET_NEXT_NPBYTE      ; XCellPos.
      LD (IX+0), A              ; XPOS.
      CALL GET_NEXT_NPBYTE      ; YCellPos.
      LD (IX+1), A              ; YPOS.
      CALL GET_NEXT_NPBYTE      ; XCellClearLength.
      LD (IX+2), A              ; XLENGTH.
      JP RETURN_TO_PARSER       ; Done.

    NODE_PARSER_MAIN_1X:        ; 1X instructions.

      DB 4      ; Instruction length.
      NODE_PARSER_MAIN_10_DRAWBITMAP:
      CALL GET_NEXT_NPBYTE      ; BitmapNum (from node data.)
      LD D, A                   ; Preserve actual BitmapNum.
      CALL GET_NEXT_NPBYTE      ; XCellPos (from node data.)
      LD B, A
      CALL GET_NEXT_NPBYTE      ; YCellPos (from node data.)
      LD C, A
      LD A, D                   ; Restore actual BitmapNum.
      CALL DRAW_BGSTREAM
      JP RETURN_TO_PARSER       ; Done.

      ; Not used, removed to save memory.
      ;DB 4      ; Instruction length.
      NODE_PARSER_MAIN_11_PRINTGFXCHAR:
      ;LD IX, D_GFX_CHAR         ; Switch to Graphics tile bank.
      ;LD (V_TILE_DRAW_BANK_START), IX
      ;CALL GET_NEXT_NPBYTE      ; GraphicNum.
      ;LD D, A                   ; Preserve actual GraphicNum.
      ;CALL GET_NEXT_NPBYTE      ; XCellPos.
      ;LD B, A
      ;CALL GET_NEXT_NPBYTE      ; YCellPos.
      ;LD C, A
      ;LD A, D                   ; Restore actual GraphicNum.
      ;CALL TILE_DRAW
      ;JP RETURN_TO_PARSER       ; Done.

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_12_DRAWAGAIN:
      CALL GET_NEXT_NPBYTE      ; BitmapNum (from node data.)
      LD BC, (V_LASTDRAWPOS)    ; Get last XY position.
      CALL DRAW_BGSTREAM
      JP RETURN_TO_PARSER       ; Done.

      ; Not used, removed to save memory.
      ;DB 3      ; Instruction length.
      NODE_PARSER_MAIN_13_SETLASTDRAW:
      ;CALL GET_NEXT_NPBYTE      ; XCellPos.
      ;LD B, A
      ;CALL GET_NEXT_NPBYTE      ; YCellPos.
      ;LD C, A
      ;LD (V_LASTDRAWPOS), BC
      ;JP RETURN_TO_PARSER       ; Done.

      DB 6      ; Instruction length.
      NODE_PARSER_MAIN_14_SETATTRIBUTES:
      LD IX, VT_ATTRIBUTES
      CALL NPS_00_COPYXYDIMS    ; Pass X/Y Starts and Lengths (four bytes.)
      CALL GET_NEXT_NPBYTE      ; AttrValue (cannot have FLASH set directly.)
      LD (IX+4), A              ; Copy to ATTRVALUE.
      CALL CHECKGVAR_ATTRIBUTES
      CALL SET_ATTRIBUTES
      JP RETURN_TO_PARSER       ; Done.

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_15_SETALLATTRIBUTES:
      CALL GET_NEXT_NPBYTE      ; AttrValue (cannot have FLASH set directly.)
      CALL SET_ALLATTRS
      JP RETURN_TO_PARSER       ; Done.

      ; Not used, removed to save memory.
      ;DB 5      ; Instruction length.
      NODE_PARSER_MAIN_16_INVATTRIBUTES:
      ;LD IX, VT_ATTRIBUTES
      ;CALL NPS_00_COPYXYDIMS    ; Pass X/Y Starts and Lengths (four bytes.)
      ;CALL CHECKGVAR_ATTRIBUTES
      ;CALL INVERT_ATTRIBUTES
      ;JP RETURN_TO_PARSER       ; Done.

      ; Not used, removed to save memory.
      ;DB 5      ; Instruction length.
      NODE_PARSER_MAIN_17_INVALLATTRIBUTES:
      ;LD IX, VT_ATTRIBUTES
      ;LD (IX+0), 0              ; XSTART.
      ;LD (IX+1), 0              ; YSTART.
      ;LD (IX+2), 32             ; XLENGTH.
      ;LD (IX+3), 24             ; YLENGTH.
      ;CALL INVERT_ATTRIBUTES
      ;JP RETURN_TO_PARSER       ; Done.

      ; Not used, removed to save memory.
      ;DB 2                      ; Instruction length.
      NODE_PARSER_MAIN_18_LOADATTRIBUTES:
      ;CALL GET_NEXT_NPBYTE      ; AttrFileNum (from node data.)
      ;CALL LOAD_ATTRIBUTES
      ;JP RETURN_TO_PARSER       ; Done.

      ; Not used, removed to save memory.
      ;DB 1      ; Instruction length.
      NODE_PARSER_MAIN_19_SETATTRSCREEN:
      ;LD DE, $5800
      ;LD (V_ATTR_START), DE
      ;JP RETURN_TO_PARSER       ; Done.

      ; Not used, removed to save memory.
      ;DB 1      ; Instruction length.
      NODE_PARSER_MAIN_1A_SETATTRBUFFER:
      ;LD DE, ATTR_BUFFER
      ;LD (V_ATTR_START), DE
      ;JP RETURN_TO_PARSER       ; Done.

      ; Not used, removed to save memory.
      ;DB 1      ; Instruction length.
      NODE_PARSER_MAIN_1B_COPYATTRBUFFER:
      ;LD HL, ATTR_BUFFER
      ;LD DE, $5800
      ;LD BC, 768
      ;HALT                      ; Wait for refresh (to avoid screen tearing.)
      ;LDIR
      ;JP RETURN_TO_PARSER       ; Done.

      ; Not used, removed to save memory.
      ;DB 5      ; Instruction length.
      NODE_PARSER_MAIN_1C_COPYATTRAREA:
      ;LD IX, VT_ATTRIBUTES
      ;CALL NPS_00_COPYXYDIMS    ; Pass X/Y Starts and Lengths (four bytes.)
      ;CALL COPYATTR_AREA
      ;JP RETURN_TO_PARSER       ; Done.

      ; Not used, removed to save memory.
      ;DB 1      ; Instruction length.
      NODE_PARSER_MAIN_1D_SCRATTRTOBUFF:
      ;LD HL, $5800
      ;LD DE, ATTR_BUFFER
      ;LD BC, 768
      ;LDIR                      ; No need to wait for refresh.
      ;JP RETURN_TO_PARSER       ; Done.

      DB 6      ; Instruction length.
      NODE_PARSER_MAIN_1E_FLIPSCRX:
      LD IX, VT_SCRFLIP
      CALL NPS_00_COPYXYDIMS    ; Pass X/Y Starts and Lengths (four bytes.)
      CALL GET_NEXT_NPBYTE      ; XCellDest.
      LD (IX+4), A              ; XTOSTART.
      CALL SCRFLIP_X
      JP RETURN_TO_PARSER       ; Done.

      DB 6      ; Instruction length.
      NODE_PARSER_MAIN_1F_FLIPSCRY:
      LD IX, VT_SCRFLIP
      CALL NPS_00_COPYXYDIMS    ; Pass X/Y Starts and Lengths (four bytes.)
      CALL GET_NEXT_NPBYTE      ; YCellDest.
      LD (IX+4), A              ; XTOSTART.
      CALL SCRFLIP_Y
      JP RETURN_TO_PARSER       ; Done.

    NODE_PARSER_MAIN_2X:        ; 2X instructions.

      DB 3                      ; Instruction length.
      NODE_PARSER_MAIN_20_SETVAR:
      CALL GET_NEXT_NPBYTE      ; Variable number.
      INC HL                    ; Variable value.
      LD E, (HL)
      LD D, 0                   ; DE now contains new variable value.
      CALL SET_GVAR_FULL
      JP RETURN_TO_PARSER       ; Done.

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_21_INCVAR:
      CALL GET_NEXT_NPBYTE      ; Variable number.
      PUSH AF                   ; Store variable number.
      CALL GET_GVAR_FULL        ; DE now contains variable value.
      INC DE
      POP AF                    ; Restore variable number.
      CALL SET_GVAR_FULL        ; Copy DE back to variable.
      JP RETURN_TO_PARSER       ; Done.

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_22_DECVAR:
      CALL GET_NEXT_NPBYTE      ; Variable number.
      PUSH AF                   ; Store variable number.
      CALL GET_GVAR_FULL        ; DE now contains variable value.
      DEC DE
      POP AF                    ; Restore variable number.
      CALL SET_GVAR_FULL        ; Copy DE back to variable.
      JP RETURN_TO_PARSER       ; Done.

      DB 3      ; Instruction length.
      NODE_PARSER_MAIN_23_COPYVAR:
      CALL GET_NEXT_NPBYTE      ; VarFrom.
      LD B, A
      CALL GET_NEXT_NPBYTE      ; VarTo.
      LD C, A
      LD A, B
      CALL GET_GVAR_FULL        ; DE now contains source variable.
      LD A, C
      CALL SET_GVAR_FULL        ; Copy DE to destination variable.
      JP RETURN_TO_PARSER       ; Done.


      ; Not used, removed to save memory.
      ;DB 4      ; Instruction length.
      NODE_PARSER_MAIN_24_PRINTVAR:
      ;CALL GET_NEXT_NPBYTE      ; Variable number.
      ;PUSH HL                   ; Preserve bytecode pointer.
      ;CALL GET_GVAR_FULL        ; DE now contains variable value.
      ;EX DE, HL
      ;CALL DOUBLE_TO_STRING     ; Variable converted to string.
      ;POP HL                    ; Restore bytecode pointer.
      ;LD IX, (V_PS_DRAWROWVT)
      ;CALL GET_NEXT_NPBYTE      ; XCellPos.
      ;LD (IX+0), A              ; XSTART.
      ;CALL GET_NEXT_NPBYTE      ; YCellPos.
      ;LD (IX+1), A              ; YSTART.
      ;LD (IX+2), 5              ; DRAWLENGTH (fixed.)
      ;CALL PS_DRAWROW
      ;JP RETURN_TO_PARSER       ; Done.

      DB 3      ; Instruction length.
      NODE_PARSER_MAIN_25_ADDVAR:
      CALL GET_NEXT_NPBYTE      ; Variable number.
      LD B, A                   ; Preserve variable number.
      CALL GET_NEXT_NPB_FULLVAR ; DE now contains value to add.
      PUSH DE                   ; Preserve value to add.
      LD A, B                   ; Restore variable number.
      CALL GET_GVAR_FULL        ; DE now contains variable value.
      POP HL                    ; Restore value to add.
      ADD HL, DE                ; HL now contains new variable value.
      EX DE, HL
      LD A, B                   ; Restore variable number.
      CALL SET_GVAR_FULL
      JP RETURN_TO_PARSER       ; Done.

      DB 3      ; Instruction length.
      NODE_PARSER_MAIN_26_SUBVAR:
      CALL GET_NEXT_NPBYTE      ; Variable number.
      LD B, A                   ; Preserve variable number.
      CALL GET_NEXT_NPB_FULLVAR ; DE now contains value to add.
      PUSH DE                   ; Preserve value to subtract.
      LD A, B                   ; Restore variable number.
      CALL GET_GVAR_FULL        ; DE now contains variable value.
      POP HL                    ; Restore value to subtract.
      EX DE, HL
      XOR A                     ; Clear carry flag.
      SBC HL, DE                ; HL now contains new variable value.
      EX DE, HL
      LD A, B                   ; Restore variable number.
      CALL SET_GVAR_FULL
      JP RETURN_TO_PARSER       ; Done.

      DB 4                      ; Instruction length.
      NODE_PARSER_MAIN_27_SETVARFULL:
      CALL GET_NEXT_NPBYTE      ; Variable number.
      INC HL                    ; Variable value (low byte.)
      LD E, (HL)
      INC HL                    ; Variable value (high byte.)
      LD D, (HL)                ; DE now contains new variable value.
      CALL SET_GVAR_FULL
      JP RETURN_TO_PARSER       ; Done.

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_28_FLIPFLAG:
      CALL GET_NEXT_NPBYTE      ; Variable number.
      PUSH AF                   ; Store variable number.
      CALL GET_GVAR_FULL        ; DE now contains variable value.
      LD A, E                   ; Get low byte.
      XOR 00000001B             ; Flip lowest bit.
      LD E, A                   ; Store modified low byte.
      POP AF                    ; Restore variable number.
      CALL SET_GVAR_FULL        ; Copy DE back to variable.
      JP RETURN_TO_PARSER       ; Done.

      ; Below are currently spare.
      NODE_PARSER_MAIN_29_???:
      NODE_PARSER_MAIN_2A_???:
      NODE_PARSER_MAIN_2B_???:
      NODE_PARSER_MAIN_2C_???:
      NODE_PARSER_MAIN_2D_???:
      NODE_PARSER_MAIN_2E_???:
      NODE_PARSER_MAIN_2F_???:

    NODE_PARSER_MAIN_3X:        ; 3X instructions.

      ; To do later - This command to be completed.
      ;DB 4      ; Instruction length.
      NODE_PARSER_MAIN_30_SETTIMER:
      ;JP RETURN_TO_PARSER

      ; To do later - This command to be completed.
      ;DB 2      ; Instruction length.
      NODE_PARSER_MAIN_31_STARTTIMER:
      ;JP RETURN_TO_PARSER

      ; To do later - This command to be completed.
      ;DB 2      ; Instruction length.
      NODE_PARSER_MAIN_32_STOPTIMER:
      ;JP RETURN_TO_PARSER

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_33_WAITONE:
      HALT
      JP RETURN_TO_PARSER       ; Done.

      DB 3      ; Instruction length.
      NODE_PARSER_MAIN_34_WAITFOR:
      CALL GET_NEXT_NPB_FULLVAR ; MinWait.
      PUSH DE
      CALL GET_NEXT_NPB_FULLVAR ; MaxWait.
      POP BC
      CALL WAITFOR
      JP RETURN_TO_PARSER       ; Done.

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_35_WAITFORKEY:
      CALL WAITFOR_ANYKEY
      JP RETURN_TO_PARSER       ; Done.

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_36_WAITBEAMLOW:
      CALL WAIT_BEAMLOW
      JP RETURN_TO_PARSER       ; Done.

      ; Below is currently spare.
      NODE_PARSER_MAIN_37_???:

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_38_GETRANDOM:
      CALL RNG
      INC HL
      LD B, (HL)                ; MaxValMask.
      LD A, (V_RNG_NUMBER)
      AND B
      LD C, A
      LD B, 0                   ; BC now contains masked random number.
      LD (V_GVAR), BC           ; Store as first global variable.
      JP NODE_PARSER_MAIN       ; Done.

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_39_WAITFIXED:
      CALL GET_NEXT_NPB_FULLVAR ; WaitTime.
      PUSH DE
      POP BC                    ; Min. and max. WaitTimes are the same.
      CALL WAITFOR
      JP RETURN_TO_PARSER       ; Done.

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_3A_WAITONESEC:
      LD DE, 50                 ; Wait for 50 frames.
      PUSH DE
      POP BC                    ; Min. and max. WaitTimes are the same.
      CALL WAITFOR
      JP RETURN_TO_PARSER       ; Done.

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_3B_WAITTWOSECS:
      LD DE, 100                ; Wait for 50 frames.
      PUSH DE
      POP BC                    ; Min. and max. WaitTimes are the same.
      CALL WAITFOR
      JP RETURN_TO_PARSER       ; Done.

      ; Below are currently spare.
      NODE_PARSER_MAIN_3C_???:
      NODE_PARSER_MAIN_3D_???:
      NODE_PARSER_MAIN_3E_???:
      NODE_PARSER_MAIN_3F_???:

    NODE_PARSER_MAIN_4X:        ; 4X instructions.

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_40_PLAYSOUND:
      CALL GET_NEXT_NPBYTE      ; SoundNum.
      LD DE, 0                  ; DurationOverride (false.)
      CALL REQUEST_SOUND        ; Pass request to Scheduler.
      JP RETURN_TO_PARSER       ; Done.

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_41_SETBORDER:
      CALL GET_NEXT_NPBYTE      ; Colour.
      AND 00000111B             ; Clear other bits.
      OUT (254), A              ; Set border.
      LD (V_SND_BORDER), A      ; Store current border colour.
      JP RETURN_TO_PARSER       ; Done.

      DB 3      ; Instruction length.
      NODE_PARSER_MAIN_42_PLAYMEMNOISE:
      CALL GET_NEXT_NPB_FULLVAR ; MemStart.
      PUSH DE                   ; Lots of PUSH/POPping to preserve HL.
      POP BC
      CALL GET_NEXT_NPB_FULLVAR ; MemLength.
      CALL PLAY_MEMNOISE
      JP RETURN_TO_PARSER       ; Done.

      ; Not used, removed to save memory.
      ;DB 7      ; Instruction length.
      NODE_PARSER_MAIN_43_ADDMEMGRAIN:
      ;LD IX, VT_ADD_MEMGRAIN
      ;CALL NPS_00_COPYXYDIMS
      ;CALL GET_NEXT_NPB_FULLVAR ; MemStart.
      ;INC HL
      ;LD A, (HL)                ; Mask.
      ;LD (IX+4), A              ; BITMASK.
      ;CALL ADD_MEMGRAIN
      ;JP RETURN_TO_PARSER       ; Done.

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_44_PLAYCLICK:
      LD B, 32                  ; Set click length (longer is louder.)
      CALL PLAY_CLICK
      JP RETURN_TO_PARSER       ; Done.

      DB 3      ; Instruction length.
      NODE_PARSER_MAIN_45_STARTSOUNDSEQ:
      LD IX, VT_SOUNDSEQ
      CALL GET_NEXT_NPBYTE      ; Pass SequenceNum...
      LD (IX+2), A              ; ...to SEQNUM.
      PUSH AF                   ; Preserve SequenceNum.
      CALL GET_NEXT_NPBYTE      ; Pass RepeatFor...
      LD (IX+3), A              ; ...to RPTCOUNT.
      POP AF                    ; Restore SequenceNum.
      CALL SET_SOUNDSEQ
      JP RETURN_TO_PARSER       ; Done.

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_46_STOPSOUNDSEQ:
      LD IX, VT_SOUNDSEQ
      LD (IX+0), 0              ; Reset ACTIVE.
      JP RETURN_TO_PARSER       ; Done.

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_47_SETVOLUME:
      CALL GET_NEXT_NPBYTE      ; Volume.
      RLCA                      ; Shift Volume bit from B0 to B4.
      RLCA
      RLCA
      RLCA
      LD (V_SND_VOLMASK), A     ; Store Volume bit as ready-to-use mask.
      JP RETURN_TO_PARSER       ; Done.

      DB 3      ; Instruction length.
      NODE_PARSER_MAIN_48_PLAYSOUNDFOR:
      CALL GET_NEXT_NPBYTE      ; SoundNum.
      PUSH AF                   ; Preserve SoundNum.
      CALL GET_NEXT_NPB_FULLVAR ; Duration.
      POP AF                    ; Restore SoundNum.
      CALL REQUEST_SOUND        ; Pass request to Scheduler.
      JP RETURN_TO_PARSER       ; Done.

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_49_WAITSOUNDSEQ:
      LD IX, VT_SOUNDSEQ
      LD A, (IX+0)              ; Check if ACTIVE.
      AND A
      JP Z, RETURN_TO_PARSER    ; No sound sequence active, so done waiting.

      ; To do - patch in "other sound playing" check.
      ; (Might need to do this before returning if no sound sequence is active.)

      LD A, (V_SOUNDSEQ_KEYBRK) ; Check if breaking allowed and key was pressed.
      LD B, A
      LD A, (V_KEYBOARD_KEY_ANY)
      AND B
      JR Z, NPM_49_JP1

      LD (IX+0), 0              ; Reset ACTIVE.
      LD BC, 1                  ; Set BROKESEQ flag (in GVAR 0.)
      LD (V_GVAR), BC
      JP RETURN_TO_PARSER       ; Done.

        NPM_49_JP1:
        LD HL, V_NODE_POINTER   ; Sound sequence active, so wait one frame...
        DEC (HL)                ; ...by repeating this instruction.
        HALT                    ;
        JP RETURN_TO_PARSER     ; Done for now.

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_4A_FORCESOUNDSEQ:
      CALL GET_NEXT_NPBYTE      ; ForceSoundSeq.
      XOR 00000001B             ; Flip flag.
      LD (V_SOUNDSEQ_KEYBRK), A ; Store as flipped.
      JP RETURN_TO_PARSER       ; Done.


      ; Below are currently spare.
      NODE_PARSER_MAIN_4B_???:
      NODE_PARSER_MAIN_4C_???:
      NODE_PARSER_MAIN_4D_???:
      NODE_PARSER_MAIN_4E_???:
      NODE_PARSER_MAIN_4F_???:

    NODE_PARSER_MAIN_5X:        ; 5X instructions.

      ; Below are currently spare.
      NODE_PARSER_MAIN_50_???:
      NODE_PARSER_MAIN_51_???:
      NODE_PARSER_MAIN_52_???:
      NODE_PARSER_MAIN_53_???:
      NODE_PARSER_MAIN_54_???:
      NODE_PARSER_MAIN_55_???:
      NODE_PARSER_MAIN_56_???:
      NODE_PARSER_MAIN_57_???:
      NODE_PARSER_MAIN_58_???:
      NODE_PARSER_MAIN_59_???:
      NODE_PARSER_MAIN_5A_???:
      NODE_PARSER_MAIN_5B_???:
      NODE_PARSER_MAIN_5C_???:
      NODE_PARSER_MAIN_5D_???:
      NODE_PARSER_MAIN_5E_???:
      NODE_PARSER_MAIN_5F_???:


    NODE_PARSER_MAIN_6X:        ; 6X instructions.

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_60_CREATEFORMNUM:
      CALL GET_NEXT_NPBYTE      ; TextVar.
      CALL CREATE_FORMNUM
      JP RETURN_TO_PARSER       ; Done.

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_61_SETPEOPLE:
      CALL SET_PEOPLE
      JP RETURN_TO_PARSER       ; Done.

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_62_SHOWTIME:
      ; Not used, removed to save memory.
      ;CALL SHOW_TIME
      JP RETURN_TO_PARSER       ; Done.

      ; Not used, removed to save memory.
      ;DB 5      ; Instruction length.
      NODE_PARSER_MAIN_63_SETUPTIME:
      ;LD IX, VT_SHOW_TIME
      ;CALL NPS_00_COPYXYDIMS    ; N.B. 4th value is XOFFSET, but order correct.
      ;JP RETURN_TO_PARSER       ; Done.

      ; Not used, removed to save memory.
      ;DB 1      ; Instruction length.
      NODE_PARSER_MAIN_64_CLEARTIME:
      ;LD HL, VT_TIMER_CLOCK
      ;PUSH HL
      ;POP DE
      ;INC HL                    ; Zero first of the four clock bytes.
      ;LD (HL), 0
      ;LD BC, 3                  ; Zero the remaining three clock bytes.
      ;LDIR
      ;JP RETURN_TO_PARSER       ; Done.

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_65_WIPEPGPH:
      CALL PS_WIPE_PGPH
      JP RETURN_TO_PARSER       ; Done.

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_66_PRINTNEXT:
      LD A, (V_PSAP_LASTSTR)
      INC A
      CALL PRINT_STRING_ASPGPH
      JP RETURN_TO_PARSER       ; Done.

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_67_SETPHONE:
      CALL GET_NEXT_NPBYTE      ; Copy PhoneType...
      LD IX, VT_DIAL_NUMBER
      LD (IX+0), A              ; ...to PHONETYPE.
      JP RETURN_TO_PARSER       ; Done.

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_68_DIALNUMBER:
      CALL GET_NEXT_NPBYTE      ; PhoneNumber.
      CALL DIAL_NUMBER
      JP RETURN_TO_PARSER       ; Done.

      DB 9      ; Instruction length.
      NODE_PARSER_MAIN_69_BEAMV:
      LD IX, VT_VFX_BEAMV
      CALL NPS_00_COPYXYDIMS    ; Pass X/Y Starts and Lengths (four bytes.)
      CALL GET_NEXT_NPBYTE      ; YCellSource.
      LD (IX+5), A              ; YSOURCE.
      CALL GET_NEXT_NPBYTE      ; Down.
      PUSH AF                   ; Preserve Down.
      CALL GET_NEXT_NPBYTE      ; Delay.
      LD (IX+7), A              ; DELAY.
      CALL GET_NEXT_NPBYTE      ; Sound.
      LD (IX+8), A              ; SOUND.
      POP AF                    ; Restore Down.
      CALL VFX_BEAMV
      JP RETURN_TO_PARSER       ; Done.

      DB 9      ; Instruction length.
      NODE_PARSER_MAIN_6A_SCROLLPIXV:
      LD IX, VT_VFX_SCROLLPIXV
      CALL GET_NEXT_NPBYTE      ; Pass XCellStart...
      LD (IX+0), A              ; ...to XFROMCL.
      CALL GET_NEXT_NPBYTE      ; Pass XCellLength...
      LD (IX+2), A              ; ...to XLENGTHCL.
      CALL GET_NEXT_NPBYTE      ; Pass Steps...
      LD (IX+5), A              ; ...to STEPS.
      CALL GET_NEXT_NPBYTE      ; Pass Down...
      LD (IX+6), A              ; ...to YVEL.
      INC HL
      LD A, (HL)                ; Pass YPixStart...
      LD (IX+1), A              ; ...to YFROMPX.
      INC HL
      LD A, (HL)                ; Pass YPixLength...
      LD (IX+3), A              ; ...to YLENGTHPX.
      INC HL
      LD A, (HL)                ; Pass YPixSource...
      LD (IX+4), A              ; ...to YSOURCEPX.
      CALL GET_NEXT_NPBYTE      ; Pass XCellSource...
      LD (IX+11), A             ; ...to XSOURCECL.
      CALL VFX_SCROLLPIXV
      JP RETURN_TO_PARSER       ; Done.

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_6B_REGENTABLES:
      CALL GEN_FLIPX_LOOKUP
      JP RETURN_TO_PARSER       ; Done.

      DB 5      ; Instruction length.
      NODE_PARSER_MAIN_6C_BENDSOUND:
      LD IX, VT_SFX_BENDSND
      CALL GET_NEXT_NPBYTE      ; Pass SourceSnd...
      LD (IX+0), A              ; ...to SOURCESND.
      CALL GET_NEXT_NPBYTE      ; Pass Steps...
      LD (IX+1), A              ; ...to STEPS.
      CALL GET_NEXT_NPB_FULLVAR ; Pass HPitchStep...
      LD (IX+2), E              ; ...to HPITCHSTEP.
      LD (IX+3), D
      CALL GET_NEXT_NPB_FULLVAR ; Pass LPitchStep...
      LD (IX+4), E              ; ...to LPITCHSTEP.
      LD (IX+5), D
      CALL SFX_BENDSND
      JP RETURN_TO_PARSER       ; Done.

      DB 5      ; Instruction length.
      NODE_PARSER_MAIN_6D_SETSCROLLV:
      LD IX, VT_VFX_SCROLLPIXV
      CALL GET_NEXT_NPBYTE      ; Pass Delay...
      LD (IX+7), A              ; ...to DELAY.
      CALL GET_NEXT_NPBYTE      ; Pass WithSound...
      LD (IX+8), A              ; ...to WITHSND.
      CALL GET_NEXT_NPBYTE      ; Pass SoundNum...
      LD (IX+9), A              ; ...to SOUNDNUM.
      CALL GET_NEXT_NPBYTE      ; Pass BendSound...
      LD (IX+10), A              ; ...to BENDSND.
      JP RETURN_TO_PARSER       ; Done.

      DB 6      ; Instruction length.
      NODE_PARSER_MAIN_6E_SHIFTAREAH:
      LD IX, VT_SHFTAREA_H
      CALL NPS_00_COPYXYDIMS    ; Pass X/Y Starts and Lengths (four bytes.)
      CALL GET_NEXT_NPBYTE      ; XPlus.
      LD (IX+4), A              ; XPLUS
      CALL SHFTAREA_H
      JP RETURN_TO_PARSER       ; Done.

      DB 3      ; Instruction length.
      NODE_PARSER_MAIN_6F_PLAYRUMBLE:
      CALL GET_NEXT_NPB_FULLVAR ; MemStart.
      CALL GET_NEXT_NPBYTE      ; Steps.
      LD C, A                   ; Move Steps.
      EX DE, HL                 ; Move MemStart.
      CALL PLAY_RUMBLE
      JP RETURN_TO_PARSER       ; Done.

    NODE_PARSER_MAIN_7X:        ; 7X instructions.

      DB 7      ; Instruction length.
      NODE_PARSER_MAIN_70_SCRCOPY:
      LD IX, VT_SCRFLIP
      CALL NPS_00_COPYXYDIMS    ; Pass X/Y Starts and Lengths (four bytes.)
      CALL GET_NEXT_NPBYTE      ; XCellDest.
      LD (IX+4), A              ; XTO.
      CALL GET_NEXT_NPBYTE      ; YCellDest.
      LD (IX+5), A              ; YTO.
      CALL SCRCOPY
      JP RETURN_TO_PARSER       ; Done.

      DB 4      ; Instruction length.
      NODE_PARSER_MAIN_71_HLINE:
      LD IX, VT_HLINE
      CALL GET_NEXT_NPBYTE      ; XPixStart.
      LD (IX+0), A              ; XSTART.
      CALL GET_NEXT_NPBYTE      ; YPixStart.
      LD (IX+1), A              ; YSTART.
      CALL GET_NEXT_NPBYTE      ; XPixLength.
      LD (IX+2), A              ; XLENGTH.
      CALL HLINE
      JP RETURN_TO_PARSER       ; Done.

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_72_HLINENEXT:
      CALL HLINE_NEXT
      JP RETURN_TO_PARSER       ; Done.

      DB 3      ; Instruction length.
      NODE_PARSER_MAIN_73_SETATTRSROW:
      CALL GET_NEXT_NPBYTE      ; AttrRow.
      PUSH AF                   ; Preserve AttrRow.
      CALL GET_NEXT_NPB_FULLVAR ; DE now contains MemStart.
      POP AF                    ; Restore AttrRow.
      CALL VFX_SETATTRSROW
      JP RETURN_TO_PARSER       ; Done.

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_74_PLAYCLICKFOR:
      CALL GET_NEXT_NPBYTE      ; ClickLength.
      LD B, A
      AND A                     ; Check non-zero.
      CALL NZ, PLAY_CLICK
      JP RETURN_TO_PARSER       ; Done.

      ; Not used, removed to save memory.
      ;DB 4      ; Instruction length.
      NODE_PARSER_MAIN_75_WIPEROWS:
      ;CALL GET_NEXT_NPBYTE      ; YPixStart.
      ;LD B, A
      ;CALL GET_NEXT_NPBYTE      ; ForRows.
      ;LD C, A
      ;CALL GET_NEXT_NPBYTE      ; YPixStep.
      ;LD D, A
      ;CALL VFX_WIPEROWS
      ;JP RETURN_TO_PARSER       ; Done.

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_76_DECBREAKZ:
      CALL GET_NEXT_NPBYTE      ; Variable number.
      PUSH AF                   ; Store variable number.
      CALL GET_GVAR_FULL        ; DE now contains variable value.
      DEC DE
      POP AF                    ; Restore variable number.
      CALL SET_GVAR_FULL        ; Copy DE back to variable.
      LD A, D                   ; Check if variable reached zero.
      OR E                      ; If so, break from subroutine.
      JP Z, NODE_PARSER_MAIN_B2_BREAKSUB
      JP RETURN_TO_PARSER       ; Otherwise, continue as normal.

      DB 3      ; Instruction length.
      NODE_PARSER_MAIN_77_INCVARLOOPAT:
      CALL GET_NEXT_NPBYTE      ; Variable number.
      LD B, A                   ; Store variable number.
      CALL GET_NEXT_NPB_FULLVAR ; Variable loop value (in DE.)
      EX DE, HL
      LD A, B                   ; Restore variable number.
      CALL GET_GVAR_FULL        ; DE now contains variable value.
      INC DE
      OR A                      ; Clear carry flag.
      SBC HL, DE                ; Loop value minus new variable value.
      JR NZ, NPM_77_JP1
      EX DE, HL                 ; If HL minus DE was zero, then set DE to zero.

        NPM_77_JP1:
        LD A, B                 ; Restore variable number.
        CALL SET_GVAR_FULL      ; Copy DE back to variable.
        JP RETURN_TO_PARSER     ; Done.

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_78_HLINERESET:
      LD IX, VT_HLINE_NEXT
      LD A, (IX+1)              ; LISTLENGTH.
      DEC A
      LD (IX+0), A              ; LISTSTEP.
      JP RETURN_TO_PARSER       ; Done.

      DB 3      ; Instruction length.
      NODE_PARSER_MAIN_79_RINGPHONE:
      CALL GET_NEXT_NPBYTE      ; Phone type.
      PUSH AF
      CALL GET_NEXT_NPBYTE      ; Repeat for.
      LD E, A
      POP AF
      CALL RING_PHONE
      JP RETURN_TO_PARSER       ; Done.


      ; Below are currently spare.
      NODE_PARSER_MAIN_7A_???:
      NODE_PARSER_MAIN_7B_???:
      NODE_PARSER_MAIN_7C_???:
      NODE_PARSER_MAIN_7D_???:
      NODE_PARSER_MAIN_7E_???:
      NODE_PARSER_MAIN_7F_???:

    NODE_PARSER_MAIN_8X:        ; 8X instructions.

      DB 3      ; Instruction length.
      NODE_PARSER_MAIN_80_IFVAR_EQ_VAR:
      LD C, 00000101B           ; Set flags for: TRUE if (Comp-Var) is Z.
      JP EVALUATE_IFVAR

      DB 3      ; Instruction length.
      NODE_PARSER_MAIN_81_IFVAR_NEQ_VAR:
      LD C, 00000100B           ; Set flags for: TRUE if (Comp-Var) is NZ.
      JP EVALUATE_IFVAR

      DB 3      ; Instruction length.
      NODE_PARSER_MAIN_82_IFVAR_GT_VAR:
      LD C, 00000111B           ; Set flags for: TRUE if (Comp-Var) is C.
      JP EVALUATE_IFVAR

      DB 3      ; Instruction length.
      NODE_PARSER_MAIN_83_IFVAR_LT_VAR:
      LD C, 00000011B           ; Set flags for: TRUE if (Var-Comp) is C.
      JP EVALUATE_IFVAR

      DB 3      ; Instruction length.
      NODE_PARSER_MAIN_84_IFVAR_GTE_VAR:
      LD C, 00000010B           ; Set flags for: TRUE if (Var-Comp) is NC.
      JP EVALUATE_IFVAR

      DB 3      ; Instruction length.
      NODE_PARSER_MAIN_85_IFVAR_LTE_VAR:
      LD C, 00000110B           ; Set flags for: TRUE if (Comp-Var) is NC.
      JP EVALUATE_IFVAR

      ; Below are currently spare.
      NODE_PARSER_MAIN_86_???:
      NODE_PARSER_MAIN_87_???:

      DB 4      ; Instruction length.
      NODE_PARSER_MAIN_88_IFVAR_EQ_NUM:
      LD C, 00001101B           ; Set flags for: TRUE if (Value-Var) is Z.
      JP EVALUATE_IFVAR

      DB 4      ; Instruction length.
      NODE_PARSER_MAIN_89_IFVAR_NEQ_NUM:
      LD C, 00001100B           ; Set flags for: TRUE if (Value-Var) is Z.
      JP EVALUATE_IFVAR

      DB 4      ; Instruction length.
      NODE_PARSER_MAIN_8A_IFVAR_GT_NUM:
      LD C, 00001111B           ; Set flags for: TRUE if (Value-Var) is C.
      JP EVALUATE_IFVAR

      DB 4      ; Instruction length.
      NODE_PARSER_MAIN_8B_IFVAR_LT_NUM:
      LD C, 00001011B           ; Set flags for: TRUE if (Var-Value) is C.
      JP EVALUATE_IFVAR
      ; TRUE if (Var-Value) is C.

      DB 4      ; Instruction length.
      NODE_PARSER_MAIN_8C_IFVAR_GTE_NUM:
      LD C, 00001010B           ; Set flags for: TRUE if (Var-Value) is NC.
      JP EVALUATE_IFVAR

      DB 4      ; Instruction length.
      NODE_PARSER_MAIN_8D_IFVAR_LTE_NUM:
      LD C, 00001110B           ; Set flags for: TRUE if (Value-Var) is NC.
      JP EVALUATE_IFVAR

      ; Below are currently spare.
      NODE_PARSER_MAIN_8E_???:
      NODE_PARSER_MAIN_8F_???:

    NODE_PARSER_MAIN_9X:        ; 9X instructions.

      ; Below are currently spare.
      NODE_PARSER_MAIN_90_???:
      NODE_PARSER_MAIN_91_???:
      NODE_PARSER_MAIN_92_???:
      NODE_PARSER_MAIN_93_???:
      NODE_PARSER_MAIN_94_???:
      NODE_PARSER_MAIN_95_???:
      NODE_PARSER_MAIN_96_???:
      NODE_PARSER_MAIN_97_???:
      NODE_PARSER_MAIN_98_???:
      NODE_PARSER_MAIN_99_???:
      NODE_PARSER_MAIN_9A_???:
      NODE_PARSER_MAIN_9B_???:
      NODE_PARSER_MAIN_9C_???:
      NODE_PARSER_MAIN_9D_???:
      NODE_PARSER_MAIN_9E_???:
      NODE_PARSER_MAIN_9F_???:

    NODE_PARSER_MAIN_AX:        ; AX instructions.

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_A0_ONKEY1:
      LD A, $00                 ; Code for "1" key.
      CALL ONKEY
      JP RETURN_TO_PARSER       ; Done.

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_A1_ONKEY2:
      LD A, $01                 ; Code for "2" key.
      CALL ONKEY
      JP RETURN_TO_PARSER       ; Done.

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_A2_ONKEY3:
      LD A, $02                 ; Code for "3" key.
      CALL ONKEY
      JP RETURN_TO_PARSER       ; Done.

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_A3_ONKEY4:
      LD A, $03                 ; Code for "4" key.
      CALL ONKEY
      JP RETURN_TO_PARSER       ; Done.

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_A4_ONKEY5:
      LD A, $04                 ; Code for "5" key.
      CALL ONKEY
      JP RETURN_TO_PARSER       ; Done.

      ; Below are currently spare.
      NODE_PARSER_MAIN_A5_???:
      NODE_PARSER_MAIN_A6_???:
      NODE_PARSER_MAIN_A7_???:
      NODE_PARSER_MAIN_A8_???:
      NODE_PARSER_MAIN_A9_???:
      NODE_PARSER_MAIN_AA_???:
      NODE_PARSER_MAIN_AB_???:
      NODE_PARSER_MAIN_AC_???:
      NODE_PARSER_MAIN_AD_???:
      NODE_PARSER_MAIN_AE_???:
      NODE_PARSER_MAIN_AF_???:

    NODE_PARSER_MAIN_BX:        ; BX instructions.

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_B0_CHANGENODE:
      INC HL
      LD A, (HL)                ; Get next Node.
      CALL GET_IFGVAR_LOWBYTE   ; Check for variable reference.
      LD (V_NODE_NEXT), A
      LD HL, PSR_CALL_STACK     ; Reset both Parser stacks.
      LD (V_NODE_CALLSTACKP), HL
      LD HL, PSR_IFKY_STACK
      LD (V_NODE_IFKYSTACKP), HL
      XOR A                     ; Reset both IF and KEY skip stacks.
      LD (V_NODE_IFDEPTH), A
      LD (V_NODE_IFSKIP), A
      LD (V_NODE_KEYDEPTH), A
      LD (V_NODE_KEYSKIP), A
      LD (V_NODE_STATE), A      ; Set Node state to ENTRY.
      JP NODE_PARSER            ; Run new Node (or same one) from top.

      DB 2      ; Instruction length.
      NODE_PARSER_MAIN_B1_CALLSUBNODE:
      INC HL
      LD A, (HL)                ; Get SubNode number.
      PUSH AF                   ; Hold SubNode number.
      CALL PUSH_CALLSTACK       ; Store next instruction position -1.
      CALL PUSH_IFKYSTACK       ; Store current IF and ONKEY depths.
      POP AF                    ; Restore SubNode number.
      CALL GET_IFGVAR_LOWBYTE   ; Check for variable reference.
      LD HL, I_NODE_SUB         ; Get SubNode instruction start address
      CALL GET_ADDRESS_UNALIGNED
      DEC HL                    ; Allow for INC before next read.
      JP NODE_PARSER_MAIN       ; - SubNodes do not have ONSTATEs to be skipped.

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_B2_BREAKSUB:
      CALL POP_IFKYSTACK        ; Restore previous IF and ONKEY depths.
      CALL POP_CALLSTACK        ; Restore previous instruction position -1.
      JP NODE_PARSER_MAIN

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_B3_REPEATSUB:
      CALL POP_IFKYSTACK        ; Restore previous IF and ONKEY depths.
      CALL POP_CALLSTACK        ; Restore previous instruction position -1.
      DEC HL                    ; Now HL points back to CALLSUBNODE instruction.
      JP NODE_PARSER_MAIN_B1_CALLSUBNODE ; Retrigger the SubNode.

      ; Below are currently spare.
      NODE_PARSER_MAIN_B4_???:
      NODE_PARSER_MAIN_B5_???:
      NODE_PARSER_MAIN_B6_???:
      NODE_PARSER_MAIN_B7_???:
      NODE_PARSER_MAIN_B8_???:
      NODE_PARSER_MAIN_B9_???:
      NODE_PARSER_MAIN_BA_???:
      NODE_PARSER_MAIN_BB_???:
      NODE_PARSER_MAIN_BC_???:
      NODE_PARSER_MAIN_BD_???:
      NODE_PARSER_MAIN_BE_???:
      NODE_PARSER_MAIN_BF_???:

    NODE_PARSER_MAIN_CX:        ; CX instructions.

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_C0_ENDIF:
      PUSH HL
      LD A, 00000000B           ; Flags: B2=0(IF), B1=0(-), B0=0(DEC DEPTH.)
      CALL SET_DEPTH_SKIP
      POP HL
      JP NODE_PARSER_MAIN       ; Using this method as V_NODE_POINTER is wrong.


      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_C1_ENDKEY:
      PUSH HL
      LD A, 00000100B           ; Flags: B2=1(KEY), B1=0(-), B0=0(DEC DEPTH.)
      CALL SET_DEPTH_SKIP
      POP HL
      JP NODE_PARSER_MAIN       ; Using this method as V_NODE_POINTER is wrong.

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_C2_ENDSTATE:
      LD DE, V_NODE_STATE
      LD A, (DE)
      CP 1                      ; Check if state was 1 - CURRENT.
      JP Z, NODE_PARSER         ; If so, repeat CURRENT state from top.
      INC A                     ; Otherwise, state was 0 - ENTRY.
      LD (DE), A                ; Change state to CURRENT.
      JP NODE_PARSER            ; Run new state from top.

      DB 1      ; Instruction length.
      NODE_PARSER_MAIN_C3_ENDSUBNODE:
      CALL POP_IFKYSTACK        ; Restore previous IF and ONKEY depths.
      CALL POP_CALLSTACK        ; Restore previous instruction position -1.
      JP NODE_PARSER_MAIN

      ; Below are currently spare.
      NODE_PARSER_MAIN_C4_???:
      NODE_PARSER_MAIN_C5_???:
      NODE_PARSER_MAIN_C6_???:
      NODE_PARSER_MAIN_C7_???:
      NODE_PARSER_MAIN_C8_???:
      NODE_PARSER_MAIN_C9_???:
      NODE_PARSER_MAIN_CA_???:
      NODE_PARSER_MAIN_CB_???:
      NODE_PARSER_MAIN_CC_???:
      NODE_PARSER_MAIN_CD_???:
      NODE_PARSER_MAIN_CE_???:
      NODE_PARSER_MAIN_CF_???:

    NODE_PARSER_MAIN_DX:        ; DX instructions.

      ; Below are currently spare.
      NODE_PARSER_MAIN_D0_???:
      NODE_PARSER_MAIN_D1_???:
      NODE_PARSER_MAIN_D2_???:
      NODE_PARSER_MAIN_D3_???:
      NODE_PARSER_MAIN_D4_???:
      NODE_PARSER_MAIN_D5_???:
      NODE_PARSER_MAIN_D6_???:
      NODE_PARSER_MAIN_D7_???:
      NODE_PARSER_MAIN_D8_???:
      NODE_PARSER_MAIN_D9_???:
      NODE_PARSER_MAIN_DA_???:
      NODE_PARSER_MAIN_DB_???:
      NODE_PARSER_MAIN_DC_???:
      NODE_PARSER_MAIN_DD_???:
      NODE_PARSER_MAIN_DE_???:
      NODE_PARSER_MAIN_DF_???:

    NODE_PARSER_MAIN_EX:        ; EX instructions.

      ; Below are currently spare.
      NODE_PARSER_MAIN_E0_???:
      NODE_PARSER_MAIN_E1_???:
      NODE_PARSER_MAIN_E2_???:
      NODE_PARSER_MAIN_E3_???:
      NODE_PARSER_MAIN_E4_???:
      NODE_PARSER_MAIN_E5_???:
      NODE_PARSER_MAIN_E6_???:
      NODE_PARSER_MAIN_E7_???:
      NODE_PARSER_MAIN_E8_???:
      NODE_PARSER_MAIN_E9_???:
      NODE_PARSER_MAIN_EA_???:
      NODE_PARSER_MAIN_EB_???:
      NODE_PARSER_MAIN_EC_???:
      NODE_PARSER_MAIN_ED_???:
      NODE_PARSER_MAIN_EE_???:
      NODE_PARSER_MAIN_EF_???:

    NODE_PARSER_MAIN_FX:        ; FX instructions.

      ; Below are currently spare.
      NODE_PARSER_MAIN_F0_???:
      NODE_PARSER_MAIN_F1_???:
      NODE_PARSER_MAIN_F2_???:
      NODE_PARSER_MAIN_F3_???:
      NODE_PARSER_MAIN_F4_???:
      NODE_PARSER_MAIN_F5_???:
      NODE_PARSER_MAIN_F6_???:
      NODE_PARSER_MAIN_F7_???:
      NODE_PARSER_MAIN_F8_???:
      NODE_PARSER_MAIN_F9_???:
      NODE_PARSER_MAIN_FA_???:
      NODE_PARSER_MAIN_FB_???:
      NODE_PARSER_MAIN_FC_???:
      NODE_PARSER_MAIN_FD_???:
      NODE_PARSER_MAIN_FE_???:
      NODE_PARSER_MAIN_FF_???:

; Crash if jump to last block of unwritten instructions.
DI
HALT

V_NODE_CURRENT:
DB 0                            ; Start at Node 0.
V_NODE_NEXT:
DB 0                            ; Start at Node 0.
V_NODE_STATE:
DB 0                            ; Start at 0=Entry (1=Current.)

V_NODE_CALLSTACKP:
DW PSR_CALL_STACK               ; Initialise pointers at start of stack spaces.
V_NODE_IFKYSTACKP:
DW PSR_IFKY_STACK

V_NODE_IFDEPTH:
DB 0                            ; For nested IF statements.
V_NODE_IFSKIP:
DB 0				; Skip until at least next IFKEY statement.
V_NODE_IFSKIPSTACK:             ; Stores T/F flag for each IFDEPTH level.
DW 0                            ; (16 bytes long, so maximum IFDEPTH is 127.)
DW 0
DW 0
DW 0
DW 0
DW 0
DW 0
DW 0

V_NODE_KEYDEPTH:
DB 0                            ; For nested ONKEY statements.
V_NODE_KEYSKIP:
DB 0				; Skip until at least next ENDKEY statement.
V_NODE_KEYSKIPSTACK:            ; Stores T/F flag for each KEYDEPTH level.
DW 0                            ; (16 bytes long, so maximum KEYDEPTH is 127.)
DW 0
DW 0
DW 0
DW 0
DW 0
DW 0
DW 0

I_NODE_PARSER_INSTRUCTIONS:     ; Index of addresses for jumps.
  DW NODE_PARSER_MAIN_00_PRINTLINE
  DW NODE_PARSER_MAIN_01_PRINTPGPH
  DW NODE_PARSER_MAIN_02_CLEARPGPH
  DW NODE_PARSER_MAIN_03_SETPGPHAREA
  DW NODE_PARSER_MAIN_04_SETPGPHSPACE
  DW NODE_PARSER_MAIN_05_SETPGPHMODE
  DW NODE_PARSER_MAIN_06_SETPGPHDELAY
  DW NODE_PARSER_MAIN_07_SETPGPHPROMPT
  DW NODE_PARSER_MAIN_08_SETLINEBANKHI
  DW NODE_PARSER_MAIN_09_SETLINEBANKLO
  DW NODE_PARSER_MAIN_0A_SETPGPHSOUND
  DW NODE_PARSER_MAIN_0B_???
  DW NODE_PARSER_MAIN_0C_FILLSCREEN
  DW NODE_PARSER_MAIN_0D_FILLAREA
  DW NODE_PARSER_MAIN_0E_PRINTFIXED
  DW NODE_PARSER_MAIN_0F_SETPRINTFIXED

  DW NODE_PARSER_MAIN_10_DRAWBITMAP
  DW NODE_PARSER_MAIN_11_PRINTGFXCHAR
  DW NODE_PARSER_MAIN_12_DRAWAGAIN
  DW NODE_PARSER_MAIN_13_SETLASTDRAW
  DW NODE_PARSER_MAIN_14_SETATTRIBUTES
  DW NODE_PARSER_MAIN_15_SETALLATTRIBUTES
  DW NODE_PARSER_MAIN_16_INVATTRIBUTES
  DW NODE_PARSER_MAIN_17_INVALLATTRIBUTES
  DW NODE_PARSER_MAIN_18_LOADATTRIBUTES
  DW NODE_PARSER_MAIN_19_SETATTRSCREEN
  DW NODE_PARSER_MAIN_1A_SETATTRBUFFER
  DW NODE_PARSER_MAIN_1B_COPYATTRBUFFER
  DW NODE_PARSER_MAIN_1C_COPYATTRAREA
  DW NODE_PARSER_MAIN_1D_SCRATTRTOBUFF
  DW NODE_PARSER_MAIN_1E_FLIPSCRX
  DW NODE_PARSER_MAIN_1F_FLIPSCRY

  DW NODE_PARSER_MAIN_20_SETVAR
  DW NODE_PARSER_MAIN_21_INCVAR
  DW NODE_PARSER_MAIN_22_DECVAR
  DW NODE_PARSER_MAIN_23_COPYVAR
  DW NODE_PARSER_MAIN_24_PRINTVAR
  DW NODE_PARSER_MAIN_25_ADDVAR
  DW NODE_PARSER_MAIN_26_SUBVAR
  DW NODE_PARSER_MAIN_27_SETVARFULL
  DW NODE_PARSER_MAIN_28_FLIPFLAG
  DW NODE_PARSER_MAIN_29_???
  DW NODE_PARSER_MAIN_2A_???
  DW NODE_PARSER_MAIN_2B_???
  DW NODE_PARSER_MAIN_2C_???
  DW NODE_PARSER_MAIN_2D_???
  DW NODE_PARSER_MAIN_2E_???
  DW NODE_PARSER_MAIN_2F_???

  DW NODE_PARSER_MAIN_30_SETTIMER
  DW NODE_PARSER_MAIN_31_STARTTIMER
  DW NODE_PARSER_MAIN_32_STOPTIMER
  DW NODE_PARSER_MAIN_33_WAITONE
  DW NODE_PARSER_MAIN_34_WAITFOR
  DW NODE_PARSER_MAIN_35_WAITFORKEY
  DW NODE_PARSER_MAIN_36_WAITBEAMLOW
  DW NODE_PARSER_MAIN_37_???
  DW NODE_PARSER_MAIN_38_GETRANDOM
  DW NODE_PARSER_MAIN_39_WAITFIXED
  DW NODE_PARSER_MAIN_3A_WAITONESEC
  DW NODE_PARSER_MAIN_3B_WAITTWOSECS
  DW NODE_PARSER_MAIN_3C_???
  DW NODE_PARSER_MAIN_3D_???
  DW NODE_PARSER_MAIN_3E_???
  DW NODE_PARSER_MAIN_3F_???

  DW NODE_PARSER_MAIN_40_PLAYSOUND
  DW NODE_PARSER_MAIN_41_SETBORDER
  DW NODE_PARSER_MAIN_42_PLAYMEMNOISE
  DW NODE_PARSER_MAIN_43_ADDMEMGRAIN
  DW NODE_PARSER_MAIN_44_PLAYCLICK
  DW NODE_PARSER_MAIN_45_STARTSOUNDSEQ
  DW NODE_PARSER_MAIN_46_STOPSOUNDSEQ
  DW NODE_PARSER_MAIN_47_SETVOLUME
  DW NODE_PARSER_MAIN_48_PLAYSOUNDFOR
  DW NODE_PARSER_MAIN_49_WAITSOUNDSEQ
  DW NODE_PARSER_MAIN_4A_FORCESOUNDSEQ
  DW NODE_PARSER_MAIN_4B_???
  DW NODE_PARSER_MAIN_4C_???
  DW NODE_PARSER_MAIN_4D_???
  DW NODE_PARSER_MAIN_4E_???
  DW NODE_PARSER_MAIN_4F_???

  DW NODE_PARSER_MAIN_50_???
  DW NODE_PARSER_MAIN_51_???
  DW NODE_PARSER_MAIN_52_???
  DW NODE_PARSER_MAIN_53_???
  DW NODE_PARSER_MAIN_54_???
  DW NODE_PARSER_MAIN_55_???
  DW NODE_PARSER_MAIN_56_???
  DW NODE_PARSER_MAIN_57_???
  DW NODE_PARSER_MAIN_58_???
  DW NODE_PARSER_MAIN_59_???
  DW NODE_PARSER_MAIN_5A_???
  DW NODE_PARSER_MAIN_5B_???
  DW NODE_PARSER_MAIN_5C_???
  DW NODE_PARSER_MAIN_5D_???
  DW NODE_PARSER_MAIN_5E_???
  DW NODE_PARSER_MAIN_5F_???

  DW NODE_PARSER_MAIN_60_CREATEFORMNUM
  DW NODE_PARSER_MAIN_61_SETPEOPLE
  DW NODE_PARSER_MAIN_62_SHOWTIME
  DW NODE_PARSER_MAIN_63_SETUPTIME
  DW NODE_PARSER_MAIN_64_CLEARTIME
  DW NODE_PARSER_MAIN_65_WIPEPGPH
  DW NODE_PARSER_MAIN_66_PRINTNEXT
  DW NODE_PARSER_MAIN_67_SETPHONE
  DW NODE_PARSER_MAIN_68_DIALNUMBER
  DW NODE_PARSER_MAIN_69_BEAMV
  DW NODE_PARSER_MAIN_6A_SCROLLPIXV
  DW NODE_PARSER_MAIN_6B_REGENTABLES
  DW NODE_PARSER_MAIN_6C_BENDSOUND
  DW NODE_PARSER_MAIN_6D_SETSCROLLV
  DW NODE_PARSER_MAIN_6E_SHIFTAREAH
  DW NODE_PARSER_MAIN_6F_PLAYRUMBLE

  DW NODE_PARSER_MAIN_70_SCRCOPY
  DW NODE_PARSER_MAIN_71_HLINE
  DW NODE_PARSER_MAIN_72_HLINENEXT
  DW NODE_PARSER_MAIN_73_SETATTRSROW
  DW NODE_PARSER_MAIN_74_PLAYCLICKFOR
  DW NODE_PARSER_MAIN_75_WIPEROWS
  DW NODE_PARSER_MAIN_76_DECBREAKZ
  DW NODE_PARSER_MAIN_77_INCVARLOOPAT
  DW NODE_PARSER_MAIN_78_HLINERESET
  DW NODE_PARSER_MAIN_79_RINGPHONE
  DW NODE_PARSER_MAIN_7A_???
  DW NODE_PARSER_MAIN_7B_???
  DW NODE_PARSER_MAIN_7C_???
  DW NODE_PARSER_MAIN_7D_???
  DW NODE_PARSER_MAIN_7E_???
  DW NODE_PARSER_MAIN_7F_???

  DW NODE_PARSER_MAIN_80_IFVAR_EQ_VAR
  DW NODE_PARSER_MAIN_81_IFVAR_NEQ_VAR
  DW NODE_PARSER_MAIN_82_IFVAR_GT_VAR
  DW NODE_PARSER_MAIN_83_IFVAR_LT_VAR
  DW NODE_PARSER_MAIN_84_IFVAR_GTE_VAR
  DW NODE_PARSER_MAIN_85_IFVAR_LTE_VAR
  DW NODE_PARSER_MAIN_86_???
  DW NODE_PARSER_MAIN_87_???
  DW NODE_PARSER_MAIN_88_IFVAR_EQ_NUM
  DW NODE_PARSER_MAIN_89_IFVAR_NEQ_NUM
  DW NODE_PARSER_MAIN_8A_IFVAR_GT_NUM
  DW NODE_PARSER_MAIN_8B_IFVAR_LT_NUM
  DW NODE_PARSER_MAIN_8C_IFVAR_GTE_NUM
  DW NODE_PARSER_MAIN_8D_IFVAR_LTE_NUM
  DW NODE_PARSER_MAIN_8E_???
  DW NODE_PARSER_MAIN_8F_???

  DW NODE_PARSER_MAIN_90_???
  DW NODE_PARSER_MAIN_91_???
  DW NODE_PARSER_MAIN_92_???
  DW NODE_PARSER_MAIN_93_???
  DW NODE_PARSER_MAIN_94_???
  DW NODE_PARSER_MAIN_95_???
  DW NODE_PARSER_MAIN_96_???
  DW NODE_PARSER_MAIN_97_???
  DW NODE_PARSER_MAIN_98_???
  DW NODE_PARSER_MAIN_99_???
  DW NODE_PARSER_MAIN_9A_???
  DW NODE_PARSER_MAIN_9B_???
  DW NODE_PARSER_MAIN_9C_???
  DW NODE_PARSER_MAIN_9D_???
  DW NODE_PARSER_MAIN_9E_???
  DW NODE_PARSER_MAIN_9F_???

  DW NODE_PARSER_MAIN_A0_ONKEY1
  DW NODE_PARSER_MAIN_A1_ONKEY2
  DW NODE_PARSER_MAIN_A2_ONKEY3
  DW NODE_PARSER_MAIN_A3_ONKEY4
  DW NODE_PARSER_MAIN_A4_ONKEY5
  DW NODE_PARSER_MAIN_A5_???
  DW NODE_PARSER_MAIN_A6_???
  DW NODE_PARSER_MAIN_A7_???
  DW NODE_PARSER_MAIN_A8_???
  DW NODE_PARSER_MAIN_A9_???
  DW NODE_PARSER_MAIN_AA_???
  DW NODE_PARSER_MAIN_AB_???
  DW NODE_PARSER_MAIN_AC_???
  DW NODE_PARSER_MAIN_AD_???
  DW NODE_PARSER_MAIN_AE_???
  DW NODE_PARSER_MAIN_AF_???

  DW NODE_PARSER_MAIN_B0_CHANGENODE
  DW NODE_PARSER_MAIN_B1_CALLSUBNODE
  DW NODE_PARSER_MAIN_B2_BREAKSUB
  DW NODE_PARSER_MAIN_B3_REPEATSUB
  DW NODE_PARSER_MAIN_B4_???
  DW NODE_PARSER_MAIN_B5_???
  DW NODE_PARSER_MAIN_B6_???
  DW NODE_PARSER_MAIN_B7_???
  DW NODE_PARSER_MAIN_B8_???
  DW NODE_PARSER_MAIN_B9_???
  DW NODE_PARSER_MAIN_BA_???
  DW NODE_PARSER_MAIN_BB_???
  DW NODE_PARSER_MAIN_BC_???
  DW NODE_PARSER_MAIN_BD_???
  DW NODE_PARSER_MAIN_BE_???
  DW NODE_PARSER_MAIN_BF_???

  DW NODE_PARSER_MAIN_C0_ENDIF
  DW NODE_PARSER_MAIN_C1_ENDKEY
  DW NODE_PARSER_MAIN_C2_ENDSTATE
  DW NODE_PARSER_MAIN_C3_ENDSUBNODE
  ;DW NODE_PARSER_MAIN_C4_???
  ;DW NODE_PARSER_MAIN_C5_???
  ;DW NODE_PARSER_MAIN_C6_???
  ;DW NODE_PARSER_MAIN_C7_???
  ;DW NODE_PARSER_MAIN_C8_???
  ;DW NODE_PARSER_MAIN_C9_???
  ;DW NODE_PARSER_MAIN_CA_???
  ;DW NODE_PARSER_MAIN_CB_???
  ;DW NODE_PARSER_MAIN_CC_???
  ;DW NODE_PARSER_MAIN_CD_???
  ;DW NODE_PARSER_MAIN_CE_???
  ;DW NODE_PARSER_MAIN_CF_???

  ;DW NODE_PARSER_MAIN_D0_???
  ;DW NODE_PARSER_MAIN_D1_???
  ;DW NODE_PARSER_MAIN_D2_???
  ;DW NODE_PARSER_MAIN_D3_???
  ;DW NODE_PARSER_MAIN_D4_???
  ;DW NODE_PARSER_MAIN_D5_???
  ;DW NODE_PARSER_MAIN_D6_???
  ;DW NODE_PARSER_MAIN_D7_???
  ;DW NODE_PARSER_MAIN_D8_???
  ;DW NODE_PARSER_MAIN_D9_???
  ;DW NODE_PARSER_MAIN_DA_???
  ;DW NODE_PARSER_MAIN_DB_???
  ;DW NODE_PARSER_MAIN_DC_???
  ;DW NODE_PARSER_MAIN_DD_???
  ;DW NODE_PARSER_MAIN_DE_???
  ;DW NODE_PARSER_MAIN_DF_???

  ;DW NODE_PARSER_MAIN_E0_???
  ;DW NODE_PARSER_MAIN_E1_???
  ;DW NODE_PARSER_MAIN_E2_???
  ;DW NODE_PARSER_MAIN_E3_???
  ;DW NODE_PARSER_MAIN_E4_???
  ;DW NODE_PARSER_MAIN_E5_???
  ;DW NODE_PARSER_MAIN_E6_???
  ;DW NODE_PARSER_MAIN_E7_???
  ;DW NODE_PARSER_MAIN_E8_???
  ;DW NODE_PARSER_MAIN_E9_???
  ;DW NODE_PARSER_MAIN_EA_???
  ;DW NODE_PARSER_MAIN_EB_???
  ;DW NODE_PARSER_MAIN_EC_???
  ;DW NODE_PARSER_MAIN_ED_???
  ;DW NODE_PARSER_MAIN_EE_???
  ;DW NODE_PARSER_MAIN_EF_???

  ;DW NODE_PARSER_MAIN_F0_???
  ;DW NODE_PARSER_MAIN_F1_???
  ;DW NODE_PARSER_MAIN_F2_???
  ;DW NODE_PARSER_MAIN_F3_???
  ;DW NODE_PARSER_MAIN_F4_???
  ;DW NODE_PARSER_MAIN_F5_???
  ;DW NODE_PARSER_MAIN_F6_???
  ;DW NODE_PARSER_MAIN_F7_???
  ;DW NODE_PARSER_MAIN_F8_???
  ;DW NODE_PARSER_MAIN_F9_???
  ;DW NODE_PARSER_MAIN_FA_???
  ;DW NODE_PARSER_MAIN_FB_???
  ;DW NODE_PARSER_MAIN_FC_???
  ;DW NODE_PARSER_MAIN_FD_???
  ;DW NODE_PARSER_MAIN_FE_???
  ;DW NODE_PARSER_MAIN_FF_???

NPS_00_COPYXYDIMS:
CALL GET_NEXT_NPBYTE        ; Copy XCellStart...
LD (IX+0), A                ; ...to XSTART.
CALL GET_NEXT_NPBYTE        ; Copy YCellStart...
LD (IX+1), A                ; ...to YSTART.
CALL GET_NEXT_NPBYTE        ; Copy XCellLength...
LD (IX+2), A                ; ...to XLENGTH.
CALL GET_NEXT_NPBYTE        ; Copy YCellLength...
LD (IX+3), A                ; ...to YLENGTH.
RET

; Called with flags set as follows in the C register:
; B3: 0 = variable comparison, 1 = value comparison.
; B2: 0 = variable-comparison, 1 = comparison-variable.
; B1-B0: 00 = TRUE if NZ, 01 = TRUE if Z, 10 = TRUE if NC, 11 = TRUE if C.
EVALUATE_IFVAR:
CALL GET_NEXT_NPBYTE        ; Variable number.
CALL GET_GVAR_FULL          ; DE now contains variable.
PUSH DE                     ; Store variable.
BIT 3, C                    ; Check "compare with variable/value" flag.
JR Z, EVALUATE_IFVAR_JP1
INC HL
LD E, (HL)
INC HL
LD D, (HL)                  ; DE now contains comparison value.
JR EVALUATE_IFVAR_JP2

  EVALUATE_IFVAR_JP1:
  CALL GET_NEXT_NPBYTE      ; Comparison variable number.
  CALL GET_GVAR_FULL        ; DE now contains comparison value.

  EVALUATE_IFVAR_JP2:
  POP HL                    ; HL contains variable, DE comparison value.
  BIT 2, C                  ; Check subtraction order flag.
  JR Z, EVALUATE_IFVAR_JP3
  EX DE, HL                 ; Swap variable and comparator.

  EVALUATE_IFVAR_JP3:
  LD A, 00000011B           ; Mask out all but lower two bits of flags byte.
  AND C
  JR Z, EVALUATE_IFVAR_JP5  ; Code was 00, carry flag is cleared.
  DEC A
  JR Z, EVALUATE_IFVAR_JP6  ; Code was 01, carry flag is cleared.
  DEC A
  JR Z, EVALUATE_IFVAR_JP7  ; Code was 10, carry flag is cleared.
  JR EVALUATE_IFVAR_JP8     ; Code was 11, carry flag is cleared.

  EVALUATE_IFVAR_JP5:       ; Evaluate as TRUE if result non-zero.
  SBC HL, DE
  JR NZ, EVALUATE_IFVAR_JPT
  JR EVALUATE_IFVAR_JPF

  EVALUATE_IFVAR_JP6:       ; Evaluate as TRUE if result zero.
  SBC HL, DE
  JR Z, EVALUATE_IFVAR_JPT
  JR EVALUATE_IFVAR_JPF

  EVALUATE_IFVAR_JP7:       ; Evaluate as TRUE if no carry result.
  SBC HL, DE
  JR NC, EVALUATE_IFVAR_JPT
  JR EVALUATE_IFVAR_JPF

  EVALUATE_IFVAR_JP8:       ; Evaluate as TRUE if carry result.
  SBC HL, DE
  JR C, EVALUATE_IFVAR_JPT
  JR EVALUATE_IFVAR_JPF

  EVALUATE_IFVAR_JPT:
  LD A, 00000001B           ; Increment IFDEPTH, set SKIP to FALSE.
  CALL SET_DEPTH_SKIP
  JP RETURN_TO_PARSER

  EVALUATE_IFVAR_JPF:
  LD A, 00000011B           ; Increment IFDEPTH, set SKIP to TRUE.
  CALL SET_DEPTH_SKIP
  JP RETURN_TO_PARSER

RETURN_TO_PARSER:
LD HL, (V_NODE_POINTER)
JP NODE_PARSER_MAIN

V_NODE_POINTER:
DW 0

; If A is variable reference (B7 set), returns low byte of variable in A.
; - otherwise, A is returned unchanged.
GET_NEXT_NPBYTE:
INC HL
LD A, (HL)
BIT 7, A
CALL NZ, GET_GVAR_LOWBYTE
RET

; If A is variable reference (B7 set), returns full variable in DE.
; - otherwise, returns value of A in DE.
GET_NEXT_NPB_FULLVAR:
INC HL
LD A, (HL)
PUSH HL
CALL GET_IFGVAR_FULLBYTE
POP HL
RET

PUSH_CALLSTACK:
PUSH DE
LD DE, (V_NODE_CALLSTACKP)
LD A, L
LD (DE), A
INC DE
LD A, H
LD (DE), A
INC DE
LD (V_NODE_CALLSTACKP), DE
XOR A
CP E                        ; Check for overflow (stack exceeds 256 bytes.)
CALL Z, PARSER_STACK_OVERFLOW
POP DE
RET

POP_CALLSTACK:
PUSH DE
LD DE, (V_NODE_CALLSTACKP)
DEC DE
LD A, (DE)
LD H, A
DEC DE
LD A, (DE)
LD L, A
LD (V_NODE_CALLSTACKP), DE
POP DE
RET

PUSH_IFKYSTACK:
PUSH DE
LD DE, (V_NODE_IFKYSTACKP)
LD A, (V_NODE_IFDEPTH)
LD (DE), A
INC DE
LD A, (V_NODE_KEYDEPTH)
LD (DE), A
INC DE
LD (V_NODE_IFKYSTACKP), DE
XOR A
CP E                        ; Check for overflow (stack exceeds 256 bytes.)
CALL Z, PARSER_STACK_OVERFLOW
POP DE
RET

POP_IFKYSTACK:
PUSH DE
LD DE, (V_NODE_IFKYSTACKP)
DEC DE
LD A, (DE)
LD (V_NODE_KEYDEPTH), A
DEC DE
LD A, (DE)
LD (V_NODE_IFDEPTH), A
LD (V_NODE_IFKYSTACKP), DE
POP DE
RET

PARSER_STACK_OVERFLOW:
DI
HALT

SET_DEPTH_SKIP:             ; Called with flags in the A register.
LD B, A                     ; Store flags.
LD HL, V_NODE_IFDEPTH
LD DE, V_NODE_IFSKIPSTACK
LD IX, V_NODE_IFSKIP
BIT 2, B                    ; BIT 2: 0=IF, 1=KEY.
JR Z, SET_DEPTH_SKIP_JP1
LD HL, V_NODE_KEYDEPTH
LD DE, V_NODE_KEYSKIPSTACK
LD IX, V_NODE_KEYSKIP

  SET_DEPTH_SKIP_JP1:
  BIT 0, B                  ; BIT 0: 0=DEC DEPTH, 1=INC DEPTH.
  JR Z, SET_DEPTH_SKIP_JP2  ; If zero, decrement DEPTH.
  INC (HL)                  ; Otherwise, increment DEPTH.
  LD A, (HL)                ; Get new DEPTH.
  BIT 7, B                  ; Check it has not exceed maximum of 127.
  JR Z, SET_DEPTH_SKIP_JP6
  DI                        ; Crash to a halt in the event of overflow.
  HALT

  SET_DEPTH_SKIP_JP2:
  DEC (HL)                  ; Decrement DEPTH.
  LD A, (HL)                ; Get new DEPTH.
  BIT 7, B                  ; Check it has not rolled back to 255.
  JR Z, SET_DEPTH_SKIP_JP3
  DI                        ; Crash to a halt in the event of overflow.
  HALT

  SET_DEPTH_SKIP_JP3:
  EX DE, HL                 ; HL now contains SKIPSTACK start.
  LD E, A                   ; Store one copy of DEPTH.
  SRL E                     ; Divide DEPTH by eight (to get byte offset.)
  SRL E
  SRL E
  LD D, 0
  ADD HL, DE                ; HL now points to the correct byte in SKIPSTACK.
  LD D, (HL)                ; D now contains correct byte from SKIPSTACK.
  AND 00000111B             ; Mask upper five DEPTH bits to get BIT position.
  INC A                     ; Going to do a zero check before shifting.

  SET_DEPTH_SKIP_JP4:       ; Shift SKIPSTACK byte in D right by A-1.
  DEC A
  JR Z, SET_DEPTH_SKIP_JP5
  SRL D
  JR SET_DEPTH_SKIP_JP4

  SET_DEPTH_SKIP_JP5:       ; Now BIT 0 of D contains the relevant SKIP flag.
  LD A, D
  AND 00000001B             ; Mask other bits.
  LD (IX+0), A              ; Set the current SKIP variable.
  RET

  SET_DEPTH_SKIP_JP6:
  EX DE, HL                 ; HL now contains SKIPSTACK start.
  LD E, A                   ; Store two copies of DEPTH:
  LD C, A                   ; - one to preserve, one for modification.
  LD A, B                   ; Operate on flags routine was called with.
  AND 00000010B             ; Mask all but BIT 1: 0=DON'T SKIP, 1=SKIP
  SRL A                     ; A now contains SKIP value in BIT 0.
  SRL E                     ; Divide DEPTH by eight (to get byte offset.)
  SRL E
  SRL E
  LD D, 0
  ADD HL, DE                ; HL now points to the correct byte in SKIPSTACK.
  LD (IX+0), A              ; Store SKIP variable.
  LD D, 11111110B           ; Bitmask pattern: skip flag bit low, others high.
  LD B, A                   ; Store flag (with upper bits set.)
  LD A, C                   ; Restore original DEPTH.
  AND 00000111B             ; Mask upper five DEPTH bits to get BIT position.
  INC A                     ; Going to do a zero check before shifting.

  SET_DEPTH_SKIP_JP7:       ; Shift SKIP flag in B, and bitmask, left by A-1.
  DEC A
  JR Z, SET_DEPTH_SKIP_JP8
  RLC B                     ; Shift SKIP flag.
  RLC D                     ; Shift bitmask.
  JR SET_DEPTH_SKIP_JP7

  SET_DEPTH_SKIP_JP8:
  LD A, (HL)                ; Get current byte from SKIPSTACK.
  AND D                     ; Use shifted bitmask to reset the destination bit.
  OR B                      ; Set the destination bit.
  LD (HL), A                ; Store the revised SKIPSTACK byte.
  RET
