; Called with BankNum in the A register, bank offset in the E register.
SET_LINEBANK:
LD HL, I_TXLINES
LD D, 0                      ; DE now contains bank offset.
SLA E                        ; Each entry is two bytes long (no overflow risk.)
ADD HL, DE                   ; HL now at start of line index copy destination.
PUSH HL                      ; Store start of line index copy destination.
LD HL, I_TXBANKS
CALL INDEX_TO_ADDRESS        ; HL now at start of bank, BC contains byte length.
POP DE                       ; Restore start of line index copy destination.
LDIR                         ; Copy block from bank index to line index.
RET

V_SLBANK_OFFSET:
DB 64

; Called with string number in the A register.
; VT_PS_DRAWROW values are all set by parser.
PRINT_STRING_ASLINE:
LD C, 32
CALL PS_SETBUFFLEN
CALL PS_SETSTRING
CALL PS_CLEARBUFF
CALL PS_FILLBUFF
EX DE, HL                       ; Swap buffer pointer to HL.
LD DE, MB_PS_NEXTROW
XOR A                           ; Clear carry flag.
SBC HL, DE                      ; HL now contains used buffer length.
LD A, L                         ; Only need low byte for PS_DRAWROW.
CALL PS_DRAWROW
RET

; Called with string number in the A register.
PRINT_STRING_ASPGPH:
LD (V_PSAP_LASTSTR), A
LD IX, VT_PS_ASPGPH
LD C, (IX+2)                    ; XLENGTH.
CALL PS_SETBUFFLEN
PUSH AF                         ; Store string number.
CALL PS_ADDROWGAP
CALL PS_SETDRAWROW              ; Also sets IY to VT_PS_DRAWROW.
POP AF                          ; Restore string number.
CALL PS_SETSTRING
BIT 0, (IX+4)                   ; LFMODE.
JR Z, PS_ASPGPH_FILL

PS_ASPGPH_FEED:
LD A, (IX+9)                    ; YEND.
LD (IX+7), A                    ; YTRACK.
LD (IY+1), A                    ; YSTART (VT_PS_DRAWROW).

 PS_ASPGPH_FEED_JP1:
 LD A, 1                        ; For PS_SCROLLUP (scroll up one row.)
 CALL PS_SCROLLUP
 CALL PS_CLEARBUFF
 CALL PS_FILLBUFF
 CALL PS_DRAWROW
 LD A, B
 OR C                           ; Reach end of string, and word buffer empty?
 JR NZ, PS_ASPGPH_FEED_JP1
 JR PS_ASPGPH_DELAY             ; Done.

PS_ASPGPH_FILL:
LD A, (IX+7)                    ; YTRACK.
INC A                           ; Allow for FILLDOWN incrementing YTRACK.
LD (IX+8), A                    ; YPGPHSTART.
LD (IY+1), A                    ; YSTART (VT_PS_DRAWROW).

 PS_ASPGPH_FILL_JP1:
 CALL PS_CLEARBUFF
 CALL PS_FILLBUFF
 CALL PS_FILLDOWN
 CALL PS_PLAYSOUND
 CALL PS_DRAWROW
 LD A, B
 OR C                           ; Reach end of string, and word buffer empty?
 JR NZ, PS_ASPGPH_FILL_JP1
 JR PS_ASPGPH_DELAY             ; Done.

PS_ASPGPH_DELAY:
LD BC, (V_PSAP_MINWAIT)
LD DE, (V_PSAP_MAXWAIT)
CALL WAITFOR
RET


VT_PS_ASPGPH:
DB 0                         ;+0 XSTART
DB 0                         ;+1 YSTART
DB 0                         ;+2 XLENGTH
DB 0                         ;+3 YLENGTH
DB 0                         ;+4 LFMODE
DB 0                         ;+5 ROWGAP
DB 0                         ;+6 DPWORD
DB 0                         ;+7 YTRACK
DB 0                         ;+8 YPGPHSTART
DB 0                         ;+9 YEND
DB 0                         ;+10 PLAYSOUND
DB 0                         ;+11 SOUNDNUM

V_PSAP_MINWAIT:
DW 0

V_PSAP_MAXWAIT:
DW 0

V_PSAP_LASTSTR:
DB 0

; Changes state of most registers.
PS_PLAYSOUND:
XOR A
CP (IX+10)                      ; PLAYSOUND.
RET Z
LD A, (IX+11)                   ; SOUNDNUM.
LD DE, 0                        ; DurationOverride (false.)
CALL REQUEST_SOUND              ; Pass request to Scheduler.
RET                             ; Done.

; Called with buffer length in the C register.
PS_SETBUFFLEN:
LD B, 0
LD HL, (V_PS_DRAWBUFFER)
ADD HL, BC
LD (V_PS_DRAWBUFF_END), HL
LD HL, V_PS_DRAWBUFF_LEN
LD (HL), C
RET

; Wipes whole paragraph area and resets tracker.
PS_CLEAR_PGPH:
CALL PS_CLEAR_LASTROW           ; Also sets IX to VT_PS_ASPGPH.
LD A, (IX+1)                    ; YSTART.
DEC A
LD (IX+7), A                    ; Reset YTRACK to YSTART-1.
LD A, (IX+3)                    ; YLENGTH.
CALL PS_SCROLLUP
RET

; Wipes whole paragraph area, without scrolling, and resets tracker.
PS_WIPE_PGPH
LD DE, VT_FILL_AREA
LD HL, VT_PS_ASPGPH
PUSH HL
POP IX
LD A, (IX+1)                    ; YSTART.
DEC A
LD (IX+7), A                    ; Reset YTRACK to YSTART-1.
LD BC, 4
LDIR                            ; Copy four XY parameters to VT_FILL_AREA.
XOR A
LD (DE), A                      ; Set VT_FILL_AREA +4 FILLVALUE to zero.
CALL FILL_AREA
RET

; Must be called with VT_PS_ASPGPH in IX.
PS_SETDRAWROW:
LD IY, VT_PS_DRAWROW
LD A, (IX+0)                    ; Pass XSTART (VT_PS_ASPGPH)...
LD (IY+0), A                    ; ...to XSTART (VT_PS_DRAWROW.)
LD A, (IX+2)                    ; Pass XLENGTH (VT_PS_ASPGPH)...
LD (IY+2), A                    ; ...to XLENGTH (VT_PS_DRAWROW.)
LD A, (IX+1)                    ; YSTART (VT_PS_ASPGPH.)
ADD A, (IX+3)                   ; YLENGTH (VT_PS_ASPGPH.)
DEC A                           ; A now contains bottom row.
LD (IX+9), A                    ; YEND (VT_PS_ASPGPH.)
RET

; Must be called with VT_PS_ASPGPH in IX.
; Destroys A, B, and C registers.
PS_ADDROWGAP:
BIT 0, (IX+5)                   ; ROWGAP.
RET Z
LD A, (IX+1)                    ; YSTART.
LD B, (IX+4)                    ; LFMODE.
LD C, (IX+7)                    ; YTRACK.
DEC A                           ; YSTART-1.
ADD A, B
CP C                            ; YTRACK == (YSTART-1), and LFMODE == 0?
RET Z                           ; No row gap required.
DEC B                           ; Check LFMODE.
JR NZ, PS_ADDROWGAP_JP1         ; Filldown mode.
LD A, 1                         ; Scroll up by one row.
CALL PS_SCROLLUP
RET                             ; Done.

  PS_ADDROWGAP_JP1:             ; Filldown mode.
  LD A, (IX+9)                  ; YEND.
  CP C                          ; YTRACK.
  JR Z, PS_ADDROWGAP_JP2        ; Already on last row, must scroll instead.
  INC (IX+7)                    ; Otherwise, simply move YTRACK down one row.
  CP (IX+7)                     ; Is YTRACK now on last row?
  JR Z, PS_ADDROWGAP_JP2        ; Reset now, rather than trigger overflow.
  RET

  PS_ADDROWGAP_JP2:
  CALL PS_CLEAR_PGPH
  RET                           ; Done.

; Clears the bottom row of the current paragraph area.
PS_CLEAR_LASTROW:
LD IX, VT_PS_ASPGPH
LD IY, VT_FILL_AREA             ; Pass values required to clear prompt.
LD A, (IX+0)                    ; Copy XSTART...
LD (IY+0), A                    ; ...to XSTART.
LD A, (IX+9)                    ; YEND.
LD (IY+1), A                    ; Set fill area YSTART.
LD A, (IX+2)                    ; Copy XLENGTH...
LD (IY+2), A                    ; ...to XLENGTH.
LD (IY+3), 1                    ; Set YLENGTH directly (one row.)
LD (IY+4), 00000000B            ; Set FILLVALUE directly (all pixels off.)
PUSH IX                         ; Store VT_PS_ASPGPH.
CALL FILL_AREA
POP IX                          ; Restore VT_PS_ASPGPH.
RET

; Prints continuation prompt, waits for key, then scrolls up print area.
; Must be called with VT_PS_ASPGPH in IX.
PS_OVERFLOW:
CALL PS_OVFL_SHOWPROMPT
CALL WAITFOR_ANYKEY
CALL PS_CLEAR_LASTROW

  PS_OVERFLOW_JP1:              ; Scrolls start of paragraph to top of screen.
  LD A, (IX+8)                  ; YPGPHSTART.
  SUB (IX+1)                    ; YSTART.
  JR Z, PS_OVERFLOW_JP2
  PUSH AF                       ; Store number of rows to scroll up.
  CALL PS_SCROLLUP
  POP BC                        ; Restore number of rows to scroll up (to B.)
  LD A, (IX+7)                  ; YTRACK.
  SUB B                         ; Move YTRACK back accordingly.
  LD (IX+7), A
  RET                           ; Done.

  PS_OVERFLOW_JP2:              ; Paragraph is larger than print area.
  LD A, 1                       ; Move up one row only.
  CALL PS_SCROLLUP
  DEC (IX+7)                    ; Move YTRACK up one row.
  RET

; Must be called with VT_PS_ASPGPH in IX.
PS_OVFL_SHOWPROMPT:
LD A, (IX+6)                    ; DPWORD.
LD HL, I_TXTOKENS               ; Token words are pure ASCII.
CALL INDEX_TO_ADDRESS           ; Returns start pointer in HL and length in BC.
LD (V_PS_DRAWBUFFER), HL        ; Redirect V_PS_DRAWBUFFER to start of word.
LD IY, VT_PS_OVERFLOW
LD (V_PS_DRAWROWVT), IY         ; Redirect PS_DRAWROW to use prompt's variables.
LD (IY+2), C                    ; Set XLENGTH to word length (low byte only.)
LD A, (IX+0)                    ; Copy XSTART across.
LD (IY+0), A
LD A, (IX+9)                    ; YEND.
LD (IY+1), A                    ; Set YSTART to bottom row.
CALL PS_DRAWROW                 ; Draw prompt.
LD HL, MB_PS_NEXTROW            ; Reset V_PS_DRAWBUFFER.
LD (V_PS_DRAWBUFFER), HL
LD HL, VT_PS_DRAWROW            ; Reset V_PS_DRAWROWVT.
LD (V_PS_DRAWROWVT), HL
RET

; N.B. only used to efficiently pass variables to PS_DRAWROW.
VT_PS_OVERFLOW:
DB 0                            ;+0 XSTART
DB 0                            ;+1 YSTART
DB 0                            ;+2 XLENGTH

; Checks for row overflow when printing strings in filldown mode.
; Must be called with VT_PS_ASPGPH in IX.
PS_FILLDOWN:
LD A, (IX+9)                    ; YEND.
CP (IX+7)                       ; Check if YTRACK already on bottom row.
JR Z, PS_FILLDOWN_JP1
INC (IX+7)                      ; Move YTRACK down one line.
CP (IX+7)                       ; Check if YTRACK now on bottom row.
JR NZ, PS_FILLDOWN_JP2

  PS_FILLDOWN_JP1:
  LD A, B                       ; Remaining word length.
  OR C                          ; Remaining string length.
  PUSH BC                       ; Store remaining word and string lengths.
  PUSH HL                       ; Store string pointer.
  CALL NZ, PS_OVERFLOW          ; More lines to follow if either was not zero.
  POP HL                        ; Restore string pointer.
  POP BC                        ; Restore remaining word and string lengths.

  PS_FILLDOWN_JP2:
  LD A, (IX+7)                  ; YTRACK.
  PUSH IX
  LD IX, VT_PS_DRAWROW
  LD (IX+1), A                  ; Pass updated YTRACK to YSTART for row drawing.
  POP IX
  RET

V_PS_DRAWBUFFER:
DW MB_PS_NEXTROW

V_PS_DRAWBUFF_END:
DW MB_PS_NEXTWORD

V_PS_DRAWBUFF_LEN:
DB 0

MB_PS_NEXTROW:
DB "................................"   ; 32 bytes (screen width.)

MB_PS_NEXTWORD:
DB "................................"   ; 32 bytes (screen width.)

MS_PS_LASTWCHR:
DB " "                                  ; 1 byte (single character.)

; Called with string number in A.
; Returns start pointer in HL, string length in C.
; Resets B to zero.
PS_SETSTRING:
LD HL, I_TXLINES
CALL INDEX_TO_ADDRESS           ; Returns start pointer in HL and length in BC.
LD B, 0                         ; Reset B.
RET

; Fills MB_PS_NEXTROW with spaces.
PS_CLEARBUFF:
PUSH HL                         ; Preserve registers.
PUSH BC
LD C, 32                        ; Length of buffer.
LD B, C                         ; Also ASCII code for space.
LD HL, MB_PS_NEXTROW

 PS_CLEARBUFF_JP1:
 LD (HL), B                     ; Write space.
 INC HL                         ; Next byte.
 DEC C
 JR NZ, PS_CLEARBUFF_JP1        ; Repeat.
 POP BC                         ; Restore registers.
 POP HL
 RET                            ; Done.

; Called with string pointer in HL, remaining word & string lengths in B & C.
; Returns modified string pointer and remaining word & string lengths.
; Returns buffer pointer offset from PS_NEXTROW in DE.
PS_FILLBUFF:
LD DE, MB_PS_NEXTROW            ; Set destination pointer to start of buffer.
XOR A                           ; Zero A.
CP B                            ; Check if any previous word to be copied.
JR NZ, PL_FILLBUFF_JP2          ; Skip getting a new word.

  PL_FILLBUFF_JP1:
  PUSH DE                       ; Preserve row buffer pointer.
  CALL PS_GETWORD               ; Render next word to buffer.
  POP DE                        ; Restore row buffer pointer.
  XOR A                         ; Zero A.
  CP B                          ; Check length of word.
  RET Z                         ; Zero-length word indicates end of string.

  PL_FILLBUFF_JP2:
  PUSH HL                       ; Preserve source pointer.
  LD HL, (V_PS_DRAWBUFF_END)
  AND A                         ; Reset carry flag.
  SBC HL, DE                    ; HL now contains remaining buffer length.
  LD A, L                       ; Only low byte required.
  POP HL                        ; Restore source pointer.
  CP B                          ; Check length of word against remaining buffer.
  JR C, PL_FILLBUFF_JP4         ; Word is larger than remaining buffer.

  PL_FILLBUFF_JP3:              ; Word fits in remaining buffer.
  CALL PS_WORDTOROW             ; Copy entire word.
  JR PL_FILLBUFF_JP1            ; Get next word.

  PL_FILLBUFF_JP4:              ; Word is larger than remaining buffer.
  INC A                         ; Check if word was one character too long.
  CP B
  JR NZ, PL_FILLBUFF_JP5        ; Word is more that one character too long.
  LD A, (MS_PS_LASTWCHR)        ; Word is exactly one character too long.
  CP 32                         ; Check if the overflow character is a space.
  JR NZ, PL_FILLBUFF_JP5        ; Not space character, so cannot truncate.
  DEC B                         ; Reduce word length to lose space character.
  CALL PS_WORDTOROW             ; Copy word excluding final character (space.)
  JR PL_FILLBUFF_JP1            ; Get next word, causing return via full buffer.

  ; TO CONSIDER: allow words longer than whole buffer to begin on a part-line.
  PL_FILLBUFF_JP5:              ; Word too long for remaining buffer.
  DEC A                         ; Compensate for INC at start of JP4.
  PUSH HL                       ; Store pointer.
  LD HL, MB_PS_NEXTROW
  AND A                         ; Reset carry flag.
  SBC HL, DE                    ; Check if at start of buffer.
  POP HL                        ; Restore pointer.
  JR Z, PL_FILLBUFF_JP6         ; Word is longer than entire buffer.
  RET                           ; Done.

  PL_FILLBUFF_JP6:              ; Word too long for entire buffer.
  PUSH BC                       ; Store remaining word and string lengths.
  LD B, 0
  LD C, A                       ; BC now contains remaining buffer length.
  PUSH HL                       ; Store string pointer.
  LD HL, MB_PS_NEXTWORD
  LD DE, MB_PS_NEXTROW
  PUSH BC                       ; Store remaining buffer length.
  LDIR                          ; Copy part of word to fill remaining buffer.
  POP BC                        ; Restore remaining buffer length.
  LD DE, MB_PS_NEXTWORD
  LDIR                          ; Shift remains of word to start of buffer.
  POP HL                        ; Restore string pointer.
  POP BC                        ; Restore remaining word and string lengths.
  LD E, A                       ; Remaining buffer length.
  LD A, B
  SUB E                         ; Reduce remaining word length accordingly.
  LD B, A
  LD DE, (V_PS_DRAWBUFF_END)    ; Restore buffer pointer (one space after end.)
  RET                           ; Done.

; Copies length B from start of word buffer to DE in row buffer.
; Returns modified DE, and unmodified BC and HL.
PS_WORDTOROW:
PUSH HL
LD HL, MB_PS_NEXTWORD
PUSH BC
LD C, B
LD B, 0
LDIR
POP BC
POP HL
RET

; Called with source pointer in HL, remaining string length in C.
; Returns modified pointer and remaining string length, and word length in B.
; WARNING: no buffer overflow test (but >32 character words are unlikely.)
PS_GETWORD:
XOR A                           ; Zero A.
LD (MS_PS_LASTWCHR), A          ; Clear reference copy of last character.
LD DE, MB_PS_NEXTWORD           ; Set destination pointer.
INC C                           ; Allow for remaining length test before copy.

  PS_GETWORD_JP1:
  DEC C                         ; Remaining string length.
  JR Z, PS_GETWORD_JP7          ; Reached end of string.
  LD A, (HL)                    ; Get next string byte.
  INC HL                        ; Move source pointer to next string byte.
  CP 0
  JR Z, PS_GETWORD_JP6          ; Global variable reference escape code.
  BIT 7, A
  JR NZ, PS_GETWORD_JP5         ; Dictionary reference.
  CP 32
  JR Z, PS_GETWORD_JP4          ; Space character.
  JR C, PS_GETWORD_JP3          ; Lower case character followed by space.

  PS_GETWORD_JP2:               ; Character is non-control/space ASCII code.
  LD (DE), A                    ; Copy directly to buffer.
  INC DE                        ; Move buffer pointer.
  JR PS_GETWORD_JP1             ; Next character.

  PS_GETWORD_JP3:               ; Character is lower case followed by space.
  ADD A, 01100000B              ; Offset to get lower case character.
  LD (DE), A                    ; Copy directly to buffer.
  INC DE                        ; Move buffer pointer.
  LD A, 32                      ; Set next character to a Space.
  JR PS_GETWORD_JP4             ; Handle next character per normal Space.

  PS_GETWORD_JP4:               ; Space character.
  LD (DE), A                    ; Copy directly to buffer.
  INC DE                        ; Move pointer (for word length calculation.)
  DEC C                         ; Reduce remaining length (as not jumping back.)
  JR PS_GETWORD_JP7             ; It's the end of the word as we know it.

  PS_GETWORD_JP5:               ; Dictionary reference.
  PUSH HL                       ; Store string source pointer.
  PUSH BC                       ; Store remaining string length.
  LD HL, I_TXDICTWD             ; Dictionary words must be pure ASCII.
  CALL INDEX_TO_ADDRESS         ; Returns start pointer in HL and length in BC.
  LDIR                          ; This copy also increments buffer pointer DE.
  POP BC                        ; Restore remaining string length.
  POP HL                        ; Restore string source pointer.
  JR PS_GETWORD_JP1             ; Next character.

  PS_GETWORD_JP6:               ; Global variable reference escape code.
  LD A, (HL)                    ; Get the global variable reference number.
  CALL GET_GVAR_LOWBYTE         ; Turn it into a token reference.
  PUSH HL                       ; Continue per dictionary reference...
  PUSH BC
  LD HL, I_TXTOKENS             ; ...except using a different index table.
  CALL INDEX_TO_ADDRESS
  LDIR
  POP BC
  POP HL
  INC HL                        ; Move source pointer to next string byte.
  DEC C                         ; - reference was two bytes long.
  JR Z, PS_GETWORD_JP7          ; Reached end of string.
  JR PS_GETWORD_JP1             ; Next character.

  PS_GETWORD_JP7:               ; Word complete, now calculate length etc.
  DEC DE                        ; Move back to last character.
  LD A, (DE)                    ; Make a reference copy of it.
  INC DE                        ; Restore buffer pointer (for length calc.)
  LD (MS_PS_LASTWCHR), A
  PUSH HL                       ; Preserve source pointer.
  EX DE, HL                     ; Swap buffer pointer to HL.
  LD DE, MB_PS_NEXTWORD
  AND A                         ; Reset carry flag.
  SBC HL, DE                    ; HL now contains word length.
  LD B, L                       ; Move low byte of word length to B (max 255.)
  POP HL                        ; Restore source pointer.
  RET

; Renders row of text to screen.
; Uses value from A register if XLENGTH is zero.
; Uses shadow registers and preserves IX.
PS_DRAWROW:
EXX                          ; Store registers.
PUSH IX
LD IX, (V_PS_DRAWROWVT)
LD HL, D_TEXT_CHAR           ; Switch to text character tilebank.
LD (V_TILE_DRAW_BANK_START), HL
LD HL, (V_PS_DRAWBUFFER)
LD B, (IX+0)                 ; XSTART.
LD C, (IX+1)                 ; YSTART.
LD D, (IX+2)                 ; XLENGTH.
INC D                        ; Check if zero.
DEC D
JR NZ, PL_DRAWROW_JP1
LD D, A                      ; XLENGTH was zero, so use value from A instead.

  PL_DRAWROW_JP1:
  LD A, 32                   ; Calculate maximum available width.
  SUB B                      ; Result = 32-XSTART.
  CP D                       ; Check against XLENGTH.
  JR NC, PL_DRAWROW_JP2      ; No greater than maximum available, continue.
  LD D, A                    ; Otherwise, set XLENGTH to maximum available.

  PL_DRAWROW_JP2:
  LD A, (HL)                 ; Get next character from buffer.
  PUSH BC                    ; Store registers.
  PUSH DE
  PUSH HL
  CALL TILE_DRAW
  POP HL                     ; Restore registers.
  POP DE
  POP BC
  INC HL                     ; Next character.
  INC B                      ; Next column.
  DEC D                      ; Row complete?
  JR NZ, PL_DRAWROW_JP2
  EXX                        ; Restore registers.
  POP IX
  RET

V_PS_DRAWROWVT:
DW VT_PS_DRAWROW

; Set by parser for ASLINE, set directly by ASPGPH.
VT_PS_DRAWROW:
DB 0                         ;+0 XSTART
DB 0                         ;+1 YSTART
DB 0                         ;+2 XLENGTH

; Set by SETPRINTFIXED.
VT_PS_DRAWFIXED:
DB 0                         ;+0 XSTART
DB 0                         ;+1 YSTART
DB 0                         ;+2 XLENGTH

; Scroll print area, as defined in table IX, up by A rows.
; The A register is corrupted.
PS_SCROLLUP:
EXX                          ; Store calling registers.
PUSH AF                      ; Store pass counter.
LD B, (IX+1)                 ; YSTART.
LD C, (IX+3)                 ; YLENGTH.
INC B                        ; Start one row below top...
DEC C                        ; ...so (YLENGTH-1) rows are copied up each pass.
JR Z, PS_SCROLLUP_JP4        ; Nothing to scroll if YLENGTH == 1.
PUSH BC                      ; Store initialised row tracker and row countdown.

  PS_SCROLLUP_JP1:           ; First pass (bottom row is only cleared once.)
  CALL PS_SCROLLUP_ONEPASS
  CALL PS_CLEAR_LASTROW
  JR PS_SCROLLUP_JP3

  PS_SCROLLUP_JP2:           ; Subsequent passes.
  CALL PS_SCROLLUP_ONEPASS

  PS_SCROLLUP_JP3:           ; Continuation point after first pass.
  POP BC                     ; Restore initialised row tracker and row countdown.
  POP AF                     ; Restore pass counter.
  DEC A                      ; All passes done?
  JR Z, PS_SCROLLUP_JP4
  PUSH AF                    ; Store pass counter.
  PUSH BC                    ; Store initialised row tracker and EndRow+1.
  JR PS_SCROLLUP_JP2

  PS_SCROLLUP_JP4:
  EXX                        ; Restore calling registers.
  RET

; Called with YSTART+1 in B, YLENGTH-1 in C, print area table in IX.
PS_SCROLLUP_ONEPASS:
PUSH BC                      ; Store current row and countdown.
LD A, B                      ; A now contains row number to be copied up.
CALL PS_ONEROWUP
POP BC                       ; Restore current row and countdown.
INC B                        ; Next row.
DEC C                        ; Reduce countdown.
JR NZ, PS_SCROLLUP_ONEPASS   ; Repeat until zero.
RET

; Copies row number A within print area, as defined in table IX, up by one row.
; All but index registers are corrupted.
PS_ONEROWUP:
LD B, (IX+0)                  ; XSTART.
LD C, A                       ; Source row.
PUSH BC                       ; Store XY.
CALL GET_SCR_CELLSTART
EX DE, HL                     ; HL now at start of source.
POP BC                        ; Restore XY.
DEC C                         ; Move to previous row.
CALL GET_SCR_CELLSTART        ; DE now at start of destination.
LD BC, 0                      ; Zero BC.
LD A, 8                       ; Each row has eight lines of pixels.

  PS_ONEROWUP_JP1:
  PUSH DE                     ; Store source and destination.
  PUSH HL
  LD C, (IX+2)                ; Set BC from zero to XLENGTH.
  LDIR                        ; Copy one line of pixels.
  POP HL                      ; Restore source and destination.
  POP DE
  INC D                       ; Move both pointers down one row.
  INC H
  DEC A                       ; All lines done?
  JR NZ, PS_ONEROWUP_JP1      ; Copy next line of pixels.
  RET
