SCHEDULER:
EI                           ; Interrupt call disables interrupts, so re-enable.
CALL PRESERVE_ALLREGS        ; Store main and shadow registers.
                             ; The following happen regardless every interrupt.

; N.B "TIMER" removed to save memory.
;CALL TIMER
LD HL, VT_TIMER_FRAMES       ; Use single byte frame counter instead.
INC (HL)


CALL KEYBOARD_READ
CALL SOUNDSEQ                ; Subthread operations go from here until JP1.
CALL SOUNDQUEUE

CALL RESTORE_ALLREGS         ; Restore main and shadow registers.
RET                          ; Return to the main thread (Node Parser.)


SOUNDQUEUE:
LD A, (V_SOUNDQUEUE_FULL)    ; Any sounds in the queue?
DEC A                        ; Also clears queue value, if it was 1.
RET NZ
LD (V_SOUNDQUEUE_FULL), A    ; Store cleared queue value.
LD A, (V_SOUNDQUEUE_NEXT)    ; Play the next sound.
LD DE, (V_SOUNDQUEUE_DURO)
DI
CALL PLAY_SOUND
EI
RET                          ; Done.

V_SOUNDQUEUE_FULL:
DB 0

V_SOUNDQUEUE_NEXT:
DB 0

V_SOUNDQUEUE_DURO:
DW 0

SOUNDSEQ:
LD IX, VT_SOUNDSEQ
XOR A
CP (IX+0)                    ; Break if not ACTIVE.
RET Z
LD A, (V_PLAY_SOUND_BUSY)    ; Break if PLAY_SOUND is currently BUSY.
AND A
RET NZ
LD B, (IX+1)                 ; Get DELAY.
DEC (IX+1)                   ; Decrement regardless.
CP B
RET NZ                       ; No sound yet if previous DELAY value not zero.
LD HL, (V_SOUNDSEQ_PTR)      ; Get then follow next instruction.
LD A, (HL)                   ; Next instruction byte.
INC HL                       ; Move pointer on.
CALL GET_IFGVAR_LOWBYTE
LD (V_SOUNDSEQ_PTR), HL      ; Store revised pointer.
CP 64                        ; Sound number?
JR NC, SOUNDSEQ_JP1
LD DE, 0                     ; Set sound duration to default.
CALL REQUEST_SOUND           ; Queue playback request.
LD A, (HL)                   ; Next instruction byte is delay until next sound.
INC HL                       ; Move pointer on.
LD (IX+1), A                 ; Pass to DELAY.
LD (V_SOUNDSEQ_PTR), HL      ; Store revised pointer.
RET                          ; Done.

  SOUNDSEQ_JP1:
  CP 127                     ; Check for "end of sequence" code.
  JR NZ, SOUNDSEQ_JP3
  XOR A
  LD (IX+0), A               ; Reset ACTIVE.
  LD B, (IX+3)               ; RPTCOUNT.
  CP B                       ; Check for zero (indefinite repeat.)
  JP Z, SOUNDSEQ_JP2         ; Leave as zero, repeat sequence.
  DEC (IX+3)                 ; RPTCOUNT.
  CP (IX+3)                  ; Check if zero after decrement.
  RET Z                      ; Done, end of sequence with no more repeats.

  SOUNDSEQ_JP2:
  LD A, (IX+2)               ; SEQNUM.
  CALL RESET_SOUNDSEQ        ; Retrigger sequence.
  RET                        ; Done.

  SOUNDSEQ_JP3:
  ; Used to draw bitmaps in rotary dialing sequence.
  CP 64                      ; Check for "draw again" code.
  JR NZ, SOUNDSEQ_BADCODE    ; Only two codes (may expand with lookup table.)
  LD A, (HL)                 ; Next instruction byte is bitmap number.
  INC HL                     ; Move pointer on.
  LD (V_SOUNDSEQ_PTR), HL    ; Store revised pointer.
  LD (IX+1), 0               ; Reset DELAY.
  LD BC, (V_LASTDRAWPOS)     ; Draw bitmap.
  CALL DRAW_BGSTREAM
  RET                        ; Done.

  SOUNDSEQ_BADCODE:          ; Crashes to halt.
  DI
  HALT

VT_SOUNDSEQ:
DB 0                         ;+0 ACTIVE
DB 0                         ;+1 DELAY
DB 0                         ;+2 SEQNUM
DB 0                         ;+3 RPTCOUNT

V_SOUNDSEQ_PTR:
DW 0                         ; Pointer to next byte of sequence data.

V_SOUNDSEQ_KEYBRK:
DB 1                         ; Flag that enables break by any keypress.

KEYBOARD_READ:
LD HL, V_KEYBOARD_KEYS
LD BC, 63486                 ; B4, B3, B2, B1, B0 = "5", "4", "3", "2", "1".
LD E, 0                      ; Used by keyhold counter (for RNG.)
IN D, (C)                    ; N.B. bits are set if key is NOT pressed.
LD B, 5                      ; Number of keys to check.

  KEYBOARD_READ_JP1:
  LD C, 0                    ; Reset the key variable to be stored.
  SRL D                      ; Check key bit.
  JR NC, KEYBOARD_READ_JP2   ; 0 = key pressed, 1 = key not pressed.
  JR KEYBOARD_READ_JP4       ; If not pressed, set the key variable to zero.

  KEYBOARD_READ_JP2:         ; Key was pressed.
  INC E                      ; Increment keyhold counter.
  LD A, (HL)
  CP C                       ; Check if key variable is zero.
  JR NZ, KEYBOARD_READ_JP3
  INC C                      ; If key variable is zero, set to 1.
  JR KEYBOARD_READ_JP4

  KEYBOARD_READ_JP3:         ; Key was pressed and variable is not zero.
  INC C                      ; C now contains 1, and A contains key variable.
  CP C
  JR Z, KEYBOARD_READ_JP4    ; If key variable was 1, keep it as 1.
  DEC A                      ; Otherwise, reduce key variable by one.
  LD C, A

  KEYBOARD_READ_JP4:
  LD (HL), C                 ; Store the new key value.
  INC HL
  DEC B
  JR NZ, KEYBOARD_READ_JP1   ; Check the next bit.

  LD A, E                    ; Move keyhold result.
  PUSH HL                    ; Store V_KEYBOARD_ANY address.
  LD HL, V_RNG_KEYHOLDCOUNT  ; All keys checked, add E to total keyhold count.
  ADD A, (HL)                ; Possible rollover is intentional for RNG usage.
  LD (HL), A
  POP HL                     ; Restore V_KEYBOARD_ANY address.
  LD A, 1                    ; Indicates key press not restricted by delay.
  LD (HL), A                 ; Set V_KEYBOARD_ANY as True.
  LD B, 5                    ; Five keys to check.

  KEYBOARD_READ_JP5:
  DEC HL
  CP (HL)
  RET Z                      ; V_KEYBOARD_ANY stays True if any key value is 1.
  DEC B
  JR NZ, KEYBOARD_READ_JP5
  XOR A                      ; Set A to zero (if no key values were 1.)
  LD (V_KEYBOARD_KEY_ANY), A ; Set V_KEYBOARD_ANY to false.
  RET

V_KEYBOARD_KEYS:
V_KBKEY_1:
DB 0

V_KBKEY_2:
DB 0

V_KBKEY_3:
DB 0

V_KBKEY_4:
DB 0

V_KBKEY_5:
DB 0

V_KEYBOARD_KEY_ANY:
DB 0

; N.B. "TIMER" REMOVED TO SAVE MEMORY.

; This will increment a timing counter.
; There are four bytes to the timer: frames, seconds, minutes, and hours.
; The run time will vary slightly.

; Update the hours/minutes/seconds/frames counter first.
;TIMER:
;LD HL, VT_TIMER_CLOCK
;INC (HL)
;LD A, (HL)
;CP 50            ; Frames per second.
;JR NZ, TIMER_JP1
;LD (HL), 0         ; Reset Frames.
;INC HL
;INC (HL)
;LD A, (HL)
;CP 60            ; Seconds per minute.
;JR NZ, TIMER_JP1
;LD (HL), 0         ; Reset Seconds.
;INC HL
;INC (HL)
;LD A, (HL)
;CP 60            ; Minutes per hour.
;JR NZ, TIMER_JP1
;LD (HL), 0         ; Reset Minutes.
;INC HL
;INC (HL)
;LD A, (HL)
;CP 24            ; Hours per day (timer resets every 24 hours.)
;JR NZ, TIMER_JP1
;LD (HL), 0         ; Reset Hours.
;
;  ; Update the pure frame counter.
;  TIMER_JP1:
;  LD HL, VT_TIMER_FRAMES
;  INC (HL)      ; Lowest byte.
;  RET Z
;  INC HL
;  INC (HL)      ; Next byte.
;  RET Z
;  INC HL
;  INC (HL)      ; Next byte.
;  RET Z
;  INC HL
;  INC (HL)      ; Highest byte.
;  RET           ; All values will be rolled over back to zero by this point.
;
;VT_TIMER_CLOCK:
;V_TIMER_CLOCK_FRAMES:
;DB 00H
;V_TIMER_CLOCK_SECONDS:
;DB 00H
;V_TIMER_CLOCK_MINUTES:
;DB 00H
;V_TIMER_CLOCK_HOURS:
;DB 00H

VT_TIMER_FRAMES:
DB 00H                       ; 32 bit frame counter - lowest byte.
DB 00H                       ;
DB 00H                       ;
DB 00H                       ; 32 bit frame counter - highest byte.
