V_SND_BORDER:
DB 0

V_SND_VOLMASK:
DB 00010000B

; Multi-channel sound generator.
; To consider later: revise to use 8 bit delay counters.
; (Would still allow 85Hz to 4,100Hz range with current delays.)
SOUND_GENMCH:
LD BC, (V_SGMCH_CYCLES)      ; Set cycle counter.
XOR A                        ; Set both channel outputs to LOW.
LD HL, (V_SGMCH_DLY_L1)      ; Channel 1 delay counter.
LD E, A                      ; Channel 1 output.
EXX
LD HL, (V_SGMCH_DLY_L2)      ; 'Channel 2 delay counter.
LD C, A                      ; 'Channel 2 output.
EXX                          ; Done filling shadow registers.

  ; Execution time (this block only) - 104 Ts.
  SOUND_GENMCH_JP1:
  CALL SGMCH_DELAY           ; 17 Ts Wait until either channel state changes.
  LD A, E                    ; 4 Ts Channel 1.
  EXX                        ; 4 Ts
  OR C                       ; '4 Ts Channel 2.
  EXX                        ; 4 Ts
  LD D, A                    ; 4 Ts Apply volume mask.
  LD A, (V_SND_VOLMASK)      ; 13 Ts
  AND D                      ; 4 Ts
  LD D, A                    ; 4 Ts Apply border colour.
  LD A, (V_SND_BORDER)       ; 13 Ts
  OR D                       ; 4 Ts
  OUT (254), A               ; 11 Ts
  LD A, B                    ; 4 Ts Cycles complete?
  OR C                       ; 4 Ts
  JP NZ, SOUND_GENMCH_JP1    ; 10 Ts

  SOUND_GENMCH_JP2:
  LD A, (V_SND_BORDER)       ; 13 Ts Ensure speaker output is LOW on exit.
  OUT (254), A               ; 11 Ts
  RET                        ; 10 Ts Done.

; Execution times:
; Neither counter reached zero - 81 Ts.
; Channel 1 reached zero       - 136/158 (L/H) Ts.
; Channel 2 reached zero       - 144/160 (L/H) Ts.
; Channels 1 & 2 reached zero  - 227 Max. (HH) Ts.
;
; NB total overhead per channel per H-L-H cycle is approximately 345 Ts/
SGMCH_DELAY:
LD D, 0                      ; 7 Ts Reset channel flip counter.
DEC HL                       ; 6 Ts Channel 1 counter.
LD A, H                      ; 4 Ts
OR L                         ; 4 Ts
JP NZ, SGMCH_DELAY_JP1       ; 10 Ts
INC D                        ; 4 Ts Channel flip counter.
LD HL, (V_SGMCH_DLY_L1)      ; 16 Ts LOW or HIGH reset chosen by output value.
LD A, E                      ; 4 Ts Flip channel output.
XOR 00010000B                ; 7 Ts
LD E, A                      ; 4 Ts
JP Z, SGMCH_DELAY_JP1        ; 10 Ts
DEC BC                       ; 6 Ts Only decrease cycle count on flip to HIGH.
LD HL, (V_SGMCH_DLY_H1)      ; 16 Ts

  SGMCH_DELAY_JP1:
  EXX                        ; 4 Ts
  DEC HL                     ; '6 Ts Channel 2 counter.
  LD A, H                    ; '4 Ts
  OR L                       ; '4 Ts
  JP NZ, SGMCH_DELAY_JP2     ; '10 Ts
  EXX                        ; '4 Ts
  INC D                      ; 4 Ts Channel flip counter.
  EXX                        ; '4 Ts
  LD HL, (V_SGMCH_DLY_L2)    ; '16 Ts Per Channel 1 counter reset.
  LD A, C                    ; '4 Ts Flip channel output.
  XOR 00010000B              ; '7 Ts
  LD C, A                    ; '4 Ts
  JP Z, SGMCH_DELAY_JP2      ; '10 Ts
  LD HL, (V_SGMCH_DLY_H2)    ; '16 Ts

  SGMCH_DELAY_JP2:
  EXX                        ; '4 Ts

  SGMCH_DELAY_JP3:
  XOR A                      ; 4 Ts
  CP D                       ; 4 Ts Channel flip counter.
  JP Z, SGMCH_DELAY          ; 10 Ts No speaker change yet.
  RET                        ; 10 Ts One or more channels flipped.

; Test values below are for 350Hz and 440Hz, both at 80% duty cycle.
V_SGMCH_CYCLES:
DW 256

V_SGMCH_DLY_H1:
DW 86

V_SGMCH_DLY_L1:
DW 29

V_SGMCH_DLY_H2:
DW 68

V_SGMCH_DLY_L2:
DW 22


SOUND_GENERATOR:
; Frequency & Phase adjusted outside this loop, at 50hz maximum:
LD A, (V_SND_VOLMASK)
LD B, A                      ; B stores HIGH speaker value, so mute if A is 0.
LD A, (V_SND_BORDER)         ; Get border colour.
  DB 17                      ; LD DE, nn
  SG_CYCLES:
  DB 0, 192
LD L, A                      ; LOW output value.
OR B                         ; Set speaker bit to volume bit.
LD H, A                      ; HIGH output value.
CALL SG_OUTLOOP
RET

SG_OUTLOOP:
; LOW phase.
  LD A, L                    ; 4 Ts (LOW output value.)
    DB 01                    ; 10 Ts - LD BC, nn
    SG_HIGHP:                ; (HIGH peroid, poked.)
    DB 3, 64                 ; Delay counter - high byte first.
  OUT (254), A               ; 11 Ts
  DEC DE                     ; 6 Ts (loop counter.)
  LD A, D                    ; 4 Ts
  OR E                       ; 4 Ts
  RET Z                      ; 5 Ts (if not breaking from loop.)
                             ; = 44 Ts for this block.

  SG_OUTLOOP_JP1:
  DJNZ SG_OUTLOOP_JP1        ; 8+(13*(B-1)) Ts
  DEC C                      ; 4 Ts
  JP NZ, SG_OUTLOOP_JP1      ; 10 Ts
                             ; = 22 Ts minimum for this block.
                             ; Minimum high phase = 66 Ts (52.9khz.)

; HIGH phase.
  LD A, H                    ; 4 Ts (HIGH output value.)
    DB 01                    ; 10 Ts - LD BC, nn
    SG_LOWP:                 ; (LOW period, poked.)
    DB 2, 192                ; Delay counter - high byte first.
  OUT (254), A               ; 11 Ts
                             ; = 25 Ts for this block.

  SG_OUTLOOP_JP2:
  DJNZ SG_OUTLOOP_JP2        ; 8+(13*(B-1)) Ts
  DEC C                      ; 4 Ts
  JP NZ, SG_OUTLOOP_JP2      ; 10 Ts
                             ; = 22 Ts minimum for this block.

  LD A, (HL)                 ; 7 Ts (balancing execution time.)
  NOP                        ; 4 Ts (balancing execution time.)
  JP SG_OUTLOOP              ; 8 Ts (jump back to HIGH phase.)
                             ; = 19 Ts for this block.
                             ; Minimum low phase = 66 Ts (52.9kHz.)

                             ; Frequency with zero delays (H-L-H) = 26.5kHz.
                             ; Frequency with 8 total delays = 14.8kHz.
                             ; Frequency with 16 total delays = 10.3kHz.
                             ; Frequency with 32 total delays = 6.4kHz.
                             ; - Adequate phase granularity at most frequencies.

; Called with sound number in the A register, duration override in DE.
REQUEST_SOUND:
LD (V_SOUNDQUEUE_NEXT), A
LD (V_SOUNDQUEUE_DURO), DE
LD A, 1
LD (V_SOUNDQUEUE_FULL), A
RET

; Called with sound number in the A register, duration override in DE.
PLAY_SOUND:
PUSH AF                     ; Set BUSY.
LD A, 1
LD (V_PLAY_SOUND_BUSY), A
POP AF
CP 32                       ; Is the sound a single tone?
JR NC, PLAY_SOUND_MCH

  PLAY_SOUND_SCH:           ; For single channel sounds.
  LD HL, I_SOUNDS_SCH
  CALL GET_ADDRESS_UNALIGNED
  LD BC, SG_CYCLES
  CALL PS_SET_DUR
  LD BC, SG_HIGHP
  CALL PS_SET_VALUE
  LD BC, SG_LOWP
  CALL PS_SET_VALUE
  CALL SOUND_GENERATOR
  XOR A                     ; Reset BUSY.
  LD (V_PLAY_SOUND_BUSY), A
  RET

  PLAY_SOUND_MCH:           ; For multi channel sounds.
  AND 00011111B             ; Clear upper three bits of sound number.
  LD HL, I_SOUNDS_MCH
  CALL GET_ADDRESS_UNALIGNED
  LD BC, V_SGMCH_CYCLES
  CALL PS_SET_DUR
  CALL PS_SET_VALUE         ; Sets V_SGMCH_DLY_H1.
  CALL PS_SET_VALUE         ; Sets V_SGMCH_DLY_L1.
  CALL PS_SET_VALUE         ; Sets V_SGMCH_DLY_H2.
  CALL PS_SET_VALUE         ; Sets V_SGMCH_DLY_L2.
  CALL SOUND_GENMCH
  XOR A                     ; Reset BUSY.
  LD (V_PLAY_SOUND_BUSY), A
  RET

V_PLAY_SOUND_BUSY:
DB 0

PS_SET_VALUE:
LD A, (HL)
LD (BC), A
INC HL
INC BC
LD A, (HL)
LD (BC), A
INC HL
INC BC
RET

PS_SET_DUR:
LD A, D                     ; DE, if not zero, is the duration override value.
OR E
JR NZ, PS_SET_DUR_JP1
CALL PS_SET_VALUE           ; Use default duration.
RET

  PS_SET_DUR_JP1:           ; Use duration in DE.
  LD A, E
  LD (BC), A
  INC HL
  INC BC
  LD A, D
  LD (BC), A
  INC HL
  INC BC
  RET


; Called with sequence number in the A register.
SET_SOUNDSEQ:
LD HL, I_SEQUENCES             ; Get start address of sequence data.
CALL GET_ADDRESS_UNALIGNED
LD (V_SET_SOUNDSEQ_PTR), HL    ; Store start address.
CALL RESET_SOUNDSEQ
RET                            ; Done.

V_SET_SOUNDSEQ_PTR:
DW 0

RESET_SOUNDSEQ:
LD IX, VT_SOUNDSEQ
LD (IX+0), 1                   ; Set ACTIVE.
LD (IX+1), 0                   ; Set DELAY to 0.
LD HL, (V_SET_SOUNDSEQ_PTR)    ; Get start address of sequence data.
LD (V_SOUNDSEQ_PTR), HL        ; Pass pointer.
LD HL, 0                       ; Reset BROKESEQ flag (in GVAR 0.)
LD (V_GVAR), HL
RET                            ; Done.

; Called with start address in BC, length in DE.
PLAY_MEMNOISE:
PUSH BC
POP HL
LD A, (V_SND_VOLMASK)
LD C, A                        ; C stores HIGH speaker value, so mute if A is 0.
LD A, (V_SND_BORDER)
SRL D                          ; Divide DE by two.
SRL E

  PLAY_MEMNOISE_JP1:           ; High period.
  OR C                         ; Set speaker bit to volume bit.
  OUT (254), A                 ; Set speaker level.
  LD B, (HL)                   ; Get byte from memory.

  PLAY_MEMNOISE_JP2:           ; Delay loop.
  DEC B
  JR NZ, PLAY_MEMNOISE_JP2

  PLAY_MEMNOISE_JP3:           ; Low period.
  AND 11101111B                ; Reset speaker bit.
  OUT (254), A                 ; Set speaker level.
  INC HL
  LD B, (HL)                   ; Get byte from memory.
  INC HL                       ; N.B. doing commands here to help balance...
  DEC DE                       ; ...High and Low period overheads.

  PLAY_MEMNOISE_JP4:           ; Delay loop.
  DEC B
  JR NZ, PLAY_MEMNOISE_JP4

  LD B, A                      ; Preserve A register.
  LD A, E                      ; End of length?
  OR D
  RET Z
  LD A, B                      ; Restore A register.
  JR PLAY_MEMNOISE_JP1

; Called with click length in B (longer value gives stronger click.)
PLAY_CLICK:
LD A, (V_SND_VOLMASK)
CP 00010000B
RET NZ
LD A, (V_SND_BORDER)           ; Get border colour (speaker bit will be zero.)
XOR 00010000B                  ; Flip speaker bit (to one.)
OUT (254), A

  PLAY_CLICK_JP1:
  DJNZ PLAY_CLICK_JP1
  XOR 00010000B                ; Flip speaker bit (back to zero.)
  OUT (254), A
  RET                          ; Done.

; Copies Sound A to position zero (for modification.)
; BC, DE, and HL registers will be affected if sound can be copied.
MODCOPY_SOUND:
CP 32                          ; Check that sound is single-channel.
RET NC
LD HL, I_SOUNDS_SCH
CALL INDEX_TO_ADDRESS          ; Returns Sound A address in HL, length in BC.
LD DE, (I_SOUNDS_SCH)          ; DE now points to Sound 0.
LDIR                           ; Copy data from Sound A to Sound 0.
RET                            ; Done.


