import os.path, sys
from PIL import Image

class gV:
    FileExtension = ".png"
    OutputName = "BITMAPS"
    MaxImages = 127 # For testing only, usually 127
    MaxWidthP = 256
    MaxHeightP = 192

    # CodeRefs will be stored in the range MaxShortRefs to MaxCodeRef.
    MaxShortRefs = 64
    MaxCodeRef = 95 # N.B. limit of 255.

    FileNumber = 0
    ImageCount = 0

    BlackRGB = (  0,   0,   0)
    WhiteRGB = (255, 255, 255)
    KeyRGB   = (255,   0, 255)  # Magenta.

    CellDim = 8
    BaseOffset = 0
    MasterSize = 0

    MapBits = 2
    ImgHeaderSize = 4

class gD:
    Imgs = []
    ImgIndex = ""
    ImgData = bytearray()
    ImgCells = []

class cSpecImage:
    def __init__(self):
        self.Image = None
        self.Number = None
        self.Width = None
        self.Height = None
        self.Alpha = None
        self.MapSize = None
        
        self.AlpChn = bytearray()
        self.MonChn = bytearray()
        self.DatOut = bytearray()

        self.Cells = []
        self.CellRefs = []
        self.CellSteps = []
        self.Map = []

    def CreateFrom(self, File, Number = None):
        filein = Image.open(File)        
        # Check dimensions.
        if (filein.width % 8 > 0) or (filein.height % 8 > 0):
            print("Error at Image", Number)
            print("Dimensions", filein.size, "are not both multiples of 8.")
            sys.exit(1)
        if (filein.width > gV.MaxWidthP) or (filein.height > gV.MaxHeightP):
            print("Error at Image", Number)
            print("Dimensions", filein.size, "exceed maximum width and/or height.")
            sys.exit(1)        
        self.Width = filein.width
        self.Height = filein.height
        self.Number = Number
        # Store as list of (r,g,b) values.
        rgbimg = filein.convert()
        self.Image = list(rgbimg.getdata())
        filein.close()        
        # Determine if alpha channel required.
        if gV.KeyRGB in self.Image:
            self.Alpha = True
        else:
            self.Alpha = False
        return

    def PackBits(self):
        AlphaByte = 0
        MonoByte = 0
        BitCount = 0        
        # Pack every eight pixels into one byte.
        for pixel in self.Image:
            # Shift current bits, setting last bit to zero.
            AlphaByte += AlphaByte
            MonoByte += MonoByte
            # Evaluate current pixel.
            if pixel == gV.KeyRGB:
                AlphaByte +=1
                MonoByte +=0
            elif pixel == gV.WhiteRGB:
                AlphaByte +=0
                MonoByte +=0
            elif pixel == gV.BlackRGB:
                AlphaByte +=0
                MonoByte +=1
            else:
                print("Invalid pixel colour:", pixel)
                sys.exit(1)
            # Check if byte is full.
            BitCount +=1
            if BitCount == 8:
                # Store then reset.
                self.AlpChn.extend(AlphaByte.to_bytes(1, byteorder="big"))
                AlphaByte = 0
                self.MonChn.extend(MonoByte.to_bytes(1, byteorder="big"))
                MonoByte = 0
                BitCount = 0
        return

    def ToCellRefs(self, CellDict):
        for cell in self.Cells:
            for entry in CellDict:                
                if cell == entry[0]: # Cell contents.
                    self.CellRefs.append(entry[2]) # Cell reference code.

    def ToCells(self):    
        width = int(self.Width / gV.CellDim)
        height = int(self.Height / gV.CellDim)
        offset = 0
        step = 0
        for row in range(height):
            # Alpha channel row, if present.
            if self.Alpha:
                for column in range(width):
                    offset = column + (row * width * gV.CellDim)
                    self.ToCell(offset, width, self.AlpChn, step)
                    step +=1
            # Mono channel row.
            for column in range(width):
                offset = column + (row * width * gV.CellDim)
                self.ToCell(offset, width, self.MonChn, step)
                step +=1
        return

    def ToCell(self, Offset, Width, Data, Step):
        cell = bytearray()
        for byte in range (gV.CellDim):
            cell.append(Data[Offset])
            Offset += Width
        # Add corresponding cell code to map.
        code = GetMapCode(cell)
        self.Map.append(code)
        if code == 0:
            # Store non-blank cell and current step.
            self.Cells.append(cell)
            self.CellSteps.append(Step)
        return

    def ReviseMap(self, CellDict):
        for c in range(len(self.CellRefs)):
            ref = self.CellRefs[c]
            if CellDict[ref][1] > 1: # Cell frequency.
                # Set map code to indicate pointer (repeated cell.)
                self.Map[self.CellSteps[c]] = 1
            else:
                # Set map code to indicate data (unique cell.)
                self.Map[self.CellSteps[c]] = 0
        return

    # Debugging in progress...
    def ToRunlens(self):
        LastRef = None
        ReptStep = None
        Repeating = False
        ReptRef = None
        RefStep = 0
        NewRefs = []
        NewSteps = []
        
        for ref in self.CellRefs:
            # Check that cells are sequential and identical.
            if (ref == LastRef) and (self.CellSteps[RefStep] - self.CellSteps[RefStep-1] == 1):                    
                    if ReptRef == gV.MaxCodeRef:
                        # Store run code then reset to avoid overflow.
                        NewRefs.append(ReptRef)
                        NewSteps.append(ReptStep)
                        ReptRef = gV.MaxShortRefs                        
                        # Set temporary code signifying RLE, and set step to store RLE result in later.
                        self.Map[self.CellSteps[RefStep]] = 4                        
                        ReptStep = self.CellSteps[RefStep]
                    else:
                        if Repeating:
                            ReptRef += 1                        
                            # Map code not required for this cell.
                            self.Map[self.CellSteps[RefStep]] = None
                        else:
                            ReptRef = gV.MaxShortRefs
                            Repeating = True
                            # Set temporary code signifying RLE, and set step to store RLE result in later.
                            self.Map[self.CellSteps[RefStep]] = 4                            
                            ReptStep = self.CellSteps[RefStep]
            else:
                if Repeating:                    
                    # Store run code.
                    NewRefs.append(ReptRef)
                    NewSteps.append(ReptStep)
                    Repeating = False
                # Store current cell reference.
                NewRefs.append(ref)
                NewSteps.append(self.CellSteps[RefStep])                    
            # Update LastRef and RefStep regardless before next pass.
            LastRef = ref
            RefStep +=1

        # In progress:
        # Store run code if run is unbroken at the end of the image.
        if Repeating:
            NewRefs.append(ReptRef)
            NewSteps.append(ReptStep)            

        # Store revised CellRefs and CellSteps.
        self.CellRefs = NewRefs
        self.CellSteps = NewSteps
        
        return

    # Appears to work.
    def SetMapSize(self):
        Length = 0        
        for entry in self.Map:
            if entry is None:
                pass
            else:
                Length +=1                
        Size = Length / (gV.CellDim / gV.MapBits)
        # Adjust for part-filled final byte.
        if Size.is_integer():
            pass
        else:            
            Size +=1            
        self.MapSize = int(Size)
        return

    # Appears to work, but output not fully verified.
    def CellsToBytes(self, GlobalOrg, GlobalData, AllCells):
        LocalData = bytearray()
        LocalOffset = GlobalOrg + len(GlobalData) + gV.ImgHeaderSize + self.MapSize
        Step = 0

        print("Starting exhaustive search, Global Data length:", len(GlobalData))
        
        for cellref in self.CellRefs:                
            if self.Map[self.CellSteps[Step]] == 0: # Code for data.                
                # Get corresponding byte array.
                celldata = gD.ImgCells[cellref][0]
                # Search global data.
                pointer = GlobalOrg            
                result = SearchFor(celldata, GlobalData)
                if not result[0]:
                    # Search local data.
                    pointer = LocalOffset
                    result = SearchFor(celldata, LocalData)
                if result[0]:
                    # Match found (store pointer.)
                    self.Map[self.CellSteps[Step]] = 1 # Code for pointer reference.
                    pointer += result[1] # Relative offset.
                    # N.B. big-endian order used, not native Z80 little-endian.
                    # - The decompressor needs to check the high byte first.
                    # - The decompressor then converts to little-endian, if required.
                    LocalData.extend(pointer.to_bytes(2, byteorder="big"))                    
                else:
                    # No match, so inline the cell data.
                    LocalData.extend(celldata)
                    
            elif self.Map[self.CellSteps[Step]] == 1: # Pointer reference.
                if cellref < gV.MaxShortRefs:                        
                    # Add single-byte reference code to bytesteam.
                    LocalData.extend(cellref.to_bytes(1, byteorder="little"))
                else:
                    # New: In progress (insert data and set pointer if None.)
                    pointer = AllCells[cellref][3]
                    if pointer is None:
                        # Calculate and store pointer.
                        pointer = LocalOffset + len(LocalData)
                        AllCells[cellref][3] = pointer                        
                        # Inline cell data.
                        LocalData.extend(AllCells[cellref][0])                        
                        # Change Map code from pointer to Data.
                        self.Map[self.CellSteps[Step]] = 0                                 
                    else:
                        # N.B. big-endian order, for same reasons noted earlier.
                        LocalData.extend(pointer.to_bytes(2, byteorder="big"))
                    
            elif self.Map[self.CellSteps[Step]] == 4: # Temporary RLE code.
                # Reset to catch-all index/RLE/pointer code.
                self.Map[self.CellSteps[Step]] = 1
                # Add single-byte RLE code to bytesteam.
                LocalData.extend(cellref.to_bytes(1, byteorder="little"))
                
                    
            # Required for correct self.Map indexing.
            Step +=1
        
        return LocalData, AllCells

    # Appears to work, but not verified.
    def MapToBytes(self, Offset):        
        MapBlock = bytearray()
        CellWidth = int(self.Width / 8)
        CellHeight = int(self.Height / 8)
        PackLimit = int(8 / gV.MapBits)

        widthflag = CellWidth
        channels = 1
        if self.Alpha:
            channels +=1
            # Set Bit 7 of width.
            widthflag += 128
        # Store these two header bytes first.
        MapBlock.append(widthflag)
        MapBlock.append(CellHeight)

        # Allow for header bytes.
        Offset += gV.ImgHeaderSize
        # Calculate start of the data stream.
        Offset += self.MapSize
        # Store remaining two header bytes.
        MapBlock.extend(Offset.to_bytes(2, byteorder="little"))

        # Create MapBlock from map entries, including padding to end of rows.
        MapByte = 0
        PackStep = 0
        for entry in self.Map:
            if entry is None:
                # Ignore empty entries.
                pass
            else:
                # Pack code into upper two bits of the byte.
                MapByte += (entry * 64)
                PackStep +=1
                if PackStep == PackLimit:
                    # Store full byte then reset.
                    MapBlock.append(MapByte)
                    PackStep = 0
                    MapByte = 0
                else:
                    # Shift bits two places right.
                    MapByte = int(MapByte / 4)
                    
        if PackStep != 0:
            for shift in range(PackLimit-(PackStep+1)):
                # Pad part-filled final byte with zeros.
                MapByte = int(MapByte / 4)
            # Store padded byte.
            MapBlock.append(MapByte)
            
        return MapBlock

# Appears to work, but not verified.
def SearchFor(TargetData, SearchData):
    match = False
    position = 0
    winlen = len(TargetData)
    searchlen = len(SearchData)

    if winlen > searchlen:
        return (match, position)
    for offset in range(searchlen-winlen):
        if TargetData == SearchData[offset:(offset+winlen)]:
            match = True
            position = offset
            break
    return(match, position)

def GetMapCode(Cell):
    Sum = sum(Cell)
    if Sum == 0:
        # White (8*0.)
        Code = 2
    elif Sum == 2040:
        # Black (8*255.)
        Code = 3
    else:
        # Data.
        Code = 0    
    return Code

# Tested and working.
def CreateIndex(SpecImages):
    offset = gV.BaseOffset + gV.MasterSize
    for step in range(gV.MaxImages):
        entry = False
        for image in SpecImages:
            if image.Number == step:
                entry = True
                break
        if entry:
            # Write image entry.
            gD.ImgIndex += ("  DW " + str(offset) + "\n")
            offset += len(image.DatOut)
        else:
            # Write null entry.
            gD.ImgIndex += ("  DW " + str(offset) + "\n")        
    return  

# Has passed initial debugging.
def CompileCells(Cells):
    CompCells = []
    
    # Cast mutable bytearray()s to immutable bytes()s for use by set().
    ByteCells = []
    for cell in Cells:
        ByteCells.append(bytes(cell))
    
    CellTypes = set(ByteCells)
    for cell in CellTypes:
        frequency = Cells.count(cell)
        CompCells.append([cell, frequency])
        
    # Sort by descending frequency.
    #  Explanation of: key=lambda i: (i[1], i[0])
    #  - The leftmost element (i[1]) sorts by frequency.
    #  - The second element (i[0]) breaks frequency ties by cell data.
    #  If the tiebreaker was not specified, then the sub-order would be arbitrary.
    #  - This would cause slight variations in later exhaustive search results.
    CompCells = sorted(CompCells, key=lambda i: (i[1], i[0]), reverse=True)
    
    # Add reference codes corresponding to frequency rank.
    # Then add pointer field, set to "None" as currently unassigned.
    refcode = 0
    for cell in CompCells:
        cell.append(refcode)
        cell.append(None)
        refcode +=1    
    return CompCells

# Has passed initial debugging.
def RemoveRuns(Cells):
    NoRunCells = []
    LastCell = None

    for cell in Cells:
        if cell == LastCell:
            pass
        else:
            NoRunCells.append(cell)
        LastCell = cell

    return NoRunCells

# Main entry.
SpecDataStart = input("Please enter the image data hex start address:")
SpecDataStart = int(SpecDataStart, 16)

if SpecDataStart < 0 or SpecDataStart > 65535:
    print("Error: data start address", SpecDataStart, "is outside memory range.")
    sys.exit(1)

gV.BaseOffset = SpecDataStart

# Load images.
for FileNumber in range(gV.MaxImages):
    FileNumString = str(FileNumber).zfill(3) + gV.FileExtension
    if os.path.isfile(FileNumString):
        image = cSpecImage()
        image.CreateFrom(FileNumString, FileNumber)
        gD.Imgs.append(image)
        gV.ImageCount +=1        
    else:
        print("Null entry for image", FileNumber)

# Build unoptimised image maps and cell lists.
for image in gD.Imgs:
    # Build image map and cell list.
    image.PackBits()
    image.ToCells()
    # Add image cells to master cell list.
    gD.ImgCells.extend(image.Cells)
print("Raw size of unoptimised non-blank cells:", (len(gD.ImgCells)*8))

# Reduce all runs of image cells to a single cell.
# This allows production of an optimised ShortRef list.
gD.ImgCells = RemoveRuns(gD.ImgCells)

# Condense the master cell list, then create bytearray of ShortRef cells.
gD.ImgCells = CompileCells(gD.ImgCells)
for cell in gD.ImgCells:
    if cell[2] == gV.MaxShortRefs: # Cell rank.
        break
    else:
        gD.ImgData.extend(cell[0]) # Cell bytes.
# Image data will then follow the master cell list.
gV.MasterSize = len(gD.ImgData)

# Build each SpecImage's list of references to the master cell list.
for image in gD.Imgs:
    image.ToCellRefs(gD.ImgCells)

# Set non-blank image map codes to most efficient choice from pointer or inline data.
for image in gD.Imgs:
    image.ReviseMap(gD.ImgCells)

# Run length encoding pass for each image.
for image in gD.Imgs:
    image.ToRunlens()

# Calculate final image map sizes.
for image in gD.Imgs:
    image.SetMapSize()

# Convert images to bytestreams via final search for pointers.
for image in gD.Imgs:
    # N.B. ImgData grows each pass for use by exhaustive search.
    result = image.CellsToBytes(gV.BaseOffset, gD.ImgData, gD.ImgCells)
    celldata = result[0] 
    gD.ImgCells = result[1] # Pointers may have been modified.
    offset = gV.BaseOffset +len(gD.ImgData)
    mapdata = image.MapToBytes(offset)
    image.DatOut.extend(mapdata)
    image.DatOut.extend(celldata)
    gD.ImgData.extend(image.DatOut)

# Display the stats, check enough space, then write the data.
print("\nFound", gV.ImageCount, "image files.")
rawsize = 0
for image in gD.Imgs:
    size = image.Height * (image.Width / 8)
    if image.Alpha:
        size += size
    rawsize += size
print("Uncompressed size:", int(rawsize), "bytes.")
print("Generated", len(gD.ImgData), "bytes of data.")

if (SpecDataStart+len(gD.ImgData) > 65535):
    print("Error: data too large for space following start address.")
    sys.exit(1)

# Create index.
CreateIndex(gD.Imgs)

# Store data.
IndexOut = open("BITMAPS-I.asm", "w")
IndexOut.write(gD.ImgIndex)
IndexOut.close()

DataOut = open("BITMAPS.bin", "wb")
DataOut.write(gD.ImgData)
DataOut.close()

print("Index and Data files written.")
sys.exit(0)
