import os.path, sys

DefaultNodesFile = "NODES.txt"

# 0. Line parser function.
#    - takes one string and a variable-length list of integers.

def ParseVariables(Vars, Types, Maxs):
    # Check correct number of variables was supplied.
    if len(Vars) != len(Types):
        print("Error on line", LineNum, ":")
        print(len(Types), "variables required, but", len(Vars), "supplied.")
        sys.exit(1)

    # Check variables are within allowed ranges, if so add to byte list.
    VarBytes = []
    VarNum = 0
    for Var in Vars:
        # Check for GVAR reference.
        if Var[0]=="v":
            if Types[VarNum] == "b":
                # Remove v. Instead, increase by 128 to indicate GVAR reference.
                Var = int(Var[1:]) + 128
                VarBytes.extend((Var).to_bytes(1, byteorder='little'))
            else:        
                print("Parser error at instruction on line", LineNum, ":")
                print("GVAR reference must be of data type byte.")
                sys.exit(1)
        # Check for high linebank reference.
        elif Var[0]=="h":
            if Types[VarNum] == "b":
                # Remove b. Instead, offset by 64 to reference high linebank.
                Var = int(Var[1:]) + 64
                VarBytes.extend((Var).to_bytes(1, byteorder='little'))
            else:        
                print("Parser error at instruction on line", LineNum, ":")
                print("High linebank reference must be of data type byte.")
                sys.exit(1)
        else:
            # Check range.
            if int(Var) > Maxs[VarNum]:
                print("Error on line", LineNum, ":")
                print("Variable", VarNum, "exceeds value limit.")
                sys.exit(1)
            # Check type.
            if Types[VarNum] == "b":
                VarBytes.extend(int(Var).to_bytes(1, byteorder='little'))
            elif Types[VarNum] == "w":
                VarBytes.extend(int(Var).to_bytes(2, byteorder='little'))
            else:        
                print("Parser error at instruction on line", LineNum, ":")
                print("Unrecognised variable type:", Types[VarNum])
                sys.exit(1)
            
        VarNum+=1
        
    return VarBytes

def ParseLine(Line):
    global ByteLength
    global OnEntryIndex
    global HasOnEntryNode
    global OnCurrentIndex
    global HasOnCurrentNode
    global OnExitIndex
    global HasOnExitNode
    global WaitingForEndState
    global WaitingForEndNode
    global WaitingForEndSub
    global SubNodesOnly
    
    Command = Line[0]

    # Bytecode instructions (all in range 0xXX.)
    
    if Command == "PRINTLINE":
        Bytes = [0x00]
        VarTypes = "bbbb"
        VarMax = [127, 31, 23, 32]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "PRINTPGPH":
        Bytes = [0x01]
        VarTypes = "b"
        VarMax = [127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "CLEARPGPH":
        Bytes = [0x02]
        
    elif Command == "SETPGPHAREA":
        Bytes = [0x03]
        VarTypes = "bbbb"
        VarMax = [31, 23, 32, 24]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "SETPGPHSPACE":
        Bytes = [0x04]
        VarTypes = "b"
        VarMax = [1]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "SETPGPHMODE":
        Bytes = [0x05]
        VarTypes = "b"
        VarMax = [1]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "SETPGPHDELAY":
        Bytes = [0x06]
        VarTypes = "bb"
        VarMax = [127, 127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "SETPGPHPROMPT":
        Bytes = [0x07]
        VarTypes = "b"
        VarMax = [127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "SETLINEBANKHI":
        Bytes = [0x08]
        VarTypes = "b"
        VarMax = [127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "SETLINEBANKLO":
        Bytes = [0x09]
        VarTypes = "b"
        VarMax = [127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "SETPGPHSOUND":
        Bytes = [0x0A]
        VarTypes = "bb"
        VarMax = [1, 127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "FILLSCREEN":
        Bytes = [0x0C]
        VarTypes = "b"
        VarMax = [255]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "FILLAREA":
        Bytes = [0x0D]
        VarTypes = "bbbbb"
        VarMax = [31, 23, 32, 24, 255]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "PRINTFIXED":
        Bytes = [0x0E]
        VarTypes = "b"
        VarMax = [127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "SETPRINTFIXED":
        Bytes = [0x0F]
        VarTypes = "bbb"
        VarMax = [31, 23, 32]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "DRAWBITMAP":
        Bytes = [0x10]
        VarTypes = "bbb"
        VarMax = [127, 31, 23]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "PRINTGFXCHAR":
        Bytes = [0x11]
        VarTypes = "bbb"
        VarMax = [127, 31, 23]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "DRAWAGAIN":
        Bytes = [0x12]
        VarTypes = "b"
        VarMax = [127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "SETLASTDRAW":
        Bytes = [0x13]
        VarTypes = "bbb"
        VarMax = [31, 23]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "SETATTRIBUTES":
        Bytes = [0x14]
        VarTypes = "bbbbb"
        VarMax = [31, 23, 32, 24, 255]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "SETALLATTRIBUTES":
        Bytes = [0x15]
        VarTypes = "b"
        VarMax = [255]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "INVATTRIBUTES":
        Bytes = [0x16]
        VarTypes = "bbbb"
        VarMax = [31, 23, 32, 24]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "INVALLATTRIBUTES":
        Bytes = [0x17]
        
    elif Command == "LOADATTRIBUTES":
        Bytes = [0x18]
        VarTypes = "b"
        VarMax = [127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "SETATTRSCREEN":
        Bytes = [0x19]
        
    elif Command == "SETATTRBUFFER":
        Bytes = [0x1A]
        
    elif Command == "COPYATTRBUFFER":
        Bytes = [0x1B]
        
    elif Command == "COPYATTRAREA":
        Bytes = [0x1C]
        VarTypes = "bbbb"
        VarMax = [31, 23, 32, 24]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "SCRATTRTOBUFF":
        Bytes = [0x1D]
        
    elif Command == "FLIPSCRX":
        Bytes = [0x1E]
        VarTypes = "bbbbb"
        VarMax = [31, 23, 32, 24, 31]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "FLIPSCRY":
        Bytes = [0x1F]
        VarTypes = "bbbbb"
        VarMax = [31, 23, 32, 24, 23]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "SETVAR":
        Bytes = [0x20]
        VarTypes = "bb"
        VarMax = [127, 255]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "INCVAR":
        Bytes = [0x21]
        VarTypes = "b"
        VarMax = [255]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "DECVAR":
        Bytes = [0x22]
        VarTypes = "b"
        VarMax = [255]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "COPYVAR":
        Bytes = [0x23]
        VarTypes = "bb"
        VarMax = [255, 255]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "PRINTVAR":
        Bytes = [0x24]
        VarTypes = "bbb"
        VarMax = [255, 255, 255]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "ADDVAR":
        Bytes = [0x25]
        VarTypes = "bb"
        VarMax = [127, 127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "SUBVAR":
        Bytes = [0x26]
        VarTypes = "bb"
        VarMax = [127, 127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "SETVARFULL":
        Bytes = [0x27]
        VarTypes = "bw"
        VarMax = [127, 65535]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "FLIPFLAG":
        Bytes = [0x28]
        VarTypes = "b"
        VarMax = [127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "SETTIMER":
        Bytes = [0x30]
        VarTypes = "bw"
        VarMax = [127, 65535]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "STARTTIMER":
        Bytes = [0x31]
        VarTypes = "b"
        VarMax = [127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "STOPTIMER":
        Bytes = [0x32]
        VarTypes = "b"
        VarMax = [127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "WAITONE":
        Bytes = [0x33]
        
    elif Command == "WAITFOR":
        Bytes = [0x34]
        VarTypes = "bb"
        VarMax = [255, 255]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "WAITFORKEY":
        Bytes = [0x35]
        
    elif Command == "WAITBEAMLOW":
        Bytes = [0x36]
        
    elif Command == "GETRANDOM":
        Bytes = [0x38]
        VarTypes = "b"
        VarMax = [255]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "WAITFIXED":
        Bytes = [0x39]
        VarTypes = "b"
        VarMax = [127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "WAITONESEC":
        Bytes = [0x3A]
        
    elif Command == "WAITTWOSECS":
        Bytes = [0x3B]
        
    elif Command == "PLAYSOUND":
        Bytes = [0x40]
        VarTypes = "b"
        VarMax = [127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "SETBORDER":
        Bytes = [0x41]
        VarTypes = "b"
        VarMax = [255]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "PLAYMEMNOISE":
        Bytes = [0x42]
        VarTypes = "bb"
        VarMax = [127, 127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "ADDMEMGRAIN":
        Bytes = [0x43]
        VarTypes = "bbbbbb"
        VarMax = [23, 31, 24, 32, 127, 255]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "PLAYCLICK":
        Bytes = [0x44]
        
    elif Command == "STARTSOUNDSEQ":
        Bytes = [0x45]
        VarTypes = "bb"
        VarMax = [63, 127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "STOPSOUNDSEQ":
        Bytes = [0x46]
        
    elif Command == "SETVOLUME":
        Bytes = [0x47]
        VarTypes = "b"
        VarMax = [1]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "PLAYSOUNDFOR":
        Bytes = [0x48]
        VarTypes = "bb"
        VarMax = [127, 127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "WAITSOUNDSEQ":
        Bytes = [0x49]
        
    elif Command == "FORCESOUNDSEQ":
        Bytes = [0x4A]
        VarTypes = "b"
        VarMax = [1]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "CREATEFORMNUM":
        Bytes = [0x60]
        VarTypes = "b"
        VarMax = [127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "SETPEOPLE":
        Bytes = [0x61]
        
    elif Command == "SHOWTIME":
        Bytes = [0x62]
        
    elif Command == "SETUPTIME":
        Bytes = [0x63]
        VarTypes = "bbbb"
        VarMax = [31, 23, 11, 10]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "CLEARTIME":
        Bytes = [0x64]
        
    elif Command == "WIPEPGPH":
        Bytes = [0x65]
        
    elif Command == "PRINTNEXT":
        Bytes = [0x66]
        
    elif Command == "SETPHONE":
        Bytes = [0x67]
        VarTypes = "b"
        VarMax = [3]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "DIALNUMBER":
        Bytes = [0x68]
        VarTypes = "b"
        VarMax = [127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "BEAMV":
        Bytes = [0x69]
        VarTypes = "bbbbbbbb"
        VarMax = [31, 23, 32, 24, 23, 1, 127, 127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "SCROLLPIXV":
        Bytes = [0x6A]
        VarTypes = "bbbbbbbb"
        VarMax = [31, 32, 192, 1, 191, 192, 191, 31]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "REGENTABLES":
        Bytes = [0x6B]
        
    elif Command == "BENDSOUND":
        Bytes = [0x6C]
        VarTypes = "bbbb"
        VarMax = [31, 127, 127, 127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "SETSCROLLV":
        Bytes = [0x6D]
        VarTypes = "bbbb"
        VarMax = [127, 1, 127, 1]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "SHIFTAREAH":
        Bytes = [0x6E]
        VarTypes = "bbbbb"
        VarMax = [31, 23, 32, 24, 1]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "PLAYRUMBLE":
        Bytes = [0x6F]
        VarTypes = "bb"
        VarMax = [127, 127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "SCRCOPY":
        Bytes = [0x70]
        VarTypes = "bbbbbb"
        VarMax = [31, 23, 32, 24, 31, 23]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "HLINE":
        Bytes = [0x71]
        VarTypes = "bbb"
        VarMax = [127, 127, 127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "HLINENEXT":
        Bytes = [0x72]
        
    elif Command == "CRASHROW":
        Bytes = [0x73]
        VarTypes = "bb"
        VarMax = [31, 127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "PLAYCLICKFOR":
        Bytes = [0x74]
        VarTypes = "b"
        VarMax = [127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "WIPEROWS":
        Bytes = [0x75]
        VarTypes = "bbb"
        VarMax = [127, 127, 127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "DECBREAKZ":
        Bytes = [0x76]
        VarTypes = "b"
        VarMax = [127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "INCVARLOOPAT":
        Bytes = [0x77]
        VarTypes = "bb"
        VarMax = [127, 127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "HLINERESET":
        Bytes = [0x78]
        
    elif Command == "RINGPHONE":
        Bytes = [0x79]
        VarTypes = "bb"
        VarMax = [127, 127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "IFVAR==(VAR)":
        Bytes = [0x80]
        VarTypes = "bb"
        VarMax = [127, 127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "IFVAR!=(VAR)":
        Bytes = [0x81]
        VarTypes = "bb"
        VarMax = [127, 127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "IFVAR>(VAR)":
        Bytes = [0x82]
        VarTypes = "bb"
        VarMax = [127, 127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "IFVAR<(VAR)":
        Bytes = [0x83]
        VarTypes = "bb"
        VarMax = [127, 127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "IFVAR>=(VAR)":
        Bytes = [0x84]
        VarTypes = "bb"
        VarMax = [127, 127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "IFVAR<=(VAR)":
        Bytes = [0x85]
        VarTypes = "bb"
        VarMax = [127, 127]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "IFVAR==(NUM)":
        Bytes = [0x88]
        VarTypes = "bw"
        VarMax = [127, 65535]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "IFVAR!=(NUM)":
        Bytes = [0x89]
        VarTypes = "bw"
        VarMax = [127, 65535]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "IFVAR>(NUM)":
        Bytes = [0x8A]
        VarTypes = "bw"
        VarMax = [127, 65535]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "IFVAR<(NUM)":
        Bytes = [0x8B]
        VarTypes = "bw"
        VarMax = [127, 65535]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "IFVAR>=(NUM)":
        Bytes = [0x8C]
        VarTypes = "bw"
        VarMax = [127, 65535]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "IFVAR<=(NUM)":
        Bytes = [0x8D]
        VarTypes = "bw"
        VarMax = [127, 65535]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "ONKEY1":
        Bytes = [0xA0]
        
    elif Command == "ONKEY2":
        Bytes = [0xA1]
        
    elif Command == "ONKEY3":
        Bytes = [0xA2]
        
    elif Command == "ONKEY4":
        Bytes = [0xA3]
        
    elif Command == "ONKEY5":
        Bytes = [0xA4]
        
    elif Command == "CHANGENODE":
        Bytes = [0xB0]
        VarTypes = "b"
        VarMax = [255]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        if WaitingForEndSub:
            print("Reminder regarding line", LineNum, ":\n"
                  "  CHANGENODE in Subnode will cause restart from specified Node.")
        
    elif Command == "CALLSUBNODE":
        Bytes = [0xB1]
        VarTypes = "b"
        VarMax = [255]
        Bytes.extend(ParseVariables(Line[1:], VarTypes, VarMax))
        
    elif Command == "BREAKSUB":
        Bytes = [0xB2]
        if not WaitingForEndSub:
            print("Error on line", LineNum,
                  ": BREAKSUB can only be called from Subnode.")
            sys.exit(1)            
        
    elif Command == "REPEATSUB":
        Bytes = [0xB3]
        if not WaitingForEndSub:
            print("Error on line", LineNum,
                  ": REPEATSUB can only be called from Subnode.")
            sys.exit(1) 
        
    elif Command == "ENDIF":
        Bytes = [0xC0]
        
    elif Command == "ENDKEY":
        Bytes = [0xC1]
        
    elif Command == "ENDSTATE":
        Bytes = [0xC2]
        if not WaitingForEndState:
            print("Error on line", LineNum,
                  ": ENDSTATE before ONSTATE.")
            sys.exit(1)
        else:
            WaitingForEndState = False
        
    elif Command == "ENDSUBNODE":
        Bytes = [0xC3]
        if not WaitingForEndSub:
            print("Error on line", LineNum,
                  ": ENDSUBNODE before SUBNODE.")
            sys.exit(1)
        else:
            WaitingForEndSub = False
            # Need to update SubNode index as well.

    # Compiler directive words (all in range 0x1XX.)        
    elif Command == "NODE":
        Bytes = [0x100]
        VarTypes = "b"
        VarMax = [127]
        ThisNode = ParseVariables(Line[1:], VarTypes, VarMax)
        if WaitingForEndNode:
            print("Error on line", LineNum,
                  ": two NODEs before ENDNODE.")
            sys.exit(1)
        else:
            WaitingForEndNode = True
        # Check against last node number.
        if len(NodeNumbers) == 0 or ThisNode[0] > NodeNumbers[-1]:
            NodeNumbers.extend(ThisNode)
        else:
            print("Error at line", LineNum,
                  ": Node Number not greater than previous one.")
            sys.exit(1)    
        print("Node:", NodeNumbers[-1])
        # Check that file has not contained any SubNodes so far.
        if SubNodesOnly:            
            print("Error at line", LineNum,
                  ": Node after SubNode.")
            sys.exit(1)        
        # Fill spaces in index table for any gap between node numbers.
        if len(NodeNumbers) > 1:
            NodeSpacing = NodeNumbers[-1] - NodeNumbers[-2]
            NodeSpacing -=1
            while NodeSpacing > 0:
                OnEntryIndex.append(ByteLength)
                OnCurrentIndex.append(ByteLength)
                OnExitIndex.append(ByteLength)
                NodeSpacing -=1
                    
    elif Command == "ENDNODE":
        Bytes = [0x101]
        if WaitingForEndNode:
            WaitingForEndNode = False
        else:
            print("Error on line", LineNum,
                  ": ENDNODE before NODE.")
            sys.exit(1)
        # Check, then reset, ONSTATE and ENDSTATE flags.
        if WaitingForEndState:
            print("Error in Node", NodeNumbers[-1], ": missing ENDSTATE.")
            sys.exit(1)
        else:
            WaitingForEndState = False
        if HasOnEntryNode:
            HasOnEntryNode = False
        else:
            print("Error in Node", NodeNumbers[-1], ": no ENTRY state.")
            sys.exit(1)
        if HasOnCurrentNode:
            HasOnCurrentNode = False
        else:
            print("Error in Node", NodeNumbers[-1], ": no CURRENT state.")
            sys.exit(1)
                    
    elif Command == "SUBNODE":
        Bytes = [0x102]
        VarTypes = "b"
        VarMax = [127]
        ThisSub = ParseVariables(Line[1:], VarTypes, VarMax)
        if WaitingForEndNode:
            print("Error on line", LineNum,
                  ": SUBNODE before ENDNODE.")
            sys.exit(1)
        if WaitingForEndSub:
            print("Error on line", LineNum,
                  ": two SUBNODEs before ENDSUB.")
            sys.exit(1)
        WaitingForEndSub = True
        SubNodesOnly = True
        SubNodeIndex.append(ByteLength)
        # Check against last subnode number.
        if len(SubNumbers) == 0 or ThisSub[0] > SubNumbers[-1]:
            SubNumbers.extend(ThisSub)
        else:
            print("Error at line", LineNum,
                  ": SubNode Number smaller than previous one.")
            sys.exit(1)    
        print("SubNode:", SubNumbers[-1])        
        # Fill spaces in index table for any gap between subnode numbers.
        if len(SubNumbers) > 1:
            SubSpacing = SubNumbers[-1] - SubNumbers[-2]
            SubSpacing -=1
        elif len(SubNumbers) == 1:
            SubSpacing = SubNumbers[0]
        else:
            SubSpacing = 0
        while SubSpacing > 0:
            SubNodeIndex.append(ByteLength)
            SubSpacing -=1
        
    elif Command == "ONSTATEENTRY":
        Bytes = [0x103]
        if not WaitingForEndState:
            if HasOnEntryNode:
                print("Error on line", LineNum,
                      ": ONSTATEENTRY previously defined in current node.")
                sys.exit(1)
            else:
                HasOnEntryNode = True
                WaitingForEndState = True
                OnEntryIndex.append(ByteLength)
        else:
            print("Error on line", LineNum,
                  ": two ONSTATEs before ENDSTATE.")
            sys.exit(1)
        
    elif Command == "ONSTATECURRENT":
        Bytes = [0x104]
        if not WaitingForEndState:
            if HasOnCurrentNode:
                print("Error on line", LineNum,
                      ": ONSTATECURRENT previously defined in current node.")
                sys.exit(1)
            else:
                HasOnCurrentNode = True
                WaitingForEndState = True
                OnCurrentIndex.append(ByteLength)
        else:
            print("Error on line", LineNum,
                  ": two ONSTATEs before ENDSTATE.")
            sys.exit(1)

    # Unrecognised instruction or directive.
    else:
        print("Command not found!", Line)
        return

    #
    if Bytes[0] < 0x100:
        ByteLength += len(Bytes)
        BytesOut.extend(Bytes)

# 1. Check for file name in argument, try opening it (or default if no arg.)

if len(sys.argv)==1:
    print("No source filename supplied, trying default", DefaultNodesFile)
    TextIn=DefaultNodesFile
else:
    TextIn=sys.argv[1]

if os.path.isfile(DefaultNodesFile):
    FileIn=open(TextIn, "rt")
else:
    print("Error:", DefaultNodesFile, "not found.")
    sys.exit(1)

# 2. Read the source file.
#    - print the file name (confirming opened successfully.)
#    - read all the lines.

print(TextIn, " has been opened successfully.")
InLines = FileIn.readlines()
FileIn.close()


# 3. Parse each line.
#    - check for whitespace, or whitespace then ";" (in which case, skip line.)
#    - check for string followed by whitespace (command)
#      - get array of numbers until \n or ";" (ignore commas and extra spaces.)
#      - check string for match with known commands.
#        - if match, check array of numbers correspond.
#          - if all fine, return array of values and byte length for each one.
#          - variables will track ONNODE..\IF\IFKEY are balanced with ENDs.

LineNum = 1
ByteLength = 0

BytesOut = []
NodeNumbers = []
SubNumbers = []

OnEntryIndex = []
HasOnEntryNode = False
OnCurrentIndex = []
HasOnCurrentNode = False
SubNodeIndex = []
WaitingForEndState = False
WaitingForEndNode = False
WaitingForEndSub = False
SubNodesOnly = False

for Line in InLines:
    CleanLine=Line.lstrip() # removes leading spaces, tabs, and newlines.
    CleanLine=CleanLine.strip("\n") # remove end newline.
    CleanLine=CleanLine.partition(";")[0] # discard everything after ";".
    if len(CleanLine)>0: # ignore lines which are now empty.
        ItemNum=0
        LineList=CleanLine.split()
        for Item in LineList:
            LineList[ItemNum]=Item.strip(",")
            ItemNum+=1
        ParseLine(LineList)
        # Check for compiler directives.
    LineNum+=1

# Can't catch this in ParseLine()!
if WaitingForEndNode:
    print("Error: final node in file has no ENDNODE.")
    sys.exit(1)


# 4. Get the hexidecimal start address for the output data.
print("Size of compiled Node data:", ByteLength)    
NodeStart = input("Please enter the Node data hex start address:")
NodeStart = int(NodeStart, 16)

if NodeStart < 0 or NodeStart > 65535:
    print("Error: data start address", TokenStart, "is outside memory range.")
    sys.exit(1)
if (NodeStart+ByteLength) > 65535:
    print("Error: data too large for space following start address")
    sys.exit(1)

# 5. Offset each entry in the index tables by the start address.
#    - Node tables will be identical lengths (or earlier error checking failed!)
for i in range(len(OnEntryIndex)):
    OnEntryIndex[i]   = OnEntryIndex[i]   + NodeStart
    OnCurrentIndex[i] = OnCurrentIndex[i] + NodeStart
# SubNode table may be a different length.
for i in range(len(SubNodeIndex)):
    SubNodeIndex[i]   = SubNodeIndex[i]   + NodeStart

# 6. Write the data unformatted to .bin files.
BinOutName=TextIn[0:-3]+"bin"
BinOut=open(BinOutName, "wb")
BinOut.write(bytes(BytesOut))
BinOut.close()

BinOutName=TextIn[0:-4]+"-I-ENT.bin"
BinOut=open(BinOutName, "wb")
for Index in OnEntryIndex:
    BinOut.write(Index.to_bytes(2, byteorder='little'))
BinOut.close()

BinOutName=TextIn[0:-4]+"-I-CUR.bin"
BinOut=open(BinOutName, "wb")
for Index in OnCurrentIndex:
    BinOut.write(Index.to_bytes(2, byteorder='little'))
BinOut.close()

BinOutName=TextIn[0:-4]+"-I-SUB.bin"
BinOut=open(BinOutName, "wb")
for Index in SubNodeIndex:
    BinOut.write(Index.to_bytes(2, byteorder='little'))
BinOut.close()

print("Binary files written.")

# For testing only!
#print(BytesOut)
#print(OnEntryIndex)
#print(OnCurrentIndex)
#print(OnExitIndex)
#print(SubNodeIndex)

# 7. Done
sys.exit(0)
