import os.path, re, sys, time
import txtlib, zxasmlib
import ZX_Text_Compiler_config as cfg
from collections import Counter

gTokens = []
gBanks = []
gWordList = []
gDictList = []
gStringList = []

gDataSaved = 0
gDataOffset = cfg.TokenIndexOrg

# Substitutes all dictionary words for corresponding code.
def DictToEsc(line, SubWordCoding=False):
    for i, word in enumerate(gDictList):
        # Ignore padding entries.
        if word=="":
            pass
        else:
            code="\\" + str(i+128) + "\\"
            if SubWordCoding:
                line=line.replace(word, code)
            else:
                line=re.sub(r"\b"+word+r"\b", re.escape(code), line)
    return line   

# Copy text tokens from file.
tfile=cfg.TokenFile+cfg.FileInExt
if os.path.isfile(tfile):
    gTokens=txtlib.GetLines(tfile)
else:
    print("Error:", tfile, "not found.")
    sys.exit(1)

# Check number of tokens.
tklen=len(gTokens)
if tklen > cfg.MaxTokens:
    print("Error:", tklen, "tokens in file exceeds maximum of", cfg.MaxTokens, ".")
    sys.exit(1)
# Add one null token entry (for length calculation.)
gTokens.append("")

# Save tokens and index list.
index=cfg.TokenFile + cfg.IndexExt + cfg.AsmOutExt
data=cfg.TokenFile + cfg.AsmOutExt
indexFile=open(index, "wt")
dataFile=open(data, "wt")
# Adjust offset by index size (two bytes per entry.)
gDataOffset+=(len(gTokens)*2)
for i, line in enumerate(gTokens):
    indexFile.write("  DW " + str(gDataOffset) + "\n")
    dataFile.write('  DB "'+line+'" ;'+str(i)+"\n")
    gDataOffset+=len(line)    
indexFile.close()
dataFile.close()

# Copy text banks from file, filling gaps with null entries.
gapCount=0
for file in range(cfg.MaxBanks):
    filename = str(file).zfill(3) + cfg.FileInExt
    if os.path.isfile(filename):
        for g in range(gapCount):
            gBanks.append([])
        gBanks.append(txtlib.GetLines(filename))
        gapCount=0
    else:
        gapCount+=1        
# Add one null bank entry (for length calculation.)
gBanks.append([])

# Generate list of all words in banks, ignoring escape codes.
for iB, bank in enumerate(gBanks):
    for iL, line in enumerate(bank):  
        # Check line length.
        if len(line)>cfg.MaxLineLength:
            print("Error in bank", iB, ".")
            print("Line", iL, ", as follows, exceeds maximum allowed length:")
            print(line)
            sys.exit(1)
        # Get all words from line.
        line=txtlib.StripCodes(line)
        line=txtlib.StripPunct(line)
        gWordList.extend(line.split())

# Generate list of unique words, and data saved by encoding each one.
dictionary=sorted(set(gWordList))
savingList=[]
for word in dictionary:
    frequency=gWordList.count(word)
    saving=txtlib.CompressionSaving(word)*(frequency-1)
    savingList.append([word, saving])
    
# Trim list, then calculate overall data saving.
savingList=sorted(savingList, key=lambda i: i[1], reverse=True)
for i in range (cfg.MaxDict):
    if savingList[i]:
        gDictList.append(savingList[i][0])
        gDataSaved+=savingList[i][1]
    else:
        break
# Add null entry (for length calculation.)
gDictList.append("")
print("Dictionary tokenisation will save", gDataSaved, "bytes.")

# Save dictionary and index list.
index=cfg.DictFile + cfg.IndexExt + cfg.AsmOutExt
data=cfg.DictFile + cfg.AsmOutExt
indexFile=open(index, "wt")
dataFile=open(data, "wt")
# Adjust offset by index size (two bytes per entry.)
gDataOffset+=(len(gDictList)*2)
for i, line in enumerate(gDictList):
    indexFile.write("  DW "+str(gDataOffset)+"\n")
    dataFile.write('  DB "'+line+'" ;'+str(i)+"\n")
    gDataOffset+=len(line)    
indexFile.close()
dataFile.close()

# Add null entry to each bank (for length calculation.)
for bank in gBanks:
    bank.append("")

# Banks and Lines are written in the same pass.
indexBanks=cfg.BanksFile + cfg.IndexExt + cfg.AsmOutExt
indexLines=cfg.LinesFile + cfg.IndexExt + cfg.AsmOutExt
dataBanks=cfg.BanksFile + cfg.AsmOutExt
dataLines=cfg.LinesFile + cfg.AsmOutExt
iBanksFile=open(indexBanks, "wt")
iLinesFile=open(indexLines, "wt")
dBanksFile=open(dataBanks, "wt")
dLinesFile=open(dataLines, "wt")

# Track Banks and Lines offsets seperately.
bankOffset=gDataOffset
# Adjust offset by index size (two bytes per entry.)
bankOffset+=(len(gBanks)*2)
lineOffset=bankOffset
for iB, bank in enumerate(gBanks):
    # Adjust offset by total size of indexes (two bytes per entry.)
    lineOffset+=(len(bank)*2)

# GOT TO HERE.

for iB, bank in enumerate(gBanks):
    print("Bank:", iB)
    # Header text and Banks index.
    iBanksFile.write("; Bank " + str(iB) + " index start:\n")
    iBanksFile.write("  DW " + str(bankOffset) + "\n\n")
    dBanksFile.write("; Bank " + str(iB) + " index:\n")
    dLinesFile.write("; Bank " + str(iB) + ":\n")
    for iL, line in enumerate(bank):
        # Tokenise all lines.
        line=txtlib.EncodeVarRefs(line)
        line=DictToEsc(line)
        if cfg.SubWordCoding:
            line=DictToEsc(line, True)
        line=txtlib.SpacePairToEsc(line)
        lineBytes=zxasmlib.LineToBytes(line)
        # Write tokenised line and index reference.
        dBanksFile.write("  DW " + str(lineOffset) + "\n")
        bankOffset+=2
        dLinesFile.write(zxasmlib.BytesToLine(lineBytes, iL))
        lineOffset+=len(lineBytes)
        
    dBanksFile.write("\n")
    dLinesFile.write("\n")
    bankZero=False

iLinesFile.close()
iBanksFile.close()
dLinesFile.close()
dBanksFile.close()

# Done.
sys.exit(0)
